/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;
import hypercast.events.*;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

/** Class containing fields for an entry in the neighbor table. 
 * @see DT_Neighborhood
 * @author HyperCast Team
 * @version 2005 (version 3.0)
*/

final class DT_NeighborhoodEntry 
{
    /** the neighbor. */
    private DT_AddressPair neighbor;
    /** the neighbor's neighbor that is immediate CW of this node. */
    private DT_AddressPair CW;
    /** the neighbor's neighbor that is immediate CCW of this node. */
    private DT_AddressPair CCW;
    /** This is true if the neighbor has an entry for this node in its neighbor table. */
    private boolean hasThisNodeAsNeighbor;
    /** This is the time the last message was received from the neighbor. */
    private long lastContactTime;
    
    /** constructor */
    DT_NeighborhoodEntry(DT_AddressPair n, DT_AddressPair cw, DT_AddressPair ccw, boolean h, long t) 
    {   
        neighbor = n;
        CW = cw;
        CCW = ccw;
        hasThisNodeAsNeighbor = h;
        lastContactTime = t;
    }

    DT_AddressPair getNeighbor()
    {
        return neighbor;
    }
    
    DT_AddressPair getCW()
    {
        return CW;
    }

    void setCW(DT_AddressPair cw)
    {
        CW = cw;
    }
    
    DT_AddressPair getCCW()
    {
        return CCW;
    }
    
    void setCCW(DT_AddressPair ccw)
    {
        CCW = ccw;
    }

    boolean hasThisNodeAsNeighbor() 
    {
        return hasThisNodeAsNeighbor;
    }

    void setHasThisNodeAsNeighbor(boolean h)
    {
        hasThisNodeAsNeighbor = h;
    }
    
    long getLastContactTime() 
    {
        return lastContactTime;
    }

    void setLastContactTime(long t) 
    {
        lastContactTime = t;
    }
}


/** This class keeps track of the Neighborhood table and LogicalAddress
 * of a node in the Delaunay triangulation.  It also provides a large variety
 * of helper functions.  
 * <p>
 * The neighbors are kept unsorted in a Vector. The neighbors can be accessed
 * either through their index in the Vector, or by the neighbor's Physical Address.
 * For each neighbor, DT_Neighborhood stores an AddressPair and the time
 * the most recent message was received from that neighbor.
 * <p>
 * DT_Neighborhood maintains a property on the neighbors.  This property is:
 * In a Delaunay triangulation of the node's Logical Address and its neighbors'
 * Logical Addresses, there would be an edge connecting the node's LA to each
 * of its neighbor's LAs.  When a node is inserted causes this property to be 
 * violate, neighbors are removed to restore the property.
 *
 * For any problems, contact Jorg Liebeherr at jorg@cs.virginia.edu
 * @author Michael Nahas
 * @version	2.0, Feb. 11, 2001
 */

final public class DT_Neighborhood
{
    /** Source of constants & error/log message display */
    private HyperCastConfig config;

    /** Network adapter (used for times and Physical Address). */
    private I_UnicastAdapter adapter; 

    /** The physical and logical address of this node. */
    private DT_AddressPair myAddressPair;

    /** NeighborhooEntries of the node's neighbors.  This is
     unordered.  If the neighbors were sorted by angleToTheHorizon,
     many functions implementations could be significantly sped up.
     @see DT_NeighborhoodEntry */
    private Vector neighbors;

	private String addNeighborDebug;

	/**
	 * Neighborhood table change flag. It is set to true when 
	 * nodes are added to or removed from the neighbor table.
	 * It is reset to false when DT node checks and handles (the
	 * flag is used to decide if a neighborhood table stable event) 
	 * the change in the neighbor table.
	 */
	private boolean neighborTableChange = true;


	/** Notification handler. When interesting events happen, this DT_Neighborhood
	 * object reports the events to the notification handler if it is not null.
	 */
	protected NotificationHandler n_handler = null;
	
	/**
	 * The reference to the StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;

	/**
	 * Statistics name for neighbor.
	 */
	String statsName = "NeighborTable";

  
	/** Constructor.
	 * @param startingCoords The starting address of this node.
	 */
public DT_Neighborhood(HyperCastConfig c, I_UnicastAdapter ua, DT_LogicalAddress startingCoords, StatsProcessor sPro) {
	
		config = c;
		adapter = ua;

		myAddressPair = new DT_AddressPair(adapter.createPhysicalAddress(), startingCoords);

		//config.log.println("DT_Neighborhood: create myAddressPair: " + myAddressPair.toString());

		neighbors = new Vector();

		statsPro = sPro;
        
		InitStatisticsStructure();
	}


	/** Returns the DT_NeighborhoodEntry instance which has given index in the table. */
 private DT_NeighborhoodEntry getEntry(int i)
    {
        return (DT_NeighborhoodEntry) neighbors.elementAt(i);
    }



	/** Returns the address pair of neighbor with given index. */
 public DT_AddressPair getNeighborAddressPair(int index) 
	{
		if (index <= neighbors.size())
		{
			return this.getEntry(index).getNeighbor();
		}
		else
		{
			return null;
		}
	}


	/** Returns the address pair of a CW neighbor correspondin to the neighbor with given index. */
public DT_AddressPair getCWNeighborAddressPair(int index) 
   {
	   if (index <= neighbors.size())
	   {
		   return this.getEntry(index).getCW();
	   }
	   else
	   {
		   return null;
	   }
   }

   /** Returns the address pair of a CCW neighbor correspondin to the neighbor with given index. */
public DT_AddressPair getCCWNeighborAddressPair(int index) 
  {
	  if (index <= neighbors.size())
	  {
		  return this.getEntry(index).getCCW();
	  }
	  else
	  {
		  return null;
	  }
  }

    
  /** Returns address pair of this node. */
public DT_AddressPair getMyAddressPair() 
{
    return myAddressPair;
}

/** Returns logical address of this node. */
public DT_LogicalAddress getLogicalAddress() 
    {
        return myAddressPair.getDTLogicalAddress();
    }
    
/** Returns physical address of this node. */
public I_PhysicalAddress getPhysicalAddress() 
    {
        return myAddressPair.getPhysicalAddress();
    }

/** Returns the number of neighbors.  This is useful when
 * calling getNeighborAtIndex() */
public final int getNumOfNeighbors() {
    return neighbors.size();
}



/** 
 * Returns the size of neighborhood table.
 */
public final int getTableSize() {
	return neighbors.size();
}



/** Returns true if there are no neighbors. */
public boolean isEmpty() {
        return neighbors.isEmpty();
    }

/** Returns true if no neighbor's Logical Address is greaterThan this node's LogicalAddress.
 */
public boolean isLeader() 
    {
        for(int i = 0; i < neighbors.size(); i++)
        {
            DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
            if (neighborAt_i.getDTLogicalAddress().greaterThan(myAddressPair.getDTLogicalAddress()))
                return false;
        }
        return true;
    }
    
/** Returns true every neighbor's CW and CCW entry are in the neighborhood. */
public boolean isStable() 
    {
        for (int i = 0; i < neighbors.size(); i++) 
        {
//            if (!getEntry(i).hasThisNodeAsNeighbor())
//                return false;
            
            DT_AddressPair cw = getEntry(i).getCW();
            DT_AddressPair my_ccw = getCCWNeighbor(getEntry(i).getNeighbor().getDTLogicalAddress());
            if ((cw != null && my_ccw == null) ||
                (cw == null && my_ccw != null) ||
                (cw != null && my_ccw != null && !cw.getPhysicalAddress().equals(my_ccw.getPhysicalAddress())))
                    return false;
            
            DT_AddressPair ccw = getEntry(i).getCCW();
            DT_AddressPair my_cw = getCWNeighbor(getEntry(i).getNeighbor().getDTLogicalAddress());
            if ((ccw != null && my_cw == null) ||
                (ccw == null && my_cw != null) ||
                (ccw != null && my_cw != null && !ccw.getPhysicalAddress().equals(my_cw.getPhysicalAddress())))
                    return false;
        }

		/*
		 * If a node has no neighbor, it is considered stable.
		 */
        return true;
    }
    

  /**
   * Sets notification handler.
   */	
  public void setNotificationHandler(NotificationHandler nh)
  {
	  n_handler = nh;
  }

  /** Return the neighbor table change flag. */
  public boolean isNeighborTableChanged()
	  {
		  return neighborTableChange;
	  }

  /** Sets the neighbor table change flag. */
  public void setNeighborTableChangeFlag(boolean changeFlag)
	  {
		neighborTableChange = changeFlag;
	  }


    
/** Returns the Nth neighbor's AddressPair.  The neighbors are not ordered. */
public DT_AddressPair getNeighborAtIndex(int i)
    {
        return getEntry(i).getNeighbor();
    }

/** Returns the Nth neighbor's CW neighbor's AddressPair. */
public DT_AddressPair getNodeCWOfNeighborAtIndex(int i)
    {
        return getEntry(i).getCW();
    }

/** Returns the Nth neighbor's CCW neighbor's AddressPair. */
public DT_AddressPair getNodeCCWOfNeighborAtIndex(int i)
    {
        return getEntry(i).getCCW();
    }

/** Returns if Nth neighbor has this node as a neighbor. */
public boolean neighborAtIndexHasThisNodeAsNeighbor(int i)
    {
        return getEntry(i).hasThisNodeAsNeighbor();
    }


/** Returns the Nth neighbor's LastContactTime (the last time a
    message was received from that neighbor).  The neighbors are not
    ordered. */
public long getLastContactTimeOfNeighborAtIndex(int i) 
    {
        return getEntry(i).getLastContactTime();
    }
    

/** Returns true if one of the neighbors has a phyiscal address equal to the parameter.
 * Recall, a physical address uniquely identifies a node.
 */
public boolean contains(I_PhysicalAddress addr) {
        return indexOfNeighbor(addr) >= 0;
    }

/** Returns true if one of the neighbors has a logical address equal to the parameter.
 * Recall, a physical address uniquely identifies a node.
 */
public boolean contains(I_LogicalAddress addr) {
        return indexOfNeighbor(addr) >= 0;
    }
  
/** Finds the index of the neighbor whose physical address equals the parameter.
 * Recall, a physical address uniquely identifies a node.
 */
public int indexOfNeighbor(I_PhysicalAddress addr) {

        for (int i = 0; i < neighbors.size(); i++) 
        {
            DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
            if ( neighborAt_i.getPhysicalAddress().equals(addr) )
                return i;
        }

        return -1;
    }


/** Returns true if one of the neighbors has a phyiscal address equal to the parameter.
 */
public boolean contains(I_NetworkAddress addr) {
		return indexOfNeighbor(addr) >= 0;
	}

/** Finds the index of the neighbor whose physical address equals the parameter.
 */
public int indexOfNeighbor(I_NetworkAddress addr) 
	{
		for (int i = 0; i < neighbors.size(); i++) 
		{
			DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
			if ( neighborAt_i.getPhysicalAddress().equals(addr) )
				return i;
		}

		return -1;
	}



/** Finds the index of the neighbor whose logical address equals the parameter.
 * Recall, a logical address uniquely identifies a node.
 */
public int indexOfNeighbor(I_LogicalAddress addr) {

		for (int i = 0; i < neighbors.size(); i++) 
		{
			DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
			if ( neighborAt_i.getLogicalAddress().equals(addr) )
				return i;
		}

		return -1;
}
    

  
/** Returns the AddressPair of the neighbor whose physical address equals the parameter.
 * Recall, a physical address uniquely identifies a node.
 */
public DT_AddressPair getNeighborWithThisPA(I_PhysicalAddress addr) {
    int i = indexOfNeighbor(addr);
    if (i < 0)
        return null;
    else
        return getEntry(i).getNeighbor();
}

/** Returns the AddressPair of the neighbor whose logical address equals the parameter.
* Recall, a logical address uniquely identifies a node.
*/
public DT_AddressPair getNeighborWithThisLA(I_LogicalAddress addr) {
   int i = indexOfNeighbor(addr);
   if (i < 0)
	   return null;
   else
	   return getEntry(i).getNeighbor();
} 

/** Returns true if there exists a neighbor with a LogicalAddress equal to the parameter. */
public boolean neighborHasThisLA(DT_LogicalAddress la)
    {
        for(int i = 0; i < neighbors.size(); i++)    
        {
            DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
            if (la.equals(neighborAt_i.getDTLogicalAddress()))
                return true;
        }

        return false;
    }


/** Returns AddressPair of neighbor that is parent of this node in a spanning tree
 * rooted at <code>rootCoords</code>.
 * This is calculated by finding the neighbor that forms the smallest angle with <code>rootCoords</code>.
 */
public DT_AddressPair getParent(DT_LogicalAddress rootCoords) 
{
    if (rootCoords.equals(myAddressPair.getDTLogicalAddress()))
        return null;

    DT_AddressPair nodeAtAngle = getNeighborAtSameAngle(rootCoords);
    if (nodeAtAngle != null)
        return nodeAtAngle;

    DT_AddressPair CW_Neigh  = getCWNeighbor(rootCoords);
    DT_AddressPair CCW_Neigh = getCCWNeighbor(rootCoords);


    if (CW_Neigh == null && CCW_Neigh == null)
        return null;
    else if (CW_Neigh == null)
        return (CCW_Neigh);
    else if (CCW_Neigh == null)
        return (CW_Neigh);


    double  CW_angle = Math.abs( DT_Point.angleRelative(CW_Neigh.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), rootCoords));
    double CCW_angle = Math.abs( DT_Point.angleRelative(CCW_Neigh.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), rootCoords));

    if( CW_angle == CCW_angle ) 
    {
        if (CW_Neigh.getDTLogicalAddress().greaterThan(CCW_Neigh.getDTLogicalAddress()))
            return CW_Neigh;
        else
            return CCW_Neigh;
    }
    else if (CW_angle < CCW_angle)
        return CW_Neigh;
    else
        return CCW_Neigh;
}


/** Return the AddressPairs of neighbor that are children of this node in a spanning tree
 * rooted at <code>rootCoords</code>.
 * This based on if DT_Neighborhood.getParent() of the neighbor would return this node.
 */
public DT_AddressPair[] getChildren(DT_LogicalAddress root) 
{
    Vector children = new Vector(neighbors.size());

    for (int i = 0; i < neighbors.size(); i++)
    {
        if (amIParent(getEntry(i), root))
            children.addElement(getEntry(i).getNeighbor());
    }
    
    DT_AddressPair[] array = new DT_AddressPair[children.size()];
    children.copyInto(array);

    return array;
}


/** Returns true if this node is the parent of the neighbor <code>node</code> 
 * in a spanning tree rooted at <code>root</code>
 * @see #getParent(DT_LogicalAddress)
 */
private boolean amIParent(DT_NeighborhoodEntry nodeEntry, DT_LogicalAddress root)
    {
        DT_AddressPair CW_Neigh  = getCWNeighbor(nodeEntry.getNeighbor().getDTLogicalAddress());
        DT_AddressPair CCW_Neigh = getCCWNeighbor(nodeEntry.getNeighbor().getDTLogicalAddress());

        if (nodeEntry.getNeighbor().getDTLogicalAddress().equals(root))
            return false;

        double this_angle = Math.abs( DT_Point.angleRelative(root, nodeEntry.getNeighbor().getDTLogicalAddress(), myAddressPair.getDTLogicalAddress()));

        if (CW_Neigh != null) {
            double CW_angle = Math.abs( DT_Point.angleRelative(root, nodeEntry.getNeighbor().getDTLogicalAddress(), CW_Neigh.getDTLogicalAddress()));
            
            if (CW_angle < this_angle || (CW_angle == this_angle && CW_Neigh.getDTLogicalAddress().greaterThan(myAddressPair.getDTLogicalAddress())))
            {
                return false;
            }
        }
        
    
        if (CCW_Neigh != null) {
            double CCW_angle = Math.abs( DT_Point.angleRelative(root, nodeEntry.getNeighbor().getDTLogicalAddress(), CCW_Neigh.getDTLogicalAddress()));
            if (CCW_angle < this_angle || (CCW_angle == this_angle && CCW_Neigh.getDTLogicalAddress().greaterThan(myAddressPair.getDTLogicalAddress())))
            {
                return false;
            }
        }
        
	if ((CW_Neigh == null) && (CCW_Neigh == null)) {
	   double dist_thisnode = DT_Point.distance(root, myAddressPair.getDTLogicalAddress());
	   double dist_neighbor = DT_Point.distance(root, nodeEntry.getNeighbor().getDTLogicalAddress());
	   if (dist_thisnode < dist_neighbor) {
              return true;
	   } else {
              return false;
	   }
	}

        return true;

}


/** Returns true if this node is the parent of the neighbor <code>node</code> 
 * in a spanning tree rooted at <code>root</code>
 * @see #getParent(DT_LogicalAddress)
 */
public boolean amIParent(DT_LogicalAddress node, DT_LogicalAddress root)
{
	int ind = indexOfNeighbor(node);
	if ( ind < 0) 
	   return false;
	   
	return (amIParent(getEntry(ind), root));
}
    

/** Returns the AddressPairs of all the neighbors 
 */

public DT_AddressPair[] getAllNeighbors() 
{
    DT_AddressPair[] array = new DT_AddressPair[neighbors.size()];
    for (int i = 0; i < neighbors.size(); i++)
        array[i] = getEntry(i).getNeighbor();
    return array;
}


/** Returns the neighbor with the greatest coordinates.
 * @return null if this node's LogicalAddress is greater than all its neighbors.
 */
public DT_AddressPair getNextHopToLeader() 
{
    DT_AddressPair nextHop = null;

    for(int i = 0; i < neighbors.size(); i++)
    {
        DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
        if (neighborAt_i.getDTLogicalAddress().greaterThan(myAddressPair.getDTLogicalAddress()) && 
            (nextHop == null || neighborAt_i.getDTLogicalAddress().greaterThan(nextHop.getDTLogicalAddress())))
            nextHop = neighborAt_i;
    }   

    return nextHop;
}




/** Returns true if this node is at least as close to the point <code>p</code> as any of its neighbors.
 * @return true if there exists a neighbor of equal distance.
 * NEEDED ONLY FOR MULTICAST VERSIONS
 */
public boolean isThisClosest(DT_Point p) 
     {
         double mydist = DT_Point.distance(myAddressPair.getDTLogicalAddress(), p);
    
         for(int i = 0; i < neighbors.size(); i++)
         {
             DT_Point p2 = (getEntry(i).getNeighbor()).getDTLogicalAddress();
             if (DT_Point.distance(p, p2) < mydist) 
                 return false;
         }

         return true;
     }


/** Returns true if any 4 neighbors lie on a circle
 */

public boolean hasFourOnCircle() 
    {
        for (int i = 0; i < neighbors.size(); i++) 
        {
            DT_LogicalAddress la = getNeighborAtIndex(i).getDTLogicalAddress();
            
            DT_AddressPair cw =  getCWNeighbor(la);
            DT_AddressPair ccw = getCCWNeighbor(la);

            if (ccw == null || cw == null)
                continue;

            // if circle bends backwards...  (i.e. not even possible to be on circle.)
            if (DT_Point.angleRelative(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress()) < 0)
                continue;
        
            if (DT_Point.isOnCircle(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress(), la))
                return true;
        }
        return false;
    }



/** Returns true if the node with LogicalAddress <code>la</code> can be added as a neighbor.
 */

public boolean shouldBeNeighbor(DT_LogicalAddress la) {

    if (myAddressPair.getDTLogicalAddress().equals(la))
        return true;

    DT_AddressPair sa = getNeighborAtSameAngle(la);
    if (sa != null) {
        // return true if the parameter is closer than the current neighbor at that angle.
        if (DT_Point.distance(myAddressPair.getDTLogicalAddress(), la) <= DT_Point.distance(myAddressPair.getDTLogicalAddress(), sa.getDTLogicalAddress()))
            return true;
        else 
            return false;
	}

	DT_AddressPair cw  =  getCWNeighbor(la);
	DT_AddressPair ccw = getCCWNeighbor(la);

    // return true if there is no clockwise and counterclockwise neighbors
	if (ccw == null || cw == null)
	    return true;

    // return true if angle between CW & CCW neighbors is greater than 180 degrees
	if (DT_Point.angleRelative(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress()) < 0)
	    return true;


    // return true if parameter lies inside circle inscribing this node, CW, and CCW neighbors.
	if (DT_Point.isInOrOnCircle(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress(), la)) 
	    return true;
	else 
	    return false;
}



/** Returns neighbors that lies as the same angle as the parameter with respect to this node.
 * If this returns a node, then this node, parameter, and returned node lie on a line.
 */
public DT_AddressPair getNeighborAtSameAngle(DT_LogicalAddress la)
    {
        for(int i = 0; i < neighbors.size(); i++)    
        {
            DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
            if (DT_Point.angleRelative(la, myAddressPair.getDTLogicalAddress(), neighborAt_i.getDTLogicalAddress()) == 0.0)
                return neighborAt_i;
        }

        return null;
    }
    

/** Corrective value for double arithmetic when calculating CW and CCW neighbors. */
private static double MIN_ANGLE = 1e-15;
    
/** Finds the neighbor with the CCW smallest angle to the parameter with respect to this node.
 * The angle must be greater than <code>MIN_ANGLE</code>
 * @see DT_Neighborhood#MIN_ANGLE
 */
public DT_AddressPair getCCWNeighbor(DT_LogicalAddress la)
{
    double bestAngle = Math.PI - MIN_ANGLE;
    DT_AddressPair bestNeighbor = null;
    for(int i = 0; i < neighbors.size(); i++)    
    {
        DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
        double theta = DT_Point.angleRelative(la, myAddressPair.getDTLogicalAddress(), neighborAt_i.getDTLogicalAddress());
        if (theta > MIN_ANGLE && theta < bestAngle)
        {
            bestAngle = theta;
            bestNeighbor = neighborAt_i;
        }
    }


    return bestNeighbor;
}

/** Finds the neighbor with the CCW smallest angle to the parameter with respect to this node.
 * The angle must be greater than <code>MIN_ANGLE</code>
 * @see DT_Neighborhood#MIN_ANGLE
 */
public DT_AddressPair getCWNeighbor(DT_LogicalAddress la)
{
    double bestAngle = -Math.PI + MIN_ANGLE;
    DT_AddressPair bestNeighbor = null;

    for(int i = 0; i < neighbors.size(); i++)     
    {
        DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();
        double theta = DT_Point.angleRelative(la, myAddressPair.getDTLogicalAddress(), neighborAt_i.getDTLogicalAddress());
        if (theta < -MIN_ANGLE && theta > bestAngle)
        {
            bestAngle = theta;
            bestNeighbor = neighborAt_i;
        }
    }

    return bestNeighbor;
}




/** Test Function.
 * @return index of neighbor that is part of inconsistency.
 */
public int isConsistent() 
{
    int result = -1;
    
    for(int i = 0; i < neighbors.size(); i++) 
    {
        DT_AddressPair neighborAt_i = getEntry(i).getNeighbor();

        DT_AddressPair sa = getNeighborAtSameAngle(neighborAt_i.getDTLogicalAddress());
        if (sa != neighborAt_i)
        {
            config.err.println("Another node has the same angle at index "+i+"!");
            result = i;
            continue;
        }


        DT_AddressPair ccw = getCCWNeighbor(neighborAt_i.getDTLogicalAddress());
        DT_AddressPair cw =  getCWNeighbor(neighborAt_i.getDTLogicalAddress());
        if (ccw == null || cw == null)
            continue;

        double angle = DT_Point.angleRelative(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress());
//        config.err.println("Angle is " + ((angle*180)/Math.PI));
//        config.err.println("\tCCW =\t" + getCoordsOfNeighbor(ccw) + "\t angle relative = " + ((DT_Point.angleRelative(getCoordsOfNeighbor(i), myAddressPair.getDTLogicalAddress(), getCoordsOfNeighbor(ccw))*180)/Math.PI));
//        config.err.println("\ttarget=\t" + getCoordsOfNeighbor(i) );
//        config.err.println("\tCW =\t" + getCoordsOfNeighbor(cw)    + "\t angle relative = " + ((DT_Point.angleRelative(getCoordsOfNeighbor(i), myAddressPair.getDTLogicalAddress(), getCoordsOfNeighbor(cw))*180)/Math.PI));
//        config.err.println("\tmy =\t" + myAddressPair.getDTLogicalAddress());


        if (angle <= 0 || angle >= Math.PI) 
        {
//            config.err.println("Angle less than or equal to 0 or greater or equal to Pi --- marking it consistent.");
            continue;
        }
        else 
        {
//            config.err.println("Angle in range (0,Pi) --- checking circle.");
        }
        
        if (!DT_Point.isInCircle(myAddressPair.getDTLogicalAddress(), cw.getDTLogicalAddress(), ccw.getDTLogicalAddress(), neighborAt_i.getDTLogicalAddress())) 
        {
            config.err.println("Out of Circle at index "+i+"!");
            result = i;
        }
        else 
        {
//            config.err.println("In circle --- marking it consistent.");
        }
            
    }
    
    return result;
}


/** Inserts a neighbor and then removes neighbors to make neighborhood consistent
 */
public void addNeighbor(DT_AddressPair addr, DT_AddressPair cw, DT_AddressPair ccw, boolean hasThisNodeAsNeighbor)
{
	/*
	 * If the neighborhood table still contains the node which will be added again,
	 * delete old record.
	 */
	if (myAddressPair.getPhysicalAddress().addressCheck(addr.getPhysicalAddress()))
	{
		return;
	}

    // String "addNeighborDebug" was used to debug this section using the RC2/RS2 
    // programs.  The RC2 was able to access "addNeighborDebug" through the I_Stats
    // interface.
    addNeighborDebug = "Started Add Neighbor";

    // guarantee my coords do not match new node's
    while ( myAddressPair.getDTLogicalAddress().equals(addr.getDTLogicalAddress()))
    {
		//System.out.println("DT_Neighborhood:addNeighbor: shift coodrs (1).");

        addNeighborDebug = "Shifting coords (1)";
        randomShiftMyCoordinates();
    }
    
    addNeighborDebug = "adding element";
    // add new neighbor
    neighbors.addElement(new DT_NeighborhoodEntry(addr, cw, ccw, hasThisNodeAsNeighbor, adapter.getCurrentTime()));

  	//neighborhoodStats.addNeighbor(addr, cw, ccw);
	this.addNeighborToStatsProcessor(addr, cw, ccw);

    addNeighborDebug = "removing non-neighbors (1)";

    removeNonNeighbors();

    addNeighborDebug = "4 on circle (1)";

    while (hasFourOnCircle()) 
    {
		//System.out.println("DT_Neighborhood:addNeighbor: shift coodrs (2).");

        addNeighborDebug = "shifting (2)";
        randomShiftMyCoordinates();
        addNeighborDebug = "removing non-neighbors (2)";
        removeNonNeighbors();
        addNeighborDebug = "done removing neighbors (2)";
    }

    addNeighborDebug = "returned from Add Neighbor";
}

/** Removes neighbor identified by its PhysicalAddress.
 */
public void removeNeighbor(I_PhysicalAddress addr) {
    int i = indexOfNeighbor(addr);
    if (i >= 0)
    {
		removeNeighbor(i);
    }
}


/** Removes neighbor identified by its index
 */
public void removeNeighbor(int i)
{
	DT_NeighborhoodEntry entry=getEntry(i);
    neighbors.removeElementAt(i);

    //neighborhoodStats.removeNeighbor(entry.getNeighbor());
	statsPro.removeStatsElement(entry.getNeighbor());
	
}

/** Removes neighbor identified by the PhysicalAddress of its AddressPair.
 */
public void removeNeighbor(DT_AddressPair addrs) 
    {
        removeNeighbor(addrs.getPhysicalAddress());
    }
    
/** Remove all neighbors from the Neighborhood. */
public void removeAllNeighbors() 
{
    neighbors.removeAllElements();
    
	statsPro.removeAllStatsElement(statsName);
}



/** Sets the contact time for the neighbor identified by its PhysicalAddress.
 * @return true if the entry has changed
 */
boolean updateContactTable(I_PhysicalAddress addr,  DT_AddressPair cw, DT_AddressPair ccw, boolean hasThisNodeAsNeighbor){
    int i = indexOfNeighbor(addr);

    DT_NeighborhoodEntry ne = getEntry(i);
    
    DT_AddressPair  cwOld = ne.getCW();
    DT_AddressPair ccwOld = ne.getCCW();

    ne.setCW(cw);
    ne.setCCW(ccw);
    ne.setHasThisNodeAsNeighbor(hasThisNodeAsNeighbor);
    ne.setLastContactTime(adapter.getCurrentTime());
    
    boolean cwChanged = (cwOld == null && cw != null) || (cwOld != null && cw == null) || (cwOld != null && cw != null && !(cw.getPhysicalAddress().equals(cwOld.getPhysicalAddress())));

    boolean ccwChanged = (ccwOld == null && ccw != null) || (ccwOld != null && ccw == null) || (ccwOld != null && ccw != null && !(ccw.getPhysicalAddress().equals(ccwOld.getPhysicalAddress())));

    return cwChanged || ccwChanged;
}


/** Changes this node's LogicalAddress by a small amount.
 * Changing the LogicalAddress can make the neighborhood inconsistent.
 */
private void randomShiftMyCoordinates() 
    {
        DT_LogicalAddress myOldCoords = myAddressPair.getDTLogicalAddress();

        Random r = new Random();
        
        do 
        {
            addNeighborDebug = "in shift loop";
            int i = r.nextInt() % 8;
            int dx, dy;
        
            //120
            //4 3
            //756

            if (i < 4)
                dy = 1;
            else if (i > 5)
                dy = -1;
            else
                dy = 0;
        
            if (i%3 == 0)
                dx = 1;
            else if (i%3 == 1)
                dx = -1;
            else
                dx = 0;

            // make sure coordinates do not flip signs
            if ((myAddressPair.getDTLogicalAddress().getX() < 0 && (myAddressPair.getDTLogicalAddress().getX() + dx) > 0) ||
                (myAddressPair.getDTLogicalAddress().getX() > 0 &&  (myAddressPair.getDTLogicalAddress().getX() + dx) < 0))
                dx = 0;
            if ((myAddressPair.getDTLogicalAddress().getY() < 0 && (myAddressPair.getDTLogicalAddress().getY() + dy) > 0) ||
                (myAddressPair.getDTLogicalAddress().getY() > 0 &&  (myAddressPair.getDTLogicalAddress().getY() + dy) < 0))
                dy = 0;

            myAddressPair = new DT_AddressPair( myAddressPair.getPhysicalAddress(), new DT_LogicalAddress(myAddressPair.getDTLogicalAddress().getX() + dx, myAddressPair.getDTLogicalAddress().getY() + dy));

            addNeighborDebug = "going to test in shift loop with dx =" + dx + " and dy = " + dy;

        } while (neighborHasThisLA(myAddressPair.getDTLogicalAddress()) || myAddressPair.getDTLogicalAddress().equals(myOldCoords));
        
        addNeighborDebug = "done shift";
        config.log.println("Shifted node to new coordinates: " + myAddressPair.getDTLogicalAddress().toString());

		if (n_handler != null) {
			NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}
    }


/** Returns the neighborhood to consistency.
 */
private void removeNonNeighbors() 
    {
        final int initialSize = neighbors.size();
        

        addNeighborDebug = "entering RNN outer loop";
OUTER_LOOP:
        for (int i = 0; i < neighbors.size(); i++) 
        {
            DT_AddressPair subject = getEntry(i).getNeighbor();
            
            addNeighborDebug = "entering RNN inner loop";
            for (int j = 0; j < neighbors.size(); j++)
            {
                if (i == j)
                    continue;

                DT_AddressPair neighborAt_j = getEntry(j).getNeighbor();
                if (DT_Point.angleRelative(subject.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), neighborAt_j.getDTLogicalAddress()) == 0.0)
                {
                    if (DT_Point.distance(subject.getDTLogicalAddress(),      myAddressPair.getDTLogicalAddress()) <
                        DT_Point.distance(neighborAt_j.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress()))
                    {
                        removeNeighbor(j);
                        continue OUTER_LOOP;
                    }
                    else 
                    {
                        // cannot be equal.
                        removeNeighbor(i);
                        continue OUTER_LOOP;
                    }
                }
            }

            addNeighborDebug = "exited RNN inner loop";

            DT_AddressPair ccw = getCCWNeighbor(subject.getDTLogicalAddress());
            if (ccw == null) 
                continue;
            DT_AddressPair cw = getCWNeighbor(subject.getDTLogicalAddress());
            if (cw == null) 
                continue;

            // if circle bends backwards...  (i.e. not even possible to be on circle.)
            if (DT_Point.angleRelative(cw.getDTLogicalAddress(), myAddressPair.getDTLogicalAddress(), ccw.getDTLogicalAddress()) < 0)
                continue;

            if (!(DT_Point.isInOrOnCircle(myAddressPair.getDTLogicalAddress(), cw.getDTLogicalAddress(), ccw.getDTLogicalAddress(), subject.getDTLogicalAddress()))) {
                removeNeighbor(i);
            }
        }

        addNeighborDebug = "exited RNN outer loop";

        // rerun if any were removed
        // (prevents keeping track of all indices after elements are removed.)
        if (neighbors.size() != initialSize)
            removeNonNeighbors();
    }

	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsTable instance is created which contains all statistics
	 * , with each represented by an entry of the table, supported in this object.
	 */
	protected void InitStatisticsStructure()
	{
		/*
		 * For each neighbor in the current neighbor table, add a StatsElement
		 * in the StatsProcessor reference stored in this object. The table may be empty
		 * when a node is started.
		 */		
		for (int i=0; i<neighbors.size(); i++)
		{
			DT_NeighborhoodEntry entry=this.getEntry(i);
			
			this.addNeighborToStatsProcessor(entry.getNeighbor(), entry.getCW(), entry.getCCW());
		}
   }

   /**
	* Create a StatsElement instance for a neighbor and add it to the StatsProcessor object.
	*/
   private void addNeighborToStatsProcessor(I_AddressPair ap, I_AddressPair cw, I_AddressPair ccw)
   {
		statsPro.addStatsElement(statsName, new DT_NeighborhoodStats(ap, cw, ccw), 0/* minoccurs */, -1/* maxoccurs */);
   }
	
	/** Creates a string containing the state of the neighborhood. */
	public String toString()
    {
        StringBuffer sb = new StringBuffer();

        sb.append(neighbors.size());
        sb.append("_");

        for (int i = 0; i < neighbors.size(); i++) {
            DT_AddressPair addrs = getEntry(i).getNeighbor();
            long lct = getEntry(i).getLastContactTime();

            sb.append(addrs.getPhysicalAddress().toString());    
            sb.append("_");
            sb.append(addrs.getDTLogicalAddress().getX());
            sb.append(",");
            sb.append(addrs.getDTLogicalAddress().getY());
            sb.append("_");
            sb.append(Long.toString(lct - adapter.getCurrentTime()));	
            sb.append("_");
        }

        return sb.toString();
    }

	/** Change the logical address of this node to new address. */
    void updateNodeAddress( DT_LogicalAddress addr ){
		myAddressPair = new DT_AddressPair( getMyAddressPair().getPhysicalAddress(), addr );

		//Report event LACHANGED to notification handler.
		if (n_handler != null) {
			NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}
    }

}//End of class DT_Neighborhood






