/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

	/**
	* This class defines the function of OL_Message extension header. It 
	* contains extension type information and provides method to create
	* an extension. Each concrete extension extends this class and 
	* implements the abstract methods defined in this class.
	*
	* @author HyperCast Team
	* @version 2005 (version 3.0)
	*/
	public abstract class Extension {
		
		/**
		 * Extension type values. These are essentially types, but the
		 * extension already has a type field (that should probably be
		 * called "sub-type")
		 */
             public static final byte HEADER_NONE  = 0;
             public static final byte FSM          = 1;     // FSM_Extension.java
             public static final byte RAW_DATA     = 2;     // PayloadExtension.java, EncryptedPayloadExtension.java
             public static final byte ROUTE_RECORD = 3;     // RouteRecord.java
             public static final byte SECURITY     = 21;    // SecurityExtension.java

            public static String typeToString (final byte type)
            {
                switch (type)
                {
                    case HEADER_NONE:  return "HEADER_NONE";
                    case FSM:          return "FSM";
                    case RAW_DATA:     return "RAW_DATA";
                    case ROUTE_RECORD: return "ROUTE_RECORD";
                    case SECURITY:     return "SECURITY";
                    default:
                        throw new HyperCastWarningRuntimeException ("Unknown extension type: " + type);
                }
            }


		/**
		 * Constructs an Extension header object.
		 *
		 *@param extensiontype	type of the extension to be created.
		 *@param contents		byte array used to creat the extension.
		 */
		public static Extension createExtension(byte extensiontype, byte[] contents) {
			return createExtension(extensiontype, contents, null, null);
		}

		/**
		 * Constructs an Extension header object.
		 *
		 *@param extensiontype	type of the extension to be created.
		 *@param contents		byte array used to creat the extension.
		 *@param node			I_Node object needed to construct logical address when
		 *						an extension is created.
		 */
		public static Extension createExtension(byte extensiontype, byte[] contents, I_Node node) {
			return createExtension(extensiontype, contents, node, null);
		}

		/**
		 * Constructs an Extension header object.
		 *
		 *@param extensiontype	type of the extension to be created.
		 *@param contents		byte array used to creat the extension.
		 *@param kv				key vault for security computation.
		 */
		public static Extension createExtension(byte extensiontype, byte[] contents, KeyVault kv) {
			return createExtension(extensiontype, contents, null, kv);
		}

		/**
		 * Constructs an Extension header object.
		 *
		 *@param extensiontype	type of the extension to be created.
		 *@param contents		byte array used to creat the extension.
		 *@param node			I_Node object needed to construct logical address when
		 *						an extension is created.
		 *@param kv				key vault for security computation.
		 */
		public static Extension createExtension(byte extensiontype, byte[] contents, I_Node node, KeyVault kv)
		{
			Extension extension = null;

			if (extensiontype == FSM)
			{
				extension = new FSM_Extension(contents);
			}
			else if (extensiontype == ROUTE_RECORD)
			{
				extension = new RouteRecord(contents, node);
			}
			else if (extensiontype == RAW_DATA)
			{
				extension = new PayloadExtension(contents);
			}
			else if (extensiontype == SECURITY)
			{
				if (kv == null)
				{
					throw new HyperCastWarningRuntimeException ("Try to create a security extension with empty key vault!");
				}
				
				extension = new SecurityExtension(contents, kv);
			}
			else
			{
				throw new HyperCastWarningRuntimeException ("Entension type " + extensiontype + " is not supported!");
			}
			
			return extension;
		}

		/**
		 * Constructs an Extension header object.
		 *
		 *@param extensiontype	type of the extension to be created.
		 *@param contents		byte array used to creat the extension.
		 *@param node			I_Node object needed to construct logical address when
		 *						an extension is created.
		 *@param kv				key vault for security computation.
		 *@param isEncryptedFlag	flag showing the given byte array is encrypted or plaintext.
		 *	True means encrypted, false means plaintext.
		 */
		public static Extension createExtension (byte extensiontype, byte[] contents, I_Node node, KeyVault kv, boolean isEncryptedFlag)
		{
                    if (isEncryptedFlag
                        &&
                        kv == null)
                    {
                        throw new HyperCastFatalRuntimeException ("isEncrypted flag set, but keyvault is null!");
                    }

			if ((extensiontype == RAW_DATA) && (kv != null && kv.getSecPolicy().equals("privacy")))
			{
				return new EncryptedPayloadExtension(contents, isEncryptedFlag, kv);
			}
			else
			{
				return createExtension(extensiontype, contents, node, kv);
			}
		}


		/**
		 * Check a specified extension type is a valid one.
		 * @param extensiontype		extension type to be checked.
		 */
		public static boolean isValidHeader (byte extensiontype)
		{
			return
				extensiontype == RAW_DATA
				||
				extensiontype == FSM
				||
				extensiontype == ROUTE_RECORD
				||
				extensiontype == SECURITY;
		}

            /**
             * Return the value of the extension type field.
             * TODO: this should be "getType" saying "getExtensionType" is redundant;
             * obviously this is an extension object...
             */
            public abstract byte getExtensionType();
	
		/**
		 * Convert this extension to a byte array.
		 */
		public abstract byte[] toByteArray();
		
		/*
		 * This method provides a way to get the length of the byte 
		 * array of an extension without actually executing toByteArray().
		 */
		/**
		 * Return the length of the byte array of this extension.
		 */
		public abstract int getSize();

		/**
		 * Return a new extension instance by cloning this extension.
		 */
		public abstract Extension Clone();


            protected String payloadToString()
            {
                return "";
            }

            public String toString()
            {
                final StringBuffer result = new StringBuffer();
                final byte type = getExtensionType();
                final String typeString = typeToString (type);
                result.append ("\n\tExtension Type: " + typeString + "\n");
                result.append ("\tPayload Length:\t" + getSize() + "\n");
                result.append ("\tPayload:\t" + payloadToString() + "\n");
                return result.toString();
            }
	}
