/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.HC;

import hypercast.*;

/** 
 * This class encapsulates a HC protocol message.
 *
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Dongwen Wang
 * @version	2.0, Feb. 27, 2001
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */

public final class HC_Message implements I_Message {
	
		/** Hypercast protocol message type (used to ping neighbor periodically) */
	public static final byte Ping   = 0;
		/** Hypercast protocol message type (used to kill a node) */
	public static final byte Kill   = 1;
		/** Hypercast protocol message type (used to beacon all) */
	public static final byte Beacon = 2;
		/** Hypercast protocol message type (used to leave the hypercube) */
	public static final byte Leave  = 3;
	
    	/** Protocol Number value in Protocol field for HC3.0 protocol message */
    	public static final byte PROTONUM_HC30 = 1;

    	public byte Proto_num;

    	public byte Proto_sub;

		/** Hypercast protocol message type */
	private int Type;
		/** A hash of overlay properties.  It is used to detect either
          conflicting overlays or detect bad packets. */
	private int overlayHash;
		/** Source logical address */
	private HC_LogicalAddress  SrcLA;
		/** Destination logical address */
	private HC_LogicalAddress  DestLA;
		/** Source physical address */
	private I_PhysicalAddress SrcPA;
		/** Destination physical address */
	private I_PhysicalAddress DestPA;
		/** HRoot logical address */
	private HC_LogicalAddress  HRootLA;
		/** HRoot sequence number */
	private int HRootSeqNum;
	
		/** Length of I_PhysicalAddress in bytes */
	private int PASize;
		/** Length of I_LogicalAddress in bytes */
	private int LASize = 4;	
	
	/** 
	 * Constructs a HC_Message with a byte array and an I_UnicastAdapter <p>
	 * 
	 *   The following is the byte format followed <br>
	 *   Byte 0     : Type <br>          
	 *   Bytes 1-4  : overlayHash <br>
	 *   Bytes 5-10  : SrcPA <br>
	 *   Bytes 11-14:   SrcLA <br>
	 *   BYtes 15-20: 	DestPA <br>
	 *	 Bytes 21-24: DestLA <br>
	 *	 Bytes 25-28: HRootLA <br>
	 *	 Bytes 29-32: HRootSeqNum <br>
	 * 
	 * @param	data		an input byte array
	 * @param	adapter		an I_UnicastAdapter object
	 * @exception	if the packet has an invalid type
	 */  

	public HC_Message(byte[] data, int offset, I_UnicastAdapter adapter) {

		PASize = adapter.getPhysicalAddressSize();

        byte[] tempaddr = new byte[PASize];

		//The first byte is Protocol field
        Proto_num = (byte) ((data[offset] >> 4) & 0xf);
        Proto_sub = (byte) (data[offset] & 0xf);
        offset += 1; 
        if (Proto_num != PROTONUM_HC30)
            throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_HC30.");
            
        //Skip 2 bytes which stores the length of the message
        offset += 2;

		Type = data[offset];
		offset += 1;
		if ((Type!=Ping) && (Type!=Kill) && (Type!=Beacon) && (Type!=Leave)) 
			throw new IllegalArgumentException("HC_Message: Invalid message type!");

        overlayHash = ByteArrayUtility.toInteger(data, offset);
        offset += 4;

        System.arraycopy(data, offset, tempaddr, 0, PASize);
        offset += PASize;
        SrcPA = adapter.createPhysicalAddress(tempaddr, 0);

		

        SrcLA = new HC_LogicalAddress(ByteArrayUtility.toInteger(data, offset), true);
        offset += LASize;
        

        System.arraycopy(data, offset, tempaddr, 0, PASize);
        offset += PASize;
        DestPA = adapter.createPhysicalAddress(tempaddr, 0);
		
        DestLA = new HC_LogicalAddress(ByteArrayUtility.toInteger(data, offset), true);
        offset += LASize;
        
        HRootLA = new HC_LogicalAddress(ByteArrayUtility.toInteger(data, offset), true);
        offset += LASize;
        
        HRootSeqNum = ByteArrayUtility.toInteger(data, offset);
        offset += 4;
	}

    
	
  
	/** 
	 * Constructs a HC_Message with all the components <p>
	 * @param	type	the type
	 * @param	srcp	source PhysicalAddress object
	 * @param	srcl	source LogicalAddress object
	 * @param	desp	destination PhysicalAddress object
	 * @param	desl	destination LogicalAddress object
	 * @param	hrootl	HRoot LogicalAddress object
	 * @param	hrootsn	HRoot sequence number
	 * @param	data	payload data array
	 * 
	 * @exception	if the packet has an invalid type
	 */  
	public HC_Message(byte type, 
            int gh,
			I_PhysicalAddress srcp, HC_LogicalAddress srcl,
			I_PhysicalAddress desp, HC_LogicalAddress desl,
			HC_LogicalAddress hrootl, int hrootsn,
			byte[] data, 
			I_UnicastAdapter adapter) {
		
		if ((type!=Ping) && (type!=Kill) && (type!=Beacon) && (type!=Leave)) 
			throw new IllegalArgumentException("HC_Message: Invalid message type!");



        	//Set values to protocol number and protocol subfield.
        	Proto_num = PROTONUM_HC30;
        	Proto_sub = I_Node.PROTOSUB_NOSPEC; //set to default value

        	Type = type;

        	overlayHash = gh;
        
        	SrcPA  = srcp;
        	SrcLA  = srcl;    
        	DestPA  = desp;
        	DestLA = desl;
        
        	HRootLA = hrootl;
        	HRootSeqNum = hrootsn;
        
			PASize = adapter.getPhysicalAddressSize();
	}
	
	/** New Constructor of HC_Message. */
	public HC_Message(byte proto_sub,
			byte type, 
            int gh,
			I_PhysicalAddress srcp, HC_LogicalAddress srcl,
			I_PhysicalAddress desp, HC_LogicalAddress desl,
			HC_LogicalAddress hrootl, int hrootsn,
			byte[] data, 
			I_UnicastAdapter adapter) {
		
		if ((type!=Ping) && (type!=Kill) && (type!=Beacon) && (type!=Leave)) 
			throw new IllegalArgumentException("HC_Message: Invalid message type!");



        	Proto_num = PROTONUM_HC30;
        	Proto_sub = proto_sub;

        	Type = type;

        	overlayHash = gh;
        
        	SrcPA  = srcp;
        	SrcLA  = srcl;    
        	DestPA  = desp;
        	DestLA = desl;
        
        	HRootLA = hrootl;
        	HRootSeqNum = hrootsn;
        
			PASize = adapter.getPhysicalAddressSize();
	}
	

	/** Reconstruct HC message from byte array. */	
	public static HC_Message restoreMessage(byte[] recvbuff, int[] validBytesStart, int validBytesEnd, I_UnicastAdapter adapter, int thisOverlaysHash) {
		int pasize = adapter.getPhysicalAddressSize();

		final int lasize = 4;

		//Protocol, length, Type, hash, PA, LA, PA, LA, LA, Seq, 
		int msgLength = 3 + 1 + 4 + pasize*2 + lasize*3 + 4;
				
        if (validBytesEnd - validBytesStart[0] < msgLength)
            return null;

        byte proto_num = (byte) ((recvbuff[validBytesStart[0]] >> 4) & 0xf);
        byte proto_sub = (byte) (recvbuff[validBytesStart[0]] & 0xf);
        if (proto_num != PROTONUM_HC30)
            throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_HC30.");
 
        short length = ByteArrayUtility.toShort(recvbuff, validBytesStart[0]+1);
        if (length != (short) (1 + 4 + pasize*2 + lasize*3 + 4))
            throw new IllegalArgumentException("The length of the message is incorrect.");


        int incomingOverlayHash = ByteArrayUtility.toInteger(recvbuff, validBytesStart[0]+4);
        if (incomingOverlayHash != thisOverlaysHash)
            throw new IllegalArgumentException("Message is not of this Overlay.");
        
		HC_Message msg = new HC_Message(recvbuff, validBytesStart[0], adapter);
        validBytesStart[0] += msgLength;

        return msg;
    }
    
	
	

	/** 
	 * Converts message to byte array
	 * @return	a byte array which represents a packet which can be sent
	 *          out to the network
	 */		
	public byte[] toByteArray() {
		int msgLength = 3 + 1 + 4 + 3*LASize + 2*PASize + 4;  //37 bytes

		//System.out.println("HC_Message: PAsize="+PASize);

		byte[] Result = new byte[msgLength];
		int index = 0;
					   
		//This is protocol field
        Result[index] = (byte) ((Proto_num << 4) | (Proto_sub & 0xf));
        index += 1;
        
		//This is length field.
        short length = (short) (1 + 4 + 3*LASize + 2*PASize + 4);
        System.arraycopy(ByteArrayUtility.toByteArray(length), 0, Result, index, 2);
        index += 2;


		// Copy the type
		Result[index] = (byte)Type;
        index++;
        
        // Copy the OverlayHash
		System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
        index += 4;
        
		//System.out.println("HC_Message: source length="+(SrcPA.toByteArray()).length+",length of Result="+Result.length+",index="+index+",PASize="+PASize);
		// Copy the Source Information
		System.arraycopy(SrcPA.toByteArray(), 0, Result, index, PASize);		
        index += PASize;
		System.arraycopy(ByteArrayUtility.toByteArray(SrcLA.getColexIndex()), 0, Result, index, LASize);
        index += LASize;

		//System.out.println("HC_Message: source length="+(DestPA.toByteArray()).length+",length of Result="+Result.length+",index="+index+",PASize="+PASize);
		// Copy the Destination Information
		System.arraycopy(DestPA.toByteArray(), 0, Result, index, PASize);
        index += PASize;		
		System.arraycopy(ByteArrayUtility.toByteArray(DestLA.getColexIndex()), 0, Result, index, LASize);
        index += LASize;

		// Copy HRoot information
		System.arraycopy(ByteArrayUtility.toByteArray(HRootLA.getColexIndex()), 0, Result, index, LASize);
        index += LASize;
		System.arraycopy(ByteArrayUtility.toByteArray(HRootSeqNum), 0, Result, index, 4);
        index += 4;

		return Result;
	}
   

    	/** Returns the Proto_num field of the message. */
    	byte getProtoNum() {
           return Proto_num;
    	}

    	/** Returns the Proto_sub field of the message. */
    	byte getProtoSub() {
           return Proto_sub;
    	}


	/** 
	 * Returns the type of the message
	 * @return	Type of this message
	 */
	public int getType() {
		return Type;
	}


    /** Returns the overlayHash of this message. */
public int getOverlayHash() 
    {
        return overlayHash;
    }
    

	/** 
	 * Returns the destination physical address of the message
	 * @return	the destication I_PhysicalAddress object
	 */
	public I_PhysicalAddress getDestPA(){
		return DestPA;
	}


	/** 
	 * Returns the source logical address of the message
	 * @return	the source HC_LogicalAddress object
	 */
	public HC_LogicalAddress getSrcLA(){
		return SrcLA;
	}

	
	/** 
	 * Returns the destination logical address of the message
	 * @return	the destination HC_LogicalAddress object
	 */
	public HC_LogicalAddress getDestLA(){
		return DestLA;
	}
	
	
	/** 
	 * Returns the HRoot logical address of the message
	 * @return	the HRoot HC_LogicalAddress object
	 */
	public HC_LogicalAddress getHRootLA() {
		return HRootLA;
	}

	
	/** 
	 * Returns the HRoot sequence number of the message
	 * @return	the HRoot sequence number
	 */	
	public int getHRootSeqNum() {
		return HRootSeqNum;
	}


	
	/** 
	 * Returns the source physical address of the message
	 * @return	the source I_PhysicalAddress object
	 */
	public I_PhysicalAddress getSrcPA(){
		return SrcPA;
	}
		
	/** 
	 * Converts the message object to String 
	 * @return	a string representation of the HC_Message object
	 */
	public String toString() {
		String X = null;

		// get the type
		//X = "\tType:\t";
		X = "";

		switch (Type) {
		case HC_Message.Beacon:  X = X + "Beacon"; break;
		case HC_Message.Kill  :  X = X + "Kill"; break;
		case HC_Message.Ping  :  X = X + "Ping"; break;
		case HC_Message.Leave :  X = X + "Leave"; break;
		default : X= "Unknown";
		}
		return X;
	}
}



