/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.util.XmlUtil;

import java.io.IOException;
import java.io.OutputStream;


/** This is the class used to send the byte array through stream manager 
 *
 * @author HyperCast Team
 */
public class HCastOutputStream extends OutputStream
{
	/**
	 * Debug flag.
	 */
	protected boolean debug = false;

    /** the buffered byte array */
	protected byte[] buffer;
    /** the buffer size */
	protected static int buffersize;
    /** the sending threthold size */
	protected static int threthold;
    /**the current byte number in the buffer*/
	protected int availableBytes = 0;
    /**the stream manager */
	protected StreamManager parent;
     /**flag indicating the sending of data*/
	protected boolean sent = false;
    /**the current delivery mode */
	protected byte currentDeliveryMode = OL_Message.DELIVERY_MODE_MULTICAST;
    /**the destination logical address of the unicast message*/
	protected I_LogicalAddress destinationLA = null; 
    /**the hop limit */
	protected short hopLimit;
    /** stream id */
	protected byte [] streamID;
	
    /**
     * Constructs an <code>HCastOutputStream</code> with a
     * <code>StreamManager</code> and a stream id.
     *
     * @param sm  StreamManager
     * @param streamID Stream id
     */
    public HCastOutputStream (StreamManager sm, byte[] streamID)
    {
        this.parent = sm;
        this.streamID = streamID;
        this.hopLimit = parent.socket.hopLimit;
        
		if (debug)
			System.out.println("HCastOutputStream:hopLimit="+this.hopLimit);        
        
		this.buffersize = parent.config.getPositiveIntAttribute (XmlUtil.createXPath("/Public/OutputStreamBuffersize"));
        this.buffer = new byte[buffersize];
    }

    /** 
     * Flushes this output stream and forces any buffered output bytes to be written out.
     * */  
 	public void flush()throws IOException
 	{
 		if(availableBytes>0)
 		{
 			byte[] payload = new byte[availableBytes];
 			System.arraycopy(buffer,0,payload,0,availableBytes);

		if (debug)
			//System.out.println("HCastOutputStream:flush: send data out:");
			System.out.println(new String(payload, 0, availableBytes));
		
		
 			send(payload,availableBytes);
			availableBytes = 0;
 		}
 	}
    
	/** 
	  * remove the stream id from the hashtable.
	  * */  
	 public void close()throws IOException
	 {
		parent.outputtable.remove(new String(streamID));
	 }
	 
	 
	 
        /**Writes b.length bytes from the specified byte array to this output stream.*/
 	public void write(byte[] b) throws IOException 
 	{
 		//buffer the data if not reaching the threshold
 		if((b.length+availableBytes)<buffersize)
 		{
 			System.arraycopy(b,0,buffer,availableBytes,b.length);
 			availableBytes += b.length;
 		}
 		else
 		{
 		//send the data
 		// the first message contains the data in the buffer and the newly coming data
 			byte[] first = new byte[buffersize];
 			System.arraycopy(buffer,0,first,0,availableBytes);
 			System.arraycopy(b,0,first,availableBytes,buffersize-availableBytes);	
 			send(first,buffersize);
 			
 		//send the following data	
 			int len = b.length-buffersize+availableBytes;
 			int index = buffersize-availableBytes;
 			availableBytes = 0;
 			
 			while(!(len<buffersize))
 			{
 				byte[] p = new byte[buffersize];
 				System.arraycopy(b,index,p,0,buffersize);
 				send(p,buffersize);
 				index += buffersize;
 				len = len - buffersize;
 			}
 		//buffer the left data
 			if(len>0)
 			{
				System.arraycopy(b,b.length-len,buffer,0,len);
 				availableBytes = len;
 			}	
 		}
 	}

        /**
         * Writes one byte (passed as a integer) to this output stream. This 
         * method is defined in java.io.OutputStream and needs to be overrided
         * in this class.
         */   
	public void write(int b)throws IOException
	{
		buffer[availableBytes] = (byte)((b >>>  0) & 0xFF);
 		availableBytes += 1;
		//buffer the data if not reaching the threshold
 		if(availableBytes==buffersize)
 		{
 		//send the data
  			send(buffer,buffersize);
  			availableBytes = 0;
  		}
	} 

	/**Writes one byte this output stream.*/   
	public void write(byte b)throws IOException
	{
		buffer[availableBytes] = b;
		availableBytes += 1;
		//send the data if the buffer is reaching the threshold
		if(availableBytes==buffersize)
		{
		//send the data
			send(buffer,buffersize);
			availableBytes = 0;
		}
	} 

	 
	 /**Writes len bytes from the specified byte array starting at offset off to this output stream. */
	public void write(byte[] b, int off, int len)throws IOException 
	{
		if (len > (b.length - off))
		{
			throw new HyperCastWarningRuntimeException("The stream packet doesn't have specified bytes data!");	
		}

		byte[] bb = new byte[len];
		System.arraycopy(b,off,bb,0,len);
		write(bb);
		
	}
    
	/**
	 * @return stream id
	 * @deprecated use getStreamIdentifier
	 */
	public int getStreamID()
	{
		return ByteArrayUtility.toInteger(streamID,0);
	}

	 /**
	  * @return stream id*/
	public int getStreamIdentifier()
	{
		return ByteArrayUtility.toInteger(streamID,0);
	}


	/**
	 * 	Sets the delivery mode for all outgoing messages to unicast with the specified logical address as destination.
	 * @deprecated use setUnicast
	 */
	public void setSendToUnicast (I_LogicalAddress Laddr)
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_UNICAST;
		destinationLA = Laddr;
	}

	/**
	 * 	Sets the delivery mode for all outgoing messages to unicast with the specified logical address as destination.
	 */
	public void setUnicast (I_LogicalAddress Laddr)
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_UNICAST;
		destinationLA = Laddr;
	}

	/**
	 * 	Sets the delivery mode for all outgoing messages to multicast. 
	 * @deprecated use setBroadcast
	 */
	public void setSendToAll()
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_MULTICAST;
	}

	/**
	 * 	Sets the delivery mode for all outgoing messages to multicast. 
	 */
	public void setBroadcast()
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_MULTICAST;
	}
	
	/**
	 * 	Sets the delivery mode for all outgoing messages to flooding. 
	 * @deprecated use setFlood
	 */
	public void setSendFlood()
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_FLOOD;
	}

	/**
	 * 	Sets the delivery mode for all outgoing messages to flooding. 
	 *
	 */
	public void setFlood()
	{
		currentDeliveryMode = OL_Message.DELIVERY_MODE_FLOOD;
	}

	/**
	 * 	Returns the current delivery mode for outgoing messages (DELIVERY_MODE_FLOOD, DELIVERY_MODE_MULTICAST, DELIVERY_MODE_UNICAST, DELIVERY_MODE_INVALID). 
	 */
	public byte getDeliveryMode()
	{
		return currentDeliveryMode;
	}

	/**
	 * Sets the hop-length field of outgoing messages to the specified value.
	 */
	public void setHopLimit (short hopLimit)
	{
            this.hopLimit = hopLimit;
	}
   
    /**
     * create a message and send out
     * */
    synchronized private void send (byte[] payload, int length)
    {
        I_OverlayMessage msg = parent.socket.createMessage (payload, streamID);
        
        msg.setHopLimit (hopLimit);

		if (debug)
			System.out.println("HCastOutputStream: send: currentDeliveryMode is "+currentDeliveryMode);

        /*
         * messages are sent out based on the current specified delivery mode
         */
        switch(currentDeliveryMode)
        {
        	case OL_Message.DELIVERY_MODE_MULTICAST:
        	{
				if (debug)
					System.out.println("call socket.sendToAll()");

				parent.socket.sendToAll(msg);
				sent = true;
        	}
        		break;
        		
        	case OL_Message.DELIVERY_MODE_UNICAST:
        	{
				if (debug)
					System.out.println("call socket.sendToNode()");

				parent.socket.sendToNode(msg,destinationLA);
				sent = true;	
        	}
        		break;
        		
        	case OL_Message.DELIVERY_MODE_FLOOD:
        	{
				if (debug)
					System.out.println("call socket.sendToFlood()");

        		parent.socket.sendFlood(msg);
				sent = true;	
        	}
        		break;
        	default:
        	{
        		throw new HyperCastWarningRuntimeException("Invalid delivery mode");	
        	}
        }

        
    }
	/**
	 * @return the message sending flag 
	 */
    public boolean hasSent()
    {
        return sent;
    }
	
	/**
	 * reset the message sending flag
	 *
	 */
    public void reset()
    {
        sent = false;
    }
 }
 
