/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast;

import hypercast.util.XmlUtil;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;

import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.KeyGenerator;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

/**
 * This class manages the personal key of local node and all of the 
 * neighborhood keys and certificates.
 *
 * @author HyperCast Team
 * @author Josh Zaritsky
 * @version 3.0, Sep 23, 2003
 */
public class KeyVault 
{

	/** the wrapped message key length */
	private int wrappedKeySize;

	/** Overlay socket configuration object */
	HyperCastConfig config;

	/** This node's personal certificate */
	X509Certificate myCert;

	/** CA certificate */
	X509Certificate CACert;

	/** key size in bytes */
	int keySize;

	/** key distribution mode */
	String keyMode;
	
	/** neighbor key mode, if keyMode == "NeighborKey" */
	/*
	 * Three possible values:
	 *      "UniformDynamicKey"   		use dynamic neighbor key for all messsage
	 * 		"StaticNonNeighborKey"		use unique static key for each non-neighbor node, dynamic key for neighbor nodes
	 *      "SharedProtocolKey"			use shared key for protocol message, dynamic key for data message
	 */
	String neighborKeyMode;
	
	final static String NKM_UNIFORM_DYNAMIC_KEY = "UniformDynamicKey";
	final static String NKM_STATIC_NONNEIGHBOR_KEY = "StaticNonNeighborKey";
	final static String NKM_SHARED_PROTOCOL_KEY = "SharedProtocolKey";

	/** security policy applied to the overlay messages */
	String secPolicy;

	/** security policy applied to the protocol messages */
	String protoSecPolicy;

	/** algorithm for encryption/decryption */
	String confidentialityAlgorithm;

	/** algorithm for MAC calculation */
	String integrityAlgorithm;

	/** private key */
	PrivateKey myPrivateKey;

	/** This node's personal key */
	SecretKeySpec myPersonalKey;

	/** group key shared by all nodes in the same overlay. */
	SecretKeySpec groupKey;

	/** group key in string format which is read from the private attribute DOM document. */
	String groupKeyString;

	/** The creation time of the personal key */
	long myKeyTimestamp;

	/*
	 * KeyVault needs to call the (callback) functions in the SecurityProcessor,
	 * for example, start or reset WatchDogTimer. So it keeps a reference to 
	 * SecurityProcessor.
	 */
	/**
	 * Security processor which handles the sending/receiving of secure protocol
	 * and overlay messages.
	 */
	SecurityProcessor sp = null;

	/**
	 * Next valid sequence number for the message sent from local node.
	 */
	int mySequenceNum;

	/** Prefix for all public security property names read by the key vault. */
	protected static final String PUBLIC_SECURITY_PROPERTY_NAME_PREFIX = "/Public/Security/";

	/** Prefix for all private property names read by the key vault. */
	protected static final String PRIVATE_PROPERTY_NAME_PREFIX = "/Private/";

	/**
	 * Constructor. Set up the appropriate settings for the 
	 * algorithm and related parameters.
	 *
	 * @param _config 	Overlay socket configuration object.
	 *
	 * @throws IllegalArgumentException when the seetings for security implementation
	 * contains errors.
	 */
	public KeyVault(HyperCastConfig _config) 
	{
		config = _config;
		
		/**
		 * create new keyvault table
		 */
		table = new Hashtable();

		/*
		 * Add the "BC" provider (required for RSA encryption/decryption).
		 */ 
		if (Security.getProvider("BC") == null) {
			config.log.println("Adding BC provider");
			Security.addProvider(new BouncyCastleProvider());
		}

		/*
		 * Get the security level.
		 */
		String secLevel = config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX+"SecurityLevel"));
		
		/*
		 * Get the security 
		 */
		if(secLevel.equals("plaintext")){
			throw new HyperCastWarningRuntimeException ("Try to create key vault when security level is plaintext.");
		}else if(secLevel.equals("protocolIntegrity")){
			protoSecPolicy = "integrity";
			secPolicy = "plaintext";
		}else if(secLevel.equals("integrity")){
			protoSecPolicy = "integrity";
			secPolicy = "integrity";
		}else if(secLevel.equals("privacy")){
			protoSecPolicy = "integrity";
			secPolicy = "privacy";
		}else {
			throw new HyperCastWarningRuntimeException ("Invalid security level: "+secLevel);
		}
		
		/*
		 * In the following try statement, the user certificate and CA certificate
		 * are created.
		 */
		CertificateFactory certFactory;
		FileInputStream fis;
		String certFileName = "";

		try {
			/*
			 * Create user certificate
			 */
			certFileName =
				config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "CertificateLocation"));
			
			fis = new FileInputStream(certFileName);

			DataInputStream dis = new DataInputStream(fis);

			certFactory = CertificateFactory.getInstance("X.509");

			byte[] certFile = new byte[dis.available()];
			dis.readFully(certFile);

			ByteArrayInputStream bais = new ByteArrayInputStream(certFile);
			myCert = (X509Certificate) certFactory.generateCertificate(bais);

			/*
			 * Create CA certificate
			 */
			certFileName =
				config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "CACertificateLocation"));

			fis = new FileInputStream(certFileName);

			dis = new DataInputStream(fis);
			certFactory = CertificateFactory.getInstance("X.509");

			certFile = new byte[dis.available()];
			dis.readFully(certFile);

			bais = new ByteArrayInputStream(certFile);
			CACert = (X509Certificate) certFactory.generateCertificate(bais);
		} catch (FileNotFoundException fnfe) {
			config.log.println(certFileName);
			throw new IllegalArgumentException("The specified certificate file does not exist!");
		} catch (IOException ioe) {
			throw new IllegalArgumentException("Unable to read the certificate file!");
		} catch (CertificateException ce) {
			throw new IllegalArgumentException("Certificate file does not contain an X.509 Certificate!");
		}

		/*
		 *  Read in the private key associated with the certificate.
		 */
		KeyStore privateKeyStore;

		try {
			fis = new FileInputStream(config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX
															   + "KeyStoreLocation")));

			DataInputStream dis = new DataInputStream(fis);

			String ksPassword =
				config.getPrivateTextAttribute (XmlUtil.createXPath(PRIVATE_PROPERTY_NAME_PREFIX + "KeyStorePassword"));
							
			String pkAlias =
				config.getPrivateTextAttribute (XmlUtil.createXPath(PRIVATE_PROPERTY_NAME_PREFIX + "PrivateKeyAlias"));

			String pkPassword =
				config.getPrivateTextAttribute (XmlUtil.createXPath(PRIVATE_PROPERTY_NAME_PREFIX + "PrivateKeyPassword"));

			privateKeyStore = KeyStore.getInstance(KeyStore.getDefaultType());

			privateKeyStore.load(dis, ksPassword.toCharArray());

			/*
			 * Obtains the private key of local node.
			 */
			myPrivateKey =
				(PrivateKey) privateKeyStore.getKey(pkAlias,
													pkPassword.toCharArray());
		} catch (FileNotFoundException fnfe) {
			throw new IllegalArgumentException("The specified key store file does not exist!");
		} catch (IOException ioe) {
			throw new IllegalArgumentException("Unable to read the keystore file!");
		} catch (NoSuchAlgorithmException e) {
			throw new IllegalArgumentException("The required public key algorithm is unavailable!");
		} catch (KeyStoreException e) {
			e.printStackTrace();
			throw new IllegalArgumentException();
		} catch (CertificateException e) {
			e.printStackTrace();
			throw new IllegalArgumentException();
		} catch (UnrecoverableKeyException e) {
			e.printStackTrace();
			throw new IllegalArgumentException();
		}

		/*
		 * Get the algorithm for message encryption/decryption operation.
		 */
		confidentialityAlgorithm =
			config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "CryptAlgorithm"));

		/*
		 * Get the key mode value from the overlay socket properties.
		 */
		keyMode =
			config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "KeyModeMethod"));
		
		if(!keyMode.equals("GroupKeys")){
			/* three different neighbor key modes */
			if(keyMode.equals("NeighborhoodKey1")){
				neighborKeyMode = KeyVault.NKM_UNIFORM_DYNAMIC_KEY;
			}else if(keyMode.equals("NeighborhoodKey2")){
				neighborKeyMode = KeyVault.NKM_STATIC_NONNEIGHBOR_KEY;
			}else if(keyMode.equals("NeighborhoodKey3")){
				neighborKeyMode = KeyVault.NKM_SHARED_PROTOCOL_KEY;
			}else{
				throw new IllegalArgumentException("Invalid KeyMode: "+keyMode);
			}
			
			keyMode = "NeighborhoodKeys";
		}
		/*
		 * If the key mode is end-to-end, create the group key from group key string specified
		 * in the private attribute document.
		 * TODO: use different property name for shared key under neighbor key mode
		 */
		if (keyMode.equals("GroupKeys") || neighborKeyMode.equals(NKM_SHARED_PROTOCOL_KEY))
		{
			groupKeyString =
				config.getPrivateTextAttribute (XmlUtil.createXPath(PRIVATE_PROPERTY_NAME_PREFIX + "GroupKey"));

			try
			{
				groupKey = new SecretKeySpec(groupKeyString.getBytes(), confidentialityAlgorithm);
			} catch (Exception e) {
				e.printStackTrace();
				throw new HyperCastFatalRuntimeException ("Error reading group key.");
			}

		}


		/*
		 * Get the algorithm for message integrity (MAC) operation.
		 */
		integrityAlgorithm =
			config.getTextAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "MacAlgorithm"));

		/*
		 * Get the value of key size.
		 */
		keySize = config.getIntAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "SymmetricKeyLength"));

		/*
		 * convert key size from bits to bytes.
		 */
		keySize /= 8;

		/*
		 * Create a new personal key since we don't have one before.
		 */
		updateMyPersonalKey();
		
		/* 
		 * determine the wrapped message key length.
		 */
		Cipher wrapper;
		
		if(this.keyMode.equals("GroupKeys")){
			wrappedKeySize = 0;
		}else{
			try {
				wrapper = Cipher.getInstance(confidentialityAlgorithm);
				wrapper.init(Cipher.WRAP_MODE, getMyPersonalKey());
	
				wrappedKeySize = wrapper.wrap(getNewMsgKey()).length;
			} catch (InvalidKeyException e) {
				throw new IllegalArgumentException("Can't wrap key with algorithm: "
												   + confidentialityAlgorithm);
			} catch (NoSuchAlgorithmException e) {
				throw new IllegalArgumentException("Can't wrap key with algorithm: "
												   + confidentialityAlgorithm);
			} catch (NoSuchPaddingException e) {
				throw new IllegalArgumentException("Can't wrap key with algorithm: "
												   + confidentialityAlgorithm);
			} catch (IllegalStateException e) {
				throw new IllegalArgumentException("Can't wrap key with algorithm: "
												   + confidentialityAlgorithm);
			} catch (IllegalBlockSizeException e) {
				throw new IllegalArgumentException("Can't wrap key with algorithm: "
												   + confidentialityAlgorithm);
			}
		}

		/*
		 * Initializes the next valid sequence number to one. When a neighbor receives 
		 * key update message from this node, the minial allowed sequence number for
		 * this node is set to 1.
		 */
		mySequenceNum = 1;
		
	}

	/**
	 * Get the key size.
	 * 
	 * @return the key size.
	 */
	public int getKeySize()
	{
		return keySize;
	}

	/**
	 * Determines the key distribution model to be used for the overlay.  This
	 * can be hop-by-hop or end-to-end.
	 *
	 * @return key distribution mode.
	 */
	public String getKeyMode()
	{
		return keyMode;
	}
	
	public String getNeighborKeyMode()
	{
		return neighborKeyMode;
	}

	/**
	 * Gets the security policy applied to overlay messages.
	 *
	 * @return socket security polocy in string.
	 */
	public String getSecPolicy()
	{
		return secPolicy;
	}

	/**
	 * Gets the security policy applied to protocol messages.
	 *
	 * @return node security policy in string.
	 */
	public String getProtoSecPolicy()
	{
		return protoSecPolicy;
	}

	/**
	 * Gets the algorithm for encryption/decryption.
	 *
	 * @return encryption/decryption algorithm in string.
	 */
	public String getConfidentialityAlgorithm()
	{
		return confidentialityAlgorithm;
	}

	/**
	 * Gets the algorithm for MAC calculation.
	 *
	 * @return MAC algorithm in string.
	 */
	public String getIntegrityAlgorithm()
	{
		return integrityAlgorithm;
	}

	/**
	 * Gets the user certificate.
	 *
	 * @return the user certificate.
	 */
	public X509Certificate getMyCertificate()
	{
		return myCert;
	}

	/**
	 * Gets the CA certificate.
	 *
	 * @return the CA certificate.
	 */
	public X509Certificate getCACertificate()
	{
		return CACert;
	}

	/**
	 * Gets the public key of this node
	 *
	 * @return the public key
	 */
	public PublicKey getMyPublicKey()
	{
		return myCert.getPublicKey();
	}

	/**
	 * Gets the user private key.
	 *
	 * @return the private key.
	 */
	public PrivateKey getMyPrivateKey()
	{
		return myPrivateKey;
	}
	/*package*/ void setMyNoNeighborKeyFor(I_LogicalAddress la, SecretKeySpec key, long timeStamp){
		KeyVaultEntry entry = (KeyVaultEntry)this.getEntry(la);
		if(entry!=null){
			entry.setMyNoNeighborKey(key);
			entry.setMyNoNeighborKeyTimestamp(timeStamp);
			
		}
	}
	/*package*/ void generateMyNoNeighborKeyFor(I_LogicalAddress la, long timeStamp){
		KeyGenerator keyGen;
		SecretKeySpec newKey;

		try
		{
			{
				keyGen = KeyGenerator.getInstance(confidentialityAlgorithm);
				keyGen.init(keySize * 8);

				newKey =
					new SecretKeySpec(keyGen.generateKey().getEncoded(),
									  confidentialityAlgorithm);
				
				KeyVaultEntry entry = this.getEntry(la);
				if(entry!=null){
					entry.setMyNoNeighborKey(newKey);
					entry.setMyNoNeighborKeyTimestamp(timeStamp);
					
				}
				
			}
		} catch (NoSuchAlgorithmException nsae) {
			throw new IllegalArgumentException("Key algorithm specified does not exist!");
		}
	}
	/*package*/ SecretKeySpec getMyNoNeighborKeyFor(I_LogicalAddress la){
		
		KeyVaultEntry entry = this.getEntry(la);
		
		if(entry==null)
			return null;
		else
			return entry.getMyNoNeighborKey();
		
		
	}
	/*package*/ void removeMyNoNeighborKeyFor(I_LogicalAddress la){
		KeyVaultEntry entry = this.getEntry(la);
		
		if(entry!=null){
			entry.setMyNoNeighborKey(null);
		}
		
	}
	/*package*/ long getMyNoNeighborKeyTimestamp(I_LogicalAddress la){
		
		KeyVaultEntry entry = this.getEntry(la);
		
		if(entry==null || entry.getMyNoNeighborKey()==null)
			return 0;
		else 
			return entry.getMyNoNeighborKeyTimestamp();
		
	}
	/**
	 * Get the personal key of this node
	 *
	 * @return this node's personal key
	 */
	public SecretKeySpec getMyPersonalKey()
	{
		if (keyMode.equals("NeighborhoodKeys"))
		{
			return myPersonalKey;
		}
		else
		{
			/*
			 * In group key mode, each time the group key is used, check if 
			 * new group key is created by the application and stored in the 
			 * private configuration document, so that the up-to-date group
			 * key is used in the data transmission. The generation and 
			 * distribution of group key is the responsibility of application.
			 */

//TODO: It has impact on the performance when the following operation is applied to 
//all outgoing packets.
			//System.out.println("KeyVault:getMyPersonalKey: read group key string from configuration file.");
/*
			String readGroupKeyStr =
				config.getPrivateTextAttribute (HyperCastConfig.createXPath(PRIVATE_PROPERTY_NAME_PREFIX + "GroupKey"));
						
			if (!readGroupKeyStr.equals(groupKeyString))
			{	
				try
				{
					System.out.println("KeyVault:getMyPersonalKey: group key is changed, reconstruct it.");

                            // To avoid the warning:
                            //new ObjectInputStream (new ByteArrayInputStream (groupKeyString.getBytes())); ???

					ObjectInputStream keyObj =
						//new ObjectInputStream(new StringBufferInputStream(readGroupKeyStr));
						new ObjectInputStream(new ByteArrayInputStream(groupKeyString.getBytes()));
							
					groupKey = (SecretKeySpec) keyObj.readObject();
			
					keyObj.close();
				} catch (Exception e) {
					throw new HyperCastFatalRuntimeException ("Error reading group key.");
				}
				
				groupKeyString = readGroupKeyStr;
			}
*/

			return groupKey;
		}
	}

	/**
	 * Set the creation time to be associated with my personal key.
	 *
	 * @param timestamp the creation time of the key
	 */
	public void setMyKeyTimestamp(long timestamp)
	{
		myKeyTimestamp = timestamp;
	}

	/**
	 * Get the timestamp of when the personal key was created.
	 *
	 * @return the key timestamp
	 */
	public long getMyKeyTimestamp()
	{
		return myKeyTimestamp;
	}
	
	/**
	 * Sets the private key to specified value.
	 * @param pKey	a private key.
	 */
	public void setMyPrivateKey(PrivateKey pKey)
	{
		myPrivateKey=pKey;
	}

	/**
	 * Update the personal key to a new value
	 *
	 * @return the new personal key.
	 *
	 * @throws IllegalArgumentException if the algorithm specified does not
	 * 		   exist
	 */
	public SecretKeySpec updateMyPersonalKey()
	{
		KeyGenerator keyGen;

		try
		{
			/*
			 * If the key distribution mode is set to shared group key, the personal
			 * key of the local node is the group key.
			 */
			if (keyMode.equals("GroupKeys"))
			{
				myPersonalKey = groupKey;
			}
			/*
			 * Otherwise, a new personal key is created.
			 */
			else
			{
				keyGen = KeyGenerator.getInstance(confidentialityAlgorithm);
				keyGen.init(keySize * 8);

				myPersonalKey =
					new SecretKeySpec(keyGen.generateKey().getEncoded(),
									  confidentialityAlgorithm);
			}
		} catch (NoSuchAlgorithmException nsae) {
			throw new IllegalArgumentException("Key algorithm specified does not exist!");
		}

		return myPersonalKey;
	}

	/**
	 * Add a new key corresponding to the neighbor at the given address.
	 *
	 * @param LA 	The address of the neighbor whose key is being added
	 * @param key 	The neighbor's key
	 * @param timestamp The creation time of the neighbor's key
	 */
	public void addKey(I_LogicalAddress LA, SecretKeySpec key, long timestamp)
	{
		
		KeyVaultEntry entry = this.getEntry(LA);
		
		if(entry!=null){
			entry.setKey(key);
			entry.setKeyTimestamp(timestamp);
		}
		
	}

	/**
	 * Remove the key corresponding to the neighbor at the given address.
	 *
	 * @param LA The address of the neighbor whose key is being removed.
	 */
	public void removeKey(I_LogicalAddress LA)
	{
		KeyVaultEntry entry = this.getEntry(LA);
		
		if(entry!=null){
			entry.setKey(null);
		}
		
	}
	
	/**
	 * Check the key store to see if there is an entry for the given node.
	 *
	 * @param LA the Logical address of the node.
	 *
	 * @return true if the node has an entry in the key store.
	 */
	public boolean haveKey(I_LogicalAddress LA)
	{
		//if (keyMode.equals("EndToEnd")) {
		if (keyMode.equals("GroupKeys")) {
			return true;
		}
		
		KeyVaultEntry entry = this.getEntry(LA);
		if(entry!=null){
			return entry.getKey()!=null;
		}else
			return false;
		
	}
	
	/**
	 * Check the key store to see if the key has been mark as invalid.
	 *
	 * @param LA the Logical address of the node.
	 *
	 * @return true if the node has an entry in the key store.
	 */
	public boolean isKeyInvalid(I_LogicalAddress LA)
	{
		if (keyMode.equals("GroupKeys")) {
			return false;
		}
		
		KeyVaultEntry entry = this.getEntry(LA);
		if(entry!=null){
			return entry.getKey()==null || entry.isKeyInvalid();
		}else
			return true;
		
	}
	public void setKeyInvalid(I_LogicalAddress LA, boolean inValid){
		if (keyMode.equals("GroupKeys")) {
			return;
		}
		
		KeyVaultEntry entry = this.getEntry(LA);
		if(entry!=null){
			entry.setKeyInvalid(inValid);
		}
	}
	

	/**
	 * Get a neighbor's personal key.
	 *
	 * @param LA Logical Address of the neighbor whose key is requested.
	 *
	 * @return The neigbhor's personal key.
	 */
	public Key getKey(I_LogicalAddress LA)
	{
		//System.out.println("KeyVault:getKey: keyMode is "+keyMode);
		
		if (keyMode.equals("NeighborhoodKeys")) {
			
			KeyVaultEntry entry = this.getEntry(LA);
			
			if(entry==null)
				return null;
			else
				return entry.getKey();
			
		}
		else
		{
			return groupKey;
		}
	}
	
	public SecretKeySpec getGroupKey(){
		return this.groupKey;
	}

	/**
	 * Get the creation time of a neighbor's key.
	 *
	 * @param LA Logical Address of the neighbor whose key timestamp is
	 *  requested
	 *
	 * @return The neighbor's key timestamp
	 */
	public long getKeyTimestamp(I_LogicalAddress LA) {
		String address = LA.toString();
		
		KeyVaultEntry entry = this.getEntry(LA);
		
		if(entry!=null){
			return entry.getKeyTimestamp();
		}else
			return 0;
		
	}

	/**
	 * Add a new certificate corresponding to the node at the given address.
	 *
	 * @param LA 	The address of the node whose certificate is being added.
	 * @param cert 	a certificate.
	 */
	public void addCert(I_LogicalAddress LA, X509Certificate cert)
	{
//		TODO: remove debug statement if no longer needed
			  //System.out.println("KeyVault: add certificate " + LA + " to neighborCerts table.");
		
		KeyVaultEntry entry = this.getEntry(LA);
		
		if(entry!=null){
			entry.setCert(cert);
		}
		
	}

	/**
	 * Remove the certificate corresponding to the node at the given
	 * address.
	 *
	 * @param LA The address of the node whose certificate is being removed.
	 */
	public void removeCert(I_LogicalAddress LA)
	{
		KeyVaultEntry entry = this.getEntry(LA);
		
		if(entry!=null)
			entry.setCert(null);
		
	}

	

	/**
	 * Check the certificate store to see if there is an entry for the given
	 * node.
	 *
	 * @param LA the Logical address of the node.
	 *
	 * @return true if the node has an entry in the certificate store.
	 */
	public boolean haveCert(I_LogicalAddress LA)
	{
		//if (keyMode.equals("EndToEnd")) {
/*
		if (keyMode.equals("GroupKeys")) {
			return true;
		}
*/		KeyVaultEntry entry = this.getEntry(LA);

		if(entry!=null)
			return entry.getCert()!=null;
		else
			return false;
		
	}

	/**
	 * Gets the certificate of the given node.
	 *
	 * @param LA logical address of a node.
	 *
	 * @return certificate of the given node.
	 */
	public X509Certificate getCertificate(I_LogicalAddress LA)
	{
		X509Certificate cert;
		
		KeyVaultEntry entry = this.getEntry(LA);
		
		if (entry!=null)
		{
			return entry.getCert();
			
		}
		else
		{
			return null;
		}
	}

	/**
	 * Gets the public key of the given node.
	 *
	 * @param LA logical address of a node.
	 *
	 * @return public of the given node.
	 */
	public PublicKey getPublicKey(I_LogicalAddress LA)
	{
		X509Certificate cert;

		cert = getCertificate(LA);

		if (cert != null)
		{
			return cert.getPublicKey();
		}
		else
		{
			return null;
		}
	}

	/**
	 * Get the message key.
	 *
	 * @return a new message key if key distribution mode is set to 
	 *  hop-by-hop; the group key otherwise.
	 *
	 * @throws IllegalArgumentException if the encryption/decryption
	 *  algorithm is not supported.
	 */
	public SecretKey getNewMsgKey()
	{
		KeyGenerator keyGen;
		SecretKey newKey;

		if (keyMode.equals("NeighborhoodKeys"))
		{
			try
			{
				keyGen = KeyGenerator.getInstance(confidentialityAlgorithm);
				keyGen.init(keySize * 8);

				newKey =
					new SecretKeySpec(keyGen.generateKey().getEncoded(),
									  confidentialityAlgorithm);
			}
			catch (NoSuchAlgorithmException nsae)
			{
				throw new IllegalArgumentException("Confidentiality algorithm specified does not exist!");
			}

			return newKey;

			//			TODO: may not want this permanently
			//			 if it shows major speed improvement, just make new msg key less often
			//			return myPersonalKey;
		}
		else
		{
			return groupKey;
		}
	}

	/**
	 * Gets the wrapped message key length.
	 *
	 * @return the wrapped message key length.
	 */
	public int getWrappedKeySize()
	{
		return wrappedKeySize;
	}


	/**
	 * Check if a sequence number is valid for a node. In the following two cases,
	 * the sequence number is considered valid:
	 * 1. a node, from which the sequence number is obtained, is not in the 
	 *    sequence number hashtable. In this case, the node and its sequence number is
	 *    added to the table.
	 * 2. a node is in the sequence number hashtable and the received sequence number 
	 *    is greater than the sequence number currently stored in the table. In this 
	 *    case, the newly received sequence number replaces the currently stored one.
	 *
	 * @param src 	logical address of the node from which the sequence
	 *  number is obtained.
	 * @param sequenceNum 	the received sequence number.
	 *
	 * @return	true if the received sequence number is valid; false otherwise.
	 */
	public boolean checkSequenceNum(I_LogicalAddress src, int sequenceNum)
	{
		KeyVaultEntry entry = this.getEntry(src);
		
		if(entry!=null){
			if(entry.getSequenceNum()==-1){
				entry.setSequenceNum(sequenceNum);
				return true;
			}else if(sequenceNum > entry.getSequenceNum()){
				entry.setSequenceNum(sequenceNum);
				return true;
			}else{
				//System.err.println("seq invalid current="+entry.getSequenceNum()+" new="+sequenceNum);
				return false;
			}
			
		}
		
		return true;
		
	}

	/**
	 * Gets the next valid sequence number for protocol messages.
	 * 
	 * @return the next valid sequence number.
	 */
	public int getNextSequenceNum()
	{
		
		/*
		 * If sequence number exceed the maximum limit,
		 * reset the sequence number, update the neighborhood key,
		 * and send key updates to all neighbors.
		 */
		if(mySequenceNum>=2147483647){
			mySequenceNum = 0;
			sp.updatePersonalKey(null);
			sp.sendKeyUpdateToAllNeighbors(null);
		}
		
		mySequenceNum++;

		return mySequenceNum;
	}

	/**
	 * Sets the sequence number for the given node in sequence 
	 * number hashtable.
	 * 
	 * @param nodeLA 	the logical address of the node to which the
	 * 	sequence number is bound.
	 * @param seqNum 	the sequence number.
	 */
	public void setSequenceNum(I_LogicalAddress nodeLA, int seqNum)
	{
		KeyVaultEntry entry = this.getEntry(nodeLA);
		
		if(entry!=null)
			entry.setSequenceNum(seqNum);
		
		
	}

	/**
	 * Sets the sequence number to given value.
	 * 
	 * @param seqNum 	the sequence number.
	 */
	public void setSequenceNum(int seqNum)
	{
		mySequenceNum = seqNum;
	}


	/**
	 * Sets the security processor reference in the object.
	 * 
	 * @param spObj the security processor.
	 */
	public void setSecurityProcessor(SecurityProcessor spObj)
	{
		sp = spObj;
	}

//TODO: check. Should we key a reference to overlay node in the key vault?
	/**
	 * Gets the reference to the overlay node.
	 * 
	 * @return overlay node.
	 */
	public I_Node getNode()
	{
		if (sp == null)
		{
			return null;
		}
		else
		{
			return sp.getNode();
		}
	}


	/**
	 * Set the time when the last secure message was received from the given
	 * node. It calls the method with same name in SecurityProcessor.
	 *
	 * @param nodeLA 	the logical address of the node from which the secure
	 * 	message	is received.
	 */
	public void setLastSecureMessageTime(I_LogicalAddress nodeLA)
	{
		if (sp != null)
		{
			sp.setLastSecureMessageTime(nodeLA);
		}
	}
	
	
	/**
	 * The hash table that stores a set of KeyVaultEntry objects. Each object
	 * represents a node with which this node exchanges certificate with.
	 */
	Hashtable table;
	
	/**
	 * Set an entry as the overlay neighbor
	 * @param la
	 */
	public void setEntryToNeighbor(I_LogicalAddress la){
		KeyVaultEntry entry = getEntry(la);
		if(entry!=null){
			if(!entry.isOverlayNeighbor()){
				entry.setIsOverlayNeighbor(true);
				
			}
		}
	}
	/**
	 * Get the addresspair according to the physical address
	 */
	public I_AddressPair getAddressPairByPA(I_NetworkAddress pa){
		//System.err.println("Search addresspair for "+pa);
		Enumeration keys= table.keys();
		int i=0;
		while(keys.hasMoreElements())
		{
			I_AddressPair ap = ((KeyVaultEntry) table.get(keys.nextElement())).getAddressPair();
			if(ap.getPhysicalAddress().equals(pa)){
				//System.err.println("Found: "+ap);
				return ap;
			}
			 
			i++;
		}
		//System.err.println("not found");
		return null;
	}
	/**
	 * Set entry as no neighbor
	 * @param la
	 */
	public void setEntryToNoNeighbor(I_LogicalAddress la){
		KeyVaultEntry entry = getEntry(la);
		if(entry!=null){
			if(entry.isOverlayNeighbor()){
				entry.setIsOverlayNeighbor(false);
				
			}
		}
	}
	/** 
	 * get the neigborhood key for this entry
	 */
	public SecretKeySpec getMyKeyFor(I_AddressPair dst_ap, I_AddressPair src_ap){
		if(dst_ap==null || dst_ap.equals(src_ap)){
			//System.err.println("use personal key for: "+dst_ap);
			return sp.vault.getMyPersonalKey();
		}
		KeyVaultEntry entry = getEntry(dst_ap);
		if(entry!=null){
			if(entry.isOverlayNeighbor()){
				//System.err.println("Neighbor key");
				//System.err.println("use personal key for: "+dst_ap);
				return sp.vault.getMyPersonalKey();
			}else{
				//System.err.println("NoNeighbor key");
				//System.err.println("use NoNeighbor key for: "+dst_ap);
				return sp.vault.getMyNoNeighborKeyFor(dst_ap.getLogicalAddress());
			}
		}
		else return null;
	}
	/**
	 * get the neighborhood key timestamp for
	 */
	public long getMyKeyTimestampFor(I_AddressPair dst_ap, I_AddressPair src_ap){
		if(dst_ap==null || dst_ap.equals(src_ap)){
			return sp.vault.getMyKeyTimestamp();
		}
		KeyVaultEntry entry = getEntry(dst_ap);
		if(entry!=null){
			if(entry.isOverlayNeighbor()){
				return sp.vault.getMyKeyTimestamp();
			}else{
				return sp.vault.getMyNoNeighborKeyTimestamp(dst_ap.getLogicalAddress());
			}
		}
		else return 0;
	}
	
	
	/**
	 * Gets the physical address of a node if this node is in the key neighbor table.
	 * 
	 * @param la	the logical address of a neighbor.
	 * @return	the physical address of the node if it is in the
	 * 	key neighbor table. Null otherwise.
	 */
	public I_PhysicalAddress getPhysicalAddress(I_LogicalAddress la)
	{
		KeyVaultEntry entry = getEntry(la);
		if (entry != null)
		{
			return entry.getAddressPair().getPhysicalAddress();
		}
		else
		{
			return null;
		}
	}
	
	/**
	 * Add an entry in the key neighbor table.
	 * 
	 * @param ap	the address pair of a node.
	 * @param now	the time that this entry is added to the key neighbor table.
	 */
	public boolean addEntry(I_AddressPair ap, long now)
	{
		synchronized (table)
		{
			KeyVaultEntry entry = getEntry(ap);
			if (entry == null)
			{
//TODO: remove debug statement if no longer needed
				//System.out.println("addEntry: add Entry " + ap + " to KeyNeighbors");

				table.put(ap.getLogicalAddress().toString(), new KeyVaultEntry(ap, false, now));
				

				//System.out.println("Key neighbor table changes: neighbor "+ap.getLogicalAddress().toString()+" is added");
				
				return true;

			}
			else
			{
//TODO:  remove debug statement if no longer needed
				//System.out.println("update Entry " + ap + " in KeyNeighbors");

				entry.update(now);
				return false;
			}
		}
	}
	
	/**
	 * Checks (with logical address) if a node is in the neighbor key table.
	 * 
	 * @param src	the logical address of a node.
	 * @return	true if the node is in the key neighbor table; false otherwise.
	 */
	public boolean contains(I_LogicalAddress src) {
		return getEntry(src) != null;
	}
	
	/**
	 * Checks (with address pair) if a node is in the key neighbor table.
	 * 
	 * @param ap	the logical address of a node.
	 * @return	true if the node is in the key neighbor table; false otherwise.
	 */
	public boolean contains(I_AddressPair ap) {
		return getEntry(ap) != null;
	}
	
	/**
	 * Gets an entry (with logical address) from the key neighbor table.
	 * 
	 * @param la	logical address of a node.
	 * @return	the KeyVaultEntry object which has the specified logical address;
	 * 	null if no entry contains the specified logical address.
	 */
	public KeyVaultEntry getEntry(I_LogicalAddress la)
	{
		return (KeyVaultEntry) table.get(la.toString());
	}
	
	/**
	 * Gets an entry (with address pair) from the key neighbor table.
	 * 
	 * @param ap	address pair of a node.
	 * @return	the KeyVaultEntry object which has the logical address in the specified
	 * 	address pair; null if no entry is found.
	 */
	public KeyVaultEntry getEntry(I_AddressPair ap) {
		KeyVaultEntry entry = (KeyVaultEntry) table.get(ap.getLogicalAddress().toString());

		if (entry != null
			&& !entry.getAddressPair().getPhysicalAddress().equals(ap.getPhysicalAddress())) {
			removeEntry(ap.getLogicalAddress());
			entry = null;
			
		}

		return entry;
	}

	/**
	 * Remove an entry which has the specified logical address from the key neighbor table.
	 * 
	 * @param la	the logical address of a node.
	 */	
	public void removeEntry(I_LogicalAddress la)
	{
		synchronized (table)
		{
			if (table.get(la) != null)
			{
				table.remove(la);
				
			}
		}
	}
	
	/**
	 * Clean up the key neighbor table by removing all entries which are
	 * timeout.
	 * 
	 * @param key_neighbor_timeout	timeout time of an entry in the key neighbor
	 * 	table.
	 * @param now	current time.
	 */
	public void clearExpiredNeighbors(long key_neighbor_timeout, long now)
	{
		Enumeration keys= table.keys();

		while(keys.hasMoreElements())
		{
			Object key=keys.nextElement();
			KeyVaultEntry kn = ((KeyVaultEntry) table.get(key));
			
			/*
			 * If the key is time out, remove it.
			 */
			if (kn.time_stamp + key_neighbor_timeout < now)
			{
				table.remove(key);
			
				//System.out.println("KeyNeighborTable:clearExpiredNeighbors: Key neighbor table changes: neighbor "+((String) key)+" is removed");
				
				
			}

		}
	}
	
	/**
	 * @return	the address pairs of all nodes in the key neighbor table.
	 */
	public I_AddressPair[] getAllKeyNeighbors()
	{
		I_AddressPair[] neighbors = new I_AddressPair[table.size()];

		Enumeration keys= table.keys();
		int i=0;
		while(keys.hasMoreElements())
		{
			neighbors[i] = ((KeyVaultEntry) table.get(keys.nextElement())).getAddressPair();
			i++;
		}

		return neighbors;
	}
	
	
	/**
	 * Check if this node is waiting for the certificate reply from specified node.
	 * 
	 * @param ap	address pair of the node to be checked.
	 * @param cert_request_timeout	the timeout time of waiting for certificate reply.
	 * @param now	current time.
	 * @return	true if timeout happens; false otherwise.
	 */
	public boolean isWaitingCertReply(I_AddressPair ap, long cert_request_timeout, long now)
	{
		if (getEntry(ap) != null)
		{
			return getEntry(ap).isWaitingCertReply(cert_request_timeout, now);
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * Check if this node is waiting for the key update from specified node.
	 * 
	 * @param ap	address pair of the node to be checked.
	 * @param key_request_timeout	the timeout time of waiting for key update reply.
	 * @param now	current time.
	 * @return	true if timeout happens; false otherwise.
	 */
	public boolean isWaitingKeyUpdate(I_AddressPair ap, long key_request_timeout, long now) {
		if (getEntry(ap) != null)
			return getEntry(ap).isWaitingKeyUpdate(key_request_timeout, now);
		else
			return false;
	}
	
	/**
	 * Sets the time of last certificate request is sent for a entry in the key neighbor table.
	 * 
	 * @param ap	address pair of the node to which the certificate request is sent.
	 * @param now	the time when the certificate request is sent.
	 */
	public void certRequestSent(I_AddressPair ap, long now)
	{
		if (getEntry(ap) != null)
		{
			getEntry(ap).certRequestSent(now);
		}
	}
	
	/**
	 * Resets the time of last certificate request is sent to zero for a entry 
	 * in the key neighbor table.
	 * 
	 * @param ap	address pair of the node from which the certificate reply is received.
	 */
	public void certReplyReceived(I_AddressPair ap)
	{
		if (getEntry(ap) != null)
		{
			getEntry(ap).certReplyReceived();
		}
	}
	
	/**
	 * Sets the time of last key request is sent for a entry in the key neighbor table.
	 * 
	 * @param ap	address pair of the node to which the key request is sent.
	 * @param now	the time when the key request is sent.
	 */
	public void keyRequestSent(I_AddressPair ap, long now)
	{
		if (getEntry(ap) != null)
		{
			getEntry(ap).keyRequestSent(now);
		}
	}

	/**
	 * Resets the time of last key request is sent to zero for a entry 
	 * in the key neighbor table.
	 * 
	 * @param ap	address pair of the node from which the key update is received.
	 */
	public void keyUpdateReceived(I_AddressPair ap)
	{
		if (getEntry(ap) != null)
		{
			getEntry(ap).keyUpdateReceived();
		}
	}
	
	/**
	 * Updates the time for a entry in the key neighbor table when it is modified.
	 * 
	 * @param ap	address pair of a node.
	 * @param now	the time when the entry corresponding to the address pair is modified.
	 */
	public void updateEntry(I_AddressPair ap, long now)
	{
		if (getEntry(ap) != null)
		{
			getEntry(ap).update(now);
		}
	}
	
	/**
	 * Updates the time of last secure message is received for a entry in 
	 * the key neighbor table. 
	 * 
	 * @param ap	address pair of the node from which the secure message is received.
	 * @param now	the time when secure message is received.
	 */
	public void secureMessageReceived(I_AddressPair ap, long now) {
		if (getEntry(ap) != null)
		{
			getEntry(ap).secureMessageReceived(now);
		}
	}
	
	/**
	 * Gets the time of last secure message is received for a entry in 
	 * the key neighbor table.
	 * 
	 * @param ap	address pair of a node.
	 * @return	the time of last secure message is received. 0 if no
	 * 	entry is found.
	 */
	public long getLastTimeSecureMessageRecieved(I_AddressPair ap) {
		if (getEntry(ap) != null)
		{
			return getEntry(ap).getLastTimeSecureMessageRecieved();
		}
		else
		{
			return 0;
		}
	}

	
}



class KeyVaultEntry{
	/** address pair of the neighbor */
	private I_AddressPair ap = null;

	/** the time when the last message (protocol or secure) is received
	 * from the node corresponding to this entry.
	 */
	long time_stamp;

	/** the time when the last certificate message is sent out. */
	long lastTimeCertRequestSent = -100000;

	/** the time when the last key request message is sent out. */
	long lastTimeKeyRequestSent = -100000;

	/** the time when the last secure message is received. */
	long lastSecureMessgeReceived = -100000;

	/** flag showing if this entry is neighbor */
	boolean IsOverlayNeighbor;
	
	X509Certificate cert;
	
	SecretKeySpec key;
	
	long keyTimestamp;
	
	int sequenceNum;
	
	SecretKeySpec myNoNeighborKey;
	
	long myNoNeighborKeyTimestamp;
	
	boolean keyInvalid = false;
	
	/**
	 * Constructor.
	 * 
	 * @param _ap	address pair of a node.
	 * @param _isOverlayNeighbor	neighbor flag.
	 * @param ts	time when the last secure message is received.
	 */
	public KeyVaultEntry(I_AddressPair _ap, boolean _isOverlayNeighbor, long ts)
	{
		ap = _ap;
		IsOverlayNeighbor = _isOverlayNeighbor;
		time_stamp = ts;
		sequenceNum = -1;
	}
	
	/**
	 * @return	time when the last secure message is received from the node 
	 * corresponding to this entry.
	 */
	public long getTimeStamp()
	{
		return time_stamp;
	}

	/**
	 * Updates the time for this entry in the key neighbor table when it is modified.
	 * 
	 * @param now	the time when last message is received.
	 */
	public void update(long now)
	{
		time_stamp = now;
	}

	/**
	 * Checks if this node is waiting for the certificate reply message

	 * @param cert_request_timeout	timeout time of waiting for certificate request.
	 * @param now	current time.

	 * @return	true if timeout happens; false otherwise.
	 */
	public boolean isWaitingCertReply(long cert_request_timeout, long now)
	{
		return lastTimeCertRequestSent>=0 && ((lastTimeCertRequestSent + cert_request_timeout) >= now);
	}
	
	/**
	 * Checks if this node is waiting key update message.
	 * 
	 * @param key_request_timeout	timeout time of waiting for key update.
	 * @param now	current time.
	 */
	public boolean isWaitingKeyUpdate(long key_request_timeout, long now)
	{
		return lastTimeKeyRequestSent>=0 && (lastTimeKeyRequestSent + key_request_timeout) >= now;
	}
	
	/**
	 * 
	 * @return	true if this entry is a neighbor of the local node; false otherwise.
	 */
	public boolean isOverlayNeighbor()
	{
		return IsOverlayNeighbor;
	}
	public void setIsOverlayNeighbor(boolean ion){
		IsOverlayNeighbor = ion;
	}
	
	/**
	 * Sets the time when last certificate request is sent.
	 * 
	 * @param now	the time when the last certificate request message is sent out.
	 */
	public void certRequestSent(long now)
	{
		lastTimeCertRequestSent = now;
	}
	
	/**
	 * 
	 * Resets the time of when last certificate request message is sent out to zero.
	 * It is doen when the certificate reply is received.
	 */
	public void certReplyReceived()
	{
		lastTimeCertRequestSent = -10000;
	}
	
	/**
	 * Sets the time when last key request is sent.
	 * 
	 * @param now	the time when the last key request message is sent out.
	 */
	public void keyRequestSent(long now)
	{
		lastTimeKeyRequestSent = now;
	}
	
	/**
	 * 
	 * Resets the time of when last key request message is sent out to zero.
	 * It is doen when the key update is received.
	 */
	public void keyUpdateReceived()
	{
		lastTimeKeyRequestSent = -10000;
	}
	
	/**
	 * 
	 * @return	the address pair of this entry.
	 */
	public I_AddressPair getAddressPair()
	{
		return ap;
	}
	
	/**
	 * Sets the time when the last secure message is received from
	 * the node corresponding to this entry.
	 * 
	 * @param now	the time when the last secure message is received.
	 */
	public void secureMessageReceived(long now)
	{
		lastSecureMessgeReceived = now;
	}
	
	/**
	 * 
	 * @return	the time when the last secure message is received from
	 * the node corresponding to this entry.
	 */
	public long getLastTimeSecureMessageRecieved()
	{
		return lastSecureMessgeReceived;
	}
	
	/**
	 * @return Returns the cert.
	 */
	public X509Certificate getCert() {
		return cert;
	}
	/**
	 * @param cert The cert to set.
	 */
	public void setCert(X509Certificate cert) {
		this.cert = cert;
	}
	/**
	 * @return Returns the key.
	 */
	public SecretKeySpec getKey() {
		return key;
	}
	/**
	 * @param key The key to set.
	 */
	public void setKey(SecretKeySpec key) {
		this.key = key;
		if(key!=null)
			this.keyInvalid = false;
		else
			this.keyInvalid = true;
	}
	/**
	 * @return Returns the keyTimestamp.
	 */
	public long getKeyTimestamp() {
		return keyTimestamp;
	}
	/**
	 * @param keyTimestamp The keyTimestamp to set.
	 */
	public void setKeyTimestamp(long keyTimestamp) {
		this.keyTimestamp = keyTimestamp;
	}
	/**
	 * @return Returns the myNoNeighborKey.
	 */
	public SecretKeySpec getMyNoNeighborKey() {
		return myNoNeighborKey;
	}
	/**
	 * @param myNoNeighborKey The myNoNeighborKey to set.
	 */
	public void setMyNoNeighborKey(SecretKeySpec myNoNeighborKey) {
		this.myNoNeighborKey = myNoNeighborKey;
	}
	/**
	 * @return Returns the myNoNeighborKeyTimestamp.
	 */
	public long getMyNoNeighborKeyTimestamp() {
		return myNoNeighborKeyTimestamp;
	}
	/**
	 * @param myNoNeighborKeyTimestamp The myNoNeighborKeyTimestamp to set.
	 */
	public void setMyNoNeighborKeyTimestamp(long myNoNeighborKeyTimestamp) {
		this.myNoNeighborKeyTimestamp = myNoNeighborKeyTimestamp;
	}
	/**
	 * @return Returns the sequenceNum.
	 */
	public int getSequenceNum() {
		return sequenceNum;
	}
	/**
	 * @param sequenceNum The sequenceNum to set.
	 */
	public void setSequenceNum(int sequenceNum) {
		this.sequenceNum = sequenceNum;
	}
	/**
	 * @return Returns the keyInvalid.
	 */
	public boolean isKeyInvalid() {
		return keyInvalid;
	}
	/**
	 * @param keyInvalid The keyInvalid to set.
	 */
	public void setKeyInvalid(boolean keyInvalid) {
		this.keyInvalid = keyInvalid;
	}
}
