/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.io.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

import hypercast.util.XmlUtil;

/**
 * This class provides a buffer which handles all the functionality for the producer/consumer
 * problem. It assumes that there can be multiple producers and consumers.
 * This is a FIFO style Buffer.
 *
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Dongwen Wang
 * @author Michael Nahas
 * @version	2.0, Feb. 28, 2001
 */

public class MessageBuffer implements I_Stats
{
    /** HyperCastConfig object. */
	private HyperCastConfig config;

    /** The number of messages in the queue */
	private int numberOfElements;

    /** The queue of messages. */
	private I_Message[] queue;

    /** The location of the oldest message in the queue.  (The queue is circular, so the start moves.) */
	private int start;

    /** Timeout for reads. */
    private int soTimeout;
	
    /** Number of threads waiting to read. */
    private int readersWaiting;

    /** Number of threads waiting to write. */
    private int writersWaiting;

    private String name; // debugging

	/** The property prefix. */
	protected String ATTRIBUTE_PREFIX = "/Public/ReceiveBuffer/";
	
	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	StatsProcessor statsPro;
	
	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;

	
    public static final int SOCKET_TIMEOUT_INFINITY = 0;

    /** 
     * Constructs the MessageBuffer with a config file and a size.
     *
     * @param config The configuration object.
     * @param queueSize The size of the buffer (number of messages)
     */
    public MessageBuffer (HyperCastConfig config, int queueSize, String name)
    {
        this (config, queueSize, SOCKET_TIMEOUT_INFINITY);
        this.name = name;
    }

    public MessageBuffer (HyperCastConfig config, int queueSize)
    {
        this (config, queueSize, SOCKET_TIMEOUT_INFINITY);
    }

    /** 
     * Constructs the MessageBuffer with a config file and a size.
     *
     * @param config The configuration object.
     * @param queueSize The size of the buffer (number of messages)
     * @param socketTimeout The amount of time (in milliseconds) to wait before giving up on a read operation.
     */
    public MessageBuffer (HyperCastConfig config, int queueSize, int socketTimeout)
    {
        this.config = config;
        this.soTimeout = socketTimeout;

        this.queue = new I_Message [queueSize];
        this.start = 0;
        this.numberOfElements = 0;

        this.readersWaiting = 0;
        this.writersWaiting = 0;

        InitStatisticsStructure ();
    }


    /** Get the timeout time for a read operation. */
    int getSoTimeout() 
    {
        return soTimeout;
    }
    
    /** Set the timeout time for a read operation. */
    void setSoTimeout(int i) 
    {
        if (i < 0)
            throw new IllegalArgumentException("called setSoTimeout with a negative argument!");

        soTimeout = i;
    }
    


	/** 
	 * Reads a message from the BoundedBuffer. <p>
	 * Algorithm: <br>
	 *  1. Waits till the queue is non-empty <br>
	 *  2. Copies first element in the queue <br>
	 *  3. Increments the start pointer modulo queue length <br>
	 *  4. Decreases size of the queue <br>
	 *  5. Notifies any other process waiting <br>
	 *  6. Returns the copied element <br>
	 * 
	 * @return	the first message in the queue
	 */

	synchronized public I_Message read() throws InterruptedException
    {
		I_Message result;

		// 1. Waits till queue is non-empty
        if (soTimeout == 0) 
        {
            try 
            {
                readersWaiting++;
                while(numberOfElements == 0)
                    wait();
            }
            finally 
            {
                readersWaiting--;
            }
        }
        else 
        {
            // Inaccurate, but we can live with that.
            if (numberOfElements == 0)
            {
                try 
                {
                    readersWaiting++;
                    wait(soTimeout);
                } 
                finally    
                {
                    readersWaiting--;
                }
                if (numberOfElements == 0)
                    return null;
            }   
        }

		// 2. Copies first element in the queue
		result = queue[start];
		queue[start] = null;

		// 3. Increments the start pointer modulo queue length
		start = (start + 1) % queue.length;
		
		// 4. Decreases size of the queue
		numberOfElements--;

		// 5. Notifies any writing elements waiting on this (only writing threads could be waiting.)
		notify();

		// 6. Returns result
		return result;
	}



	/** 
	 * Writes a message to the DropTailMessageBuffer <P>
	 * Algorithm: <br>
	 *  1. checks if queue is full <br>
	 *  2. Places New packet at the end of the queue <br>
	 *  3. Increments number of elements <br>
	 *  4. Notifies all threads waiting <br>
	 * 
	 * @param	msg		a message to be added into the queue
	 */

	synchronized public void nonblockingWrite(I_Message msg){

		// if the queue is full, punt
		if (numberOfElements == queue.length) {
			config.err.println(name + " DropTailMessageBuffer: No room in queue, punting.");
			return;
		}

		// Places new packet at the end of the queue
		int index = (start + numberOfElements) % queue.length;
		queue[index] = msg;

		// Increments number of elements
		numberOfElements++;

		// Notifies all waiting threads
        // MDN: I don't think notifyAll is necessary.
        // MDN: For each write, only one reader can read.
		//notifyAll();
        notify();
	}





	/** 
	 * Writes a message to the BoundedBuffer <P>
	 * Algorithm: <br>
	 *  1. checks if queue is full <br>
	 *  2. Places New packet at the end of the queue <br>
	 *  3. Increments number of elements <br>
	 *  4. Notifies all threads waiting <br>
	 * 
	 * @param	msg		a message to be added into the queue
	 */

	synchronized public void blockingWrite(I_Message msg) throws InterruptedException {
		// if the queue is full, wait
		while (numberOfElements == queue.length) 
        {
            try 
            {
                try 
                {
                    writersWaiting++;
                    wait();
                }
                finally 
                {
                    writersWaiting--;
                }
            }
            catch (InterruptedException ie) 
            {
                notify();
                throw ie;
            }
        }
        

		//System.out.println("(" + numberOfElements + ") BoundedBuffer write: " + Buffer.toString());

		// Places new packet at the end of the queue
		int index = (start + numberOfElements) % queue.length;
		queue[index] = msg;

		// Increments number of elements
		numberOfElements++;

		// Notifies a waiting read thread (Only a read thread could be wait()ing.)
        notify();
	}

	

	/** 
	 * Peeks the message at the head of the queue <p>
	 * Algorithm: <br>
	 *  If queue is empty return null, else return the packet at the head of the queue
	 * @return	message at the head of the queue, but does not remove it
	 */
	synchronized public I_Message peek(){
		if (numberOfElements == 0) 
            return null;
		else
			return queue[start];
	}


	
	/** 
	 * Checks if there is any messages available in the queue.
	 * If queue is empty return false, else return true. 
	 * @return	false if queue is non empty, and true otherwise
	 */

	synchronized public boolean available(){
		if (numberOfElements > 0) 
            return true;
        else
            return false;
	}


	/** 
	 * Returns the number of messages in the queue.
	 * @return	the number of elements in the queue
	 */

	synchronized public int numInQueue() {
		return numberOfElements;
	}

	
	
	/* * 
	 * Clears the queue.
	 * Resets variables, and notifies all waiting processes.
	 */

// MDN: I'm not sure this is correct...  And we don't use it.
// 	synchronized public void clear(){
// 		numberOfElements = start = 0;

//         //wait all write threads.
// 		notifyAll();
// 	}


	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath(ATTRIBUTE_PREFIX+"StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		statsPro.addStatsElement("MaxMessages"/* name */, new MaxMessagesInQueue()/* I_Stats object */,
								1/* minoccurs */, 1/* maxoccurs */);
								
		statsPro.addStatsElement("NumOfMsgsInQueue", new NumOfMsgs(), 1, 1);

		statsPro.addStatsElement("NumOfReadersWaiting", new NumOfWaitingReaders(), 1, 1);

		statsPro.addStatsElement("NumOfWritersWaiting", new NumOfWaitingWriters(), 1, 1);
	}


	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

	/** @see I_Stats#getStats(Document, XPath)*/
	public Element[] getStats (final Document doc, final XPath xpath)
            throws HyperCastStatsException
	{
		return statsPro.getStatsResult(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("MessageBuffer:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("MessageBuffer:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("MessageBuffer:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("MessageBuffer:getSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	
	
	/*
	 * Below are a set of classes defined to manage simple statistics supported by this object.
	 */

	/**
	 * This class manages simple statistic "MaxMessages".
	 */
	class MaxMessagesInQueue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+queue.length);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			try
			{
				int size=Integer.parseInt(newValue);
				if(size!=queue.length && size>0)
				{
					config.log.println("MessageBuffer.BufferSize is reset to "+size);
					{
						queue = new I_Message[size];
						start = 0;
						numberOfElements = 0;
					}
				}
				return newValue;
			}
			catch(Exception e)
			{
				throw new HyperCastStatsException(statisticsName, HyperCastStatsException.InappropriateValue);
			}
		}


		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}

	}


	/**
	 * This class manages simple statistic "NumOfMsgsInQueue".
	 */
	class NumOfMsgs extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+numberOfElements);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}


		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "NumOfReadersWaiting".
	 */
	class NumOfWaitingReaders extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+readersWaiting);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
	}

	/**
	 * This class manages simple statistic "NumOfWritersWaiting".
	 */
	class NumOfWaitingWriters extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+writersWaiting);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	
}






