/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.*;
import hypercast.util.XmlUtil;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

/**
 * A buffer that holds messages and allows messages, or types of
 * messages, to have finite state machines associated with them. It is
 * through these finite state machine that interesting services can be
 * created. Finite state machines are created by extending the
 * I_MessageStoreFSM interface.
 * 
 * @author HyperCast Team
 * @author Greg Mattes
 * @version October 2004
 * @author Guimin Zhang
 * @version 2.0, May. 23, 2001
 */
public class MessageStore implements I_Stats
{
	/**
	 * Debug flag
	 */
	protected boolean debug = false;

    /**
     * Hashtable using uniqueId as hashcode.
     * (String -> I_MessageStoreFSM)
     *
     * Must be accessible to the finite state machines, access is
     * granted via protected access modifier rather than accessor
     * method(s). Due to this access choice synchronization must be
     * handled by the collection itself. Hashtable is used rather than
     * HashMap to address synchronization issues.
     */
    protected Hashtable store = new Hashtable();

    /**
     * The overlay socket object associated with this messagestore,
     * each socket can have at most one message store. The socket must
     * be accessible to the finite state machines.
     */
    protected OL_Socket socket;

    /**
     * The configuration object that holds information on how the
     * message store and finite state machines should be configured.
     * The configuration must be accessible to the finite state
     * machines.
     */
    protected HyperCastConfig config;

    /**
     * The Message Processor is responsible for reading messages from
     * the messageBuffer and calling MessageStore.processMessage() on
     * each message. It runs in its own thread of
     * execution. Essentially, the Message Processor *is* the "Message
     * Store Thread." Other parts of the system, e.g. the forwarding
     * engine, pass messages to the message store by invoking public
     * methods of the message store. These methods simply add the
     * messages to a queue that is read by the message processor.
     */
    private Message_Processor messageProcessor;

    /**
     * The buffer/queue used to hold the messages that are to be
     * processed; The message processor thread reads messages from
     * this buffer and process them. Obviously, this buffer must be
     * accessible to the message processor.
     */
    protected MessageBuffer buffer;

    /**
     * Used to notify an appliction when interesting events occur.
     * This value can be null if an application does not wish to
     * receive event notification. Individual finite state machines
     * must have access to this handler, they call it directly.
     */
    protected NotificationHandler notificationHandler = null;

    /**
     * Finite state machine identifiers
     */
    public static final short FSM_NO_SERVICE            = 0;
    public static final short FSM_H2H_ACK               = 1;
    public static final short FSM_E2E_ACK               = 2;
    public static final short FSM_DUPLICATE_ELIMINATION = 3;
    public static final short FSM_SYNCHRONIZATION       = 4;
    public static final short FSM_INCAST                = 5;
    public static final short FSM_BEST_EFFORT_ORDERING  = 6;
    public static final short FSM_NAMING                = 7;

	/**
	 * Message type conatsnts. The Finite state machine identifiers
	 * are grouped into message types.
	 */
	public static final short STREAM_MESSAGE            = 0;
	public static final short MESSAGESTORE_MESSAGE      = 1;
	public static final short OTHER_MESSAGE             = 2;

    private static final String serviceNumberToString (final short fsmid)
    {
        switch (fsmid)
        {
            case FSM_NO_SERVICE: return "FSM_NO_SERVICE";
            case FSM_H2H_ACK: return "FSM_H2H_ACK";
            case FSM_E2E_ACK: return "FSM_E2E_ACK";
            case FSM_DUPLICATE_ELIMINATION: return "FSM_DUPLICATE_ELIMINATION";
            case FSM_SYNCHRONIZATION: return "FSM_SYNCHRONIZATION";
            case FSM_INCAST: return "FSM_INCAST";
            case FSM_BEST_EFFORT_ORDERING: return "FSM_BEST_EFFORT_ORDERING";
            case FSM_NAMING: return "FSM_NAMING";
            default: return "<UNKNOWN FSMID: " + fsmid + ">";
        }
    }

	/**
	  * The counter to generate message id.  Counters are either two
	  * bytes or four bytes in length. Since they are unsigned
	  * quantities, we need to use a long (eight bytes) to be able to
	  * have four bytes unsigned.
	  *
	  * Counter is initialized to a random value in a manner similar to
	  * TCP sequence numbers.
	  */
	 private short counter = (short) Math.abs((short) ((new Random (System.currentTimeMillis ())).nextInt ()));

	 /**
	  * The maximum value that the counter can have.  We do not use
	  * Integer.MAX_VALUE since "int"s are signed, we need a 32 bit
	  * unsigned quantity, so we use the first 32 bits of a long.
	  */
	 private static int COUNTER_MAX_VALUE = 0xFFff;

	 /**The table to hold the sequence number when sending stream*/
	 private HashMap streamsequence = new HashMap();
	 
    /**
     * This table maps (Short -> FiniteStateMachineMetaOperations). It
     * is used to hold operations that need to be defined for types of
     * finite state machines rather than instances of finite state
     * machines. Each finite state machine type registers a meta
     * operations object with this table. Whenever the message store
     * needs to know how to handle a message based on its type - in
     * the absence of any particular finite state machine instance -
     * it will use the meta operations.
     */
    private static HashMap FSMMetaOpsTable = new HashMap();

    /*
     * This code registers the various meta operations objects with
     * the meta operations table.
     */
    static
    {
        FSMMetaOpsTable.put (new Short (FSM_E2E_ACK),               MessageStoreFSM_E2EACK.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_H2H_ACK),               MessageStoreFSM_H2HACK.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_DUPLICATE_ELIMINATION), MessageStoreFSM_DuplicateElimination.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_SYNCHRONIZATION),       MessageStoreFSM_Sync.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_INCAST),                MessageStoreFSM_Incast.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_BEST_EFFORT_ORDERING),  MessageStoreFSM_Ordering.registerMetaOperations());
        FSMMetaOpsTable.put (new Short (FSM_NAMING),                MessageStoreFSM_Naming.registerMetaOperations());
    }

    /* ------------------------------------
     * BEGIN variables set by configuration
     */

    /**True if the message store is used for synchronization*/
    private boolean synchronization;

    /** sleep time before the init synchronization*/
    private long initialSynchronizatinDelay;

    /** Time at which the most recent query was sent */
    private long startQueryTime;

    /*
     * END variables set by configuration
     * ----------------------------------
     */

    /*
     * Message Store states
     */
    protected final static int STOPPED  = 0;
    protected final static int STARTING = 1;
    protected final static int STARTED  = 2;
    protected final static int STOPPING = 3;
    
    /**
     * Current state of the store.  It goes from STOPPED to STARTING
     * to STARTED to STOPPING and then back to STOPPED.
     */
    private int state;

    /**
     * Used while starting and stoping the message processor thread to
     * ensure proper synchronization; protects <code>state</code> when
     * its value is changed.
     */
    private Object stateGuard = new Object();

    /**
     * The object used to synchronize thread shutdown.
     * Must be accessible by message processor.
     *
     */
    protected Object deathGuard = new Object();

    /* --------------------------
     * BEGIN Statistics variables
     */

    private int  dataMessagesReceived    = 0;
    private long dataBytesReceived       = 0L;
    private long controlMessagesReceived = 0L;
    private int  numberOfRetransmissions = 0;

    /* 
     * END Statistics variables
     * ------------------------
     */

    /**
     * A table of callbacks. Each type of finite state machine can
     * have one callback registered for it.
     * (Short -> I_ReceiveCallback) fsmId -> callback
     */
    // TODO: perhaps this should be synchronized if FSMs will have
    // access to it. Synchronized collection or just hashtable? or
    // make private and have methods to access it
    protected static HashMap fsmCallbacks = new HashMap();

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	private StatsProcessor statsPro;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;


	/**
	 * The object that stores pre-allocated or no-longer used 
	 * FSM instances. The FSM instances stored in this object are
	 * reused when new FSM instances are required.
	 */
	protected FSMPool fsmPool;


    /**
     * Constructs a MessageStore object with HyperCastConfig and OL_Socket.
     * @param config HyperCastConfig     
     * @param socket OL_Socket
     */

    public MessageStore (final OL_Socket socket, final HyperCastConfig config)
    {
        this.socket = socket;
        this.config = config;
        this.state = STOPPED;

		this.synchronization = (config.getTextAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Sync"))).equals("Synchronization");
		this.initialSynchronizatinDelay = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Sync/Synchronization/InitSync")); // TODO: should this be part of the Sync FSM?
		MessageStoreFSM_Sync.synchronizationPeriod = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Sync/Synchronization/SyncPeriod"));
		MessageStoreFSM_Sync.maximumTransmissionRate = config.getPositiveIntAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Sync/Synchronization/MaxTransmissionRate"));
		int bufferSize = config.getIntAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/MessageBufferSize"));

        this.buffer = new MessageBuffer (config, bufferSize, "Message Store");

//TODO: Remove debug statement
/*
			  System.out.println("this.synchronization="+this.synchronization);
			  System.out.println("this.initialSynchronizatinDelay="+this.initialSynchronizatinDelay);
			  System.out.println("MessageStoreFSM_Sync.synchronizationPeriod="+MessageStoreFSM_Sync.synchronizationPeriod);
			  System.out.println("MessageStoreFSM_Sync.maximumTransmissionRate="+MessageStoreFSM_Sync.maximumTransmissionRate);
			  System.out.println("bufferSize="+bufferSize);
*/

        InitStatisticsStructure();
        this.createFSMPool();

        final String activateNamingString = config.getTextAttribute (XmlUtil.createXPath ("/Public/NetworkService/Naming/StartService"));
        namingServiceStarted = config.booleanFromYesNoString (activateNamingString);
    }


	/**
	 * Create a pool to store pre-allocated FSM instances based on the configuration file.
	 */
	public void createFSMPool()
	{
		fsmPool = new FSMPool();

		int E2EACK_PredefinedNum = config.getIntAttribute 
			(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/PreAllocatedFSMs"));
		int H2HACK_PredefinedNum = config.getIntAttribute 
			(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/LocalACK/PreAllocatedFSMs"));
		int DuplicateElimination_PredefinedNum = config.getIntAttribute 
			(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/DuplicateElimination/PreAllocatedFSMs"));
		int Incast_PredefinedNum = config.getIntAttribute 
			(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Incast/PreAllocatedFSMs"));;
		int Ordering_PredefinedNum = config.getIntAttribute 
			(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Best-Effort-Ordering/PreAllocatedFSMs"));;

		//System.out.println("MessageStore.createFSMPool: create "+E2EACK_PredefinedNum+" MessageStoreFSM_E2EACK instances.");
		for (int i=0; i<E2EACK_PredefinedNum; i++)
		{
			fsmPool.addFSMInstance(new MessageStoreFSM_E2EACK(socket, this, config) ,FSM_E2E_ACK);
		}
		for (int i=0; i<H2HACK_PredefinedNum; i++)
		{
			fsmPool.addFSMInstance(new MessageStoreFSM_H2HACK(socket, this, config) ,FSM_H2H_ACK);
		}
		for (int i=0; i<DuplicateElimination_PredefinedNum; i++)
		{
			fsmPool.addFSMInstance(new MessageStoreFSM_DuplicateElimination(socket, this, config) ,FSM_DUPLICATE_ELIMINATION);
		}
		for (int i=0; i<Incast_PredefinedNum; i++)
		{
			fsmPool.addFSMInstance(new MessageStoreFSM_Incast(socket, this, config) ,FSM_INCAST);
		}
		for (int i=0; i<Ordering_PredefinedNum; i++)
		{
			fsmPool.addFSMInstance(new MessageStoreFSM_Ordering(socket, this, config) ,FSM_BEST_EFFORT_ORDERING);
		}

		if (this.synchronization)
		{
			int Sync_PredefinedNum = config.getIntAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/Sync/Synchronization/PreAllocatedFSMs"));
			for (int i=0; i<Sync_PredefinedNum; i++)
			{
				fsmPool.addFSMInstance(new MessageStoreFSM_Sync(socket, this, config) ,FSM_SYNCHRONIZATION);
			}
		}

                // "Pooling" of naming FSMs - since there can be only one naming FSM per message store, it is a special case. Its "pool" has a single object.
                final String preallocateNamingFSMString = config.getTextAttribute (XmlUtil.createXPath ("/Public/NetworkService/Naming/PreallocateFSM"));
                final boolean preallocateNamingFSM = config.booleanFromYesNoString (preallocateNamingFSMString);
                if (preallocateNamingFSM)
                {
                    this.naming = new MessageStoreFSM_Naming (socket, this, config);
                    statsPro.addStatsElement ("Naming", naming, 1, 1);
                }
	}


    /*
     * Class used when setting the timer for the initial
     * synchronization message.
     */
    private class QueryTimer { }

    /**
     * Sets the callback (which is a StreamManager) for stream messages
     */
    protected static synchronized void setFSMCallback (final short fsmId, final I_ReceiveCallback callback)
    {
        /*
         * Validate the callback
         */
        switch (fsmId)
        {
            case FSM_BEST_EFFORT_ORDERING:
            {
                if (! (callback instanceof StreamManager))
                    throw new HyperCastFatalRuntimeException ("Stream callback must be a StreamManager");
            }
            break;

            case FSM_H2H_ACK: /* fall-through */
            case FSM_E2E_ACK: /* fall-through */
            case FSM_DUPLICATE_ELIMINATION: /* fall-through */
            case FSM_SYNCHRONIZATION: /* fall-through */
            case FSM_INCAST: /* fall-through */
            case FSM_NAMING: /* fall-through */
            {
                /* do nothing */
            }
            break;

            case FSM_NO_SERVICE:
                throw new HyperCastWarningRuntimeException ("Invalid service specified");

            default:
                //assert false : "unreachable code";
        }

        fsmCallbacks.put (new Short (fsmId), callback);
    }

    /**
     * Start the timer thread in the message store
     * The code here are copied from adpter class
     */
    public void start()
    {
        synchronized (stateGuard) 
        {
            while (true) 
            {
                if (state == STOPPED) 
                {
                    state = STARTING;
                    break;
                }
                else if (state == STARTED) 
                {
                    throw new IllegalStateException("Tried to start message store twice!");
                }
                // TODO: else other states?

                try
                {
                    stateGuard.wait();
                }
                catch (InterruptedException ignored) 
                {
                }
            }
        }
        messageProcessor = new Message_Processor (this);
        new Thread (messageProcessor, "Message Store Message Processing Thread").start ();

        /*
         * If synchronization is required the transmission of the
         * QUERY_SYNC_ALL message must be scheduled.
         */

        if (synchronization)
        {
            socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new QueryTimer()),initialSynchronizatinDelay); 
        }

        synchronized (stateGuard) 
        {
            state = STARTED;
            stateGuard.notify();
        }
    }

    /**Stop the thread
     */
    public void stop(){
        synchronized (stateGuard) 
        {
            while (true) 
            {
                if (state == STARTED) 
                {
                    state = STOPPING;
                    break;
                }
                else if (state == STOPPED) 
                {
                    throw new IllegalStateException("Tried to stop message store twice!");
                }

                try {
                    stateGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }

        myStop();

        synchronized (stateGuard) 
        {
            state = STOPPED;
            stateGuard.notify();
        }
    }

    /**  It is called by Stop(). */

    protected void myStop() 
    {
        messageProcessor.die();
        synchronized (deathGuard)
        {
            while (! messageProcessor.isDead()) 
            {
                try 
                {
                    deathGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }
    }
    
    /**
     * @return an service extension for individual message
     */
    protected FSM_Extension getMessageServiceExtension(short serviceNumber)
    {
		/*
		  * A message id is the concatenation of a four byte hash of a
		  * unique identifier of this node and a four byte counter. The
		  * counter is unsigned and wraps to zero after reaching its
		  * maximum value.
		  */



		/*
		 * Useing a random number to replace a hash value as the part of the
		 * the message ID makes more sense, because it can be printed in 
		 * the application.
		 */
		 short randomNum = (short) Math.abs((short) (new Random()).nextInt());
		 	
		 byte[] idBytes = ByteArrayUtility.toByteArray (randomNum); //2 bytes
		 

		 byte [] adfpayload = new byte [FSM_Extension.EXTENSION_HEADER_LENGTH];

		 /*
		  * This index will track the current insertion position as the
		  * payload extension array is populated.
		  */
		 int payloadIndex = 0;

	 	 /*
		  * The next two bytes encode a finite state machine id, also known as
		  * "service number" by the application.
		  */
		 adfpayload [payloadIndex++] = (byte)((serviceNumber >>> 8) & 0xFF);
		 adfpayload [payloadIndex++] = (byte)((serviceNumber >>> 0) & 0xFF);


		/*
		 * The next byte encodes a type. TODO: it is unclear why this is here,
		 * there seems to be only a single constant defined for this field so
		 * it is essentially hard-coded? What other values can this field have?
		 * What do those values mean? Why are those values not defined as
		 * constants as well?
		 */
		adfpayload [payloadIndex++] = FSM_Extension.EXTENSION_TYPE_PAYLOAD;
		 /*
		  * END of message store extension header.
		  */

		 /*
		  * Add the value of the hash to the payload array.
		  */
		 //System.arraycopy (hash, 0, adfpayload, payloadIndex, FSM_Extension.ID_HASH_LENGTH);
		 System.arraycopy (idBytes, 0, adfpayload, payloadIndex, FSM_Extension.ID_HASH_LENGTH);
		 payloadIndex += FSM_Extension.ID_HASH_LENGTH;

		 /*
		  * Add the value of the counter to the payload array.
		  */
		 adfpayload [payloadIndex++] = (byte)((counter >>>  8) & 0xFF);
		 adfpayload [payloadIndex++] = (byte)((counter >>>  0) & 0xFF);



		 counter = (counter == COUNTER_MAX_VALUE) ? 0 : ++counter;
		 
		 return new FSM_Extension(adfpayload);
    }

    /**
     * @return a service extension for a stream message.
     */
    protected FSM_Extension getStreamServiceExtension (byte [] streamid)
    {
        if (streamid == null
            ||
            streamid.length != 4)
        {
            throw new HyperCastWarningRuntimeException ("Stream ID must be 4 bytes in length: " + streamid);
        }

        // TODO: do a streamid lookup to see if this stream has been created?
// does that lookup return a stream type? if so, use it to set this
// variable, that way we wouldn't need to hard code fsm specific stuff
// here. would have to put a "dummy" value in the lookup table for
// things like naming

        final short fsmid;

        long sequencenumber = 0;
        Long sequencenumberLong = (Long) streamsequence.get (streamid);
        if (sequencenumberLong != null)
        {
            sequencenumber = sequencenumberLong.longValue ();
        }

        // TODO(NAMING): this is FSM specific, shouldn't be in
        // message store, don't know where else it should go...

        // TODO: Is there a better  way in Java to do array compare?
        if (streamid[0] == FSM_Extension.NAMING_STREAM_ID[0]
            &&
            streamid[1] == FSM_Extension.NAMING_STREAM_ID[1]
            &&
            streamid[2] == FSM_Extension.NAMING_STREAM_ID[2]
            &&
            streamid[3] == FSM_Extension.NAMING_STREAM_ID[3])
        {
            // NAMING
                
            // use 0 as a sequence number since we don't need a sequence number
            sequencenumber = 0;
                
            fsmid = MessageStore.FSM_NAMING;
        }
        else
        {
            fsmid = MessageStore.FSM_BEST_EFFORT_ORDERING;
        }

        // TODO: why not create an fsm here if streamid was not known?

        byte [] streampayload = new byte [FSM_Extension.EXTENSION_HEADER_LENGTH +
                                          FSM_Extension.SEQUENCE_NUMBER_LENGTH];

        /*
         * This index will track the current insertion position as the
         * payload extension array is populated.
         */
        int payloadIndex = 0;

        /*
         * Populate the fields of the message store extension header
         */
        streampayload [payloadIndex++] = (byte)((fsmid >>> 8) & 0xFF);
        streampayload [payloadIndex++] = (byte)((fsmid >>> 0) & 0xFF);
        streampayload [payloadIndex++] = FSM_Extension.EXTENSION_TYPE_PAYLOAD;

        System.arraycopy (streamid, 0, streampayload, payloadIndex, streamid.length);
        payloadIndex += streamid.length;

        System.arraycopy (ByteArrayUtility.toByteArray (sequencenumber), 0, streampayload, payloadIndex, FSM_Extension.SEQUENCE_NUMBER_LENGTH);

        sequencenumber = (sequencenumber == Long.MAX_VALUE) ? 0 : ++sequencenumber;
        streamsequence.put (streamid, new Long (sequencenumber + 1));

        return new FSM_Extension (streampayload);
    }

    /**
     * Test if a service is supported by the MessageStore
     * @param  serviceNumber service number     
     * @return true/false
     * 
     */
    public static boolean isValidServiceNumber(short serviceNumber)
    {
        return (serviceNumber == FSM_H2H_ACK
                ||
                serviceNumber == FSM_E2E_ACK
                ||
                serviceNumber == FSM_SYNCHRONIZATION
                ||
                serviceNumber == FSM_DUPLICATE_ELIMINATION
                ||
                serviceNumber == FSM_INCAST
                ||
                serviceNumber == FSM_BEST_EFFORT_ORDERING
                ||
                serviceNumber == FSM_NAMING);
    }

	/**
	 * Returns the message type given a service identifier.
	 * @param  serviceNumber service number     
	 * @return message type
	 * 
	 */
	public static short getMsgType(short serviceNumber)
	{
		if (serviceNumber == FSM_H2H_ACK
				||
				serviceNumber == FSM_E2E_ACK
				||
				serviceNumber == FSM_SYNCHRONIZATION
				||
				serviceNumber == FSM_DUPLICATE_ELIMINATION
				||
				serviceNumber == FSM_INCAST)
		{
			return MESSAGESTORE_MESSAGE;
		}
		else if (serviceNumber == FSM_BEST_EFFORT_ORDERING
					||
					serviceNumber == FSM_NAMING)
		{
			return STREAM_MESSAGE;
		}
		else
		{
			return OTHER_MESSAGE;
		}
	}

    /**
     * Factory method that creates finite state machines based on the identifier given.
     * @param finiteStateMachineId A number that identified a type of finite state machine
     * @return a finite state machine object
     */
    private I_MessageStoreFSM createFSM (short finiteStateMachineId)
    {
        switch (finiteStateMachineId)
        {
            case FSM_H2H_ACK:
                return new MessageStoreFSM_H2HACK (socket, this, config);
            case FSM_E2E_ACK:
                return new MessageStoreFSM_E2EACK (socket, this, config);
            case FSM_DUPLICATE_ELIMINATION:
                return new MessageStoreFSM_DuplicateElimination (socket, this, config);
            case FSM_SYNCHRONIZATION:
                return new MessageStoreFSM_Sync (socket, this, config);
            case FSM_INCAST:
                return new MessageStoreFSM_Incast (socket, this, config);
            case FSM_BEST_EFFORT_ORDERING:
                return new MessageStoreFSM_Ordering (socket, this, config);
            case FSM_NAMING:
            {
                if (namingServiceStarted)
                {
                    if (naming == null)                    
                    {
                        naming = new MessageStoreFSM_Naming (socket, this, config);
                        statsPro.addStatsElement ("Naming", naming, 1, 1);
                    }
                    return naming;
                }
                else
                    return null;
            }
            default:
                throw new HyperCastFatalRuntimeException ("Unsupported service");
        }
    }

    private MessageStoreFSM_Naming naming = null;
    private boolean namingServiceStarted = false;

    /**
     * Sets the notification handler that is used to pass events from
     * the message store to an application.
     */
    public void setNotificationHandler (NotificationHandler notificationHandler)
    {
        this.notificationHandler = notificationHandler;
    }

    /**
     * Increments the retransmission count; used by the individual
     * finite state machines.
     */
    protected void incrementRetransmissionCount ()
    {
        ++numberOfRetransmissions;
    }

    /**
     * This method is called from the OL_Socket whenever a new message
     * is sent, i.e. when one of the "send*" methods is called. Since
     * the send methods are part of the HyperCast API, this method is
     * called by an application thread.
     *
     * @param appmsg The OL_Message that is scheduled for sending.
     */
    protected /* TODO: synchronized? */ void sendMessage (OL_Message appmsg)
    {
        /*
         * The message is stored at the buffer to allow
         * the message processor thread to process it.
         */
        //buffer.nonblockingWrite (appmsg);
		processMessage (appmsg);
/*
		if (debug)
			System.out.println("MessageStore:sendMessage: block writing message into buffer.");
*/

		try 
		{
			buffer.blockingWrite (appmsg);
		}
		catch (InterruptedException ignored)
		{
		}

    }

    /**
     * Accepts a message having a finite state machine extension into
     * the message store. The message must have a finite state machine
     * extension. The message is either a stream or message based
     * message; it is either application data or control information
     * (out-of-band/meta information).
     *
     * Called from the forwarding engine when a finite state machine
     * extension is detected (the socket adapter thread). Also called
     * by the naming API to inject "dummy" messages into the
     * message store (The application thread).
     *
     * @param fsmMessage Message assumed to have a finite state machine extension
     * 
     * @return <code>true</code> if the message store will assume responsibility
     * for forwarding this message. When <code>false</code> the responsibility of forwarding
     * is on the caller (i.e. forwarding engine).
     */
    protected synchronized boolean receiveMessage (final OL_Message fsmMessage)
    {
        //if (debug)
        //	System.err.println ("MessageStore.receiveMessage() invoked with this message: " + fsmMessage);

        /*
         * Validate message
         */
        final FSM_Extension fsmExtension = (FSM_Extension) fsmMessage.getFirstExtensionByType (Extension.FSM);
        if (fsmExtension == null)
            throw new HyperCastWarningRuntimeException ("Message sent to the message store without FSM extension: " + fsmMessage);

        /*
         * First, the message is stored at the receive buffer to allow
         * the message processor thread to process it, i.e. create a
         * new finite state machine or update the existing finite
         * state machine.
         */

        if (debug)
            System.out.println("MessageStore:receiveMessage: block writing message into buffer.");

        try { buffer.blockingWrite (fsmMessage); }
        catch (InterruptedException ignored) { }

		if (debug)
			System.out.println("MessageStore:receiveMessage: serviceNumber is "+fsmMessage.getServiceIdentifier());

        /*
         * Then the message is checked to decide if the messagestore
         * should forward this message. This check is done based on
         * the type of finite state machine that is associated with
         * the message.
         */
        final short serviceNumber = fsmMessage.getServiceIdentifier();
        final FiniteStateMachineMetaOperations metaOps = getMetaOps (serviceNumber);
        return metaOps.messageStoreWillForwardMessage (fsmMessage);
    }

    private FiniteStateMachineMetaOperations getMetaOps (short serviceNumber)
    {
        Short serviceNumberShort = new Short (serviceNumber);
        FiniteStateMachineMetaOperations metaOps = (FiniteStateMachineMetaOperations) FSMMetaOpsTable.get (serviceNumberShort);
        if (metaOps == null)
            throw new HyperCastFatalRuntimeException ("No meta operations object registered for service number " + serviceNumber);
        return metaOps;
    }

    /**
     * This is the method used to process a message that is sent or received by the messagestore.
     * @param appmsg OL_Message     
     */
    protected void processMessage (OL_Message appmsg)
    {

		if (debug)
			System.out.println("MessageStore:processMessage: Received message from "+((OL_Message) appmsg).getSourceAddress().toString());

        FSM_Extension fsmExtension = (FSM_Extension)appmsg.getFirstExtensionByType(Extension.FSM);

        /*
         * Validate that there is information carried by the FSM
         * extension.
         */
        if (fsmExtension == null)
            throw new HyperCastWarningRuntimeException ("No FSM Extension");

        /*
         * Validate that the message contains the right combination of
         * service number and delivery mode.
         */
        short serviceNumber = appmsg.getServiceIdentifier();
        FiniteStateMachineMetaOperations metaOps = getMetaOps (serviceNumber);

		/*
		 * The condition for different FSM message to be Valid is not the same. See 
		 * class AcknowledgmentMetaOperations.
		 */
        if (! metaOps.isValidMessage (appmsg)) 
            throw new HyperCastWarningRuntimeException ("Message using this service number " + appmsg.getServiceIdentifier() + " can not be delivered with delivery mode " + appmsg.getDeliveryMode());

		/*
		  * For unicast message, it is the messagestore's
		  * responsibility to check if the message has arrived at the
		  * destination and if the messagestore should process it.
		  */
		 I_LogicalAddress destinationLogicalAddress = appmsg.getDestinationAddress();
		 boolean messageHasReachedDestination = socket.getLogicalAddress().equals (destinationLogicalAddress);

		 boolean isSource = socket.getLogicalAddress().equals (appmsg.getSourceAddress());

		 if ((appmsg.getDeliveryMode() == OL_Message.DELIVERY_MODE_UNICAST)
			 &&
			 (! isSource)
			 &&
			 (! messageHasReachedDestination))
		 {
			if (debug)
				System.out.println("metaOps.processIntermediateUnicastMessage (appmsg) is "+  metaOps.processIntermediateUnicastMessage (appmsg));

			if (! metaOps.processIntermediateUnicastMessage (appmsg)) 
	 			return;
		 }

        byte type = fsmExtension.getType();
        short fsmid = fsmExtension.getFSMID();
        byte [] uniqueId = fsmExtension.getUniqueID();

        /*
         * process Query Messages for synchronization
         * The query message is treated differently than other messages because the 
         * processing of the query message need the information of all available messages.
         * Therefore, it is processed at the messagestore, not in individual finite state machine.
         */ 
        if (type == MessageStoreFSM_Sync.QUERY_SYNC_ALL)
        {
            MessageStoreFSM_Sync.receiveQuery (store, socket, appmsg);
        }
        else
        {
        	if (debug)
				System.out.println("uniqueId=" + HTTP_ServerUtility.toHexString(uniqueId)+".");

            /*
             * Get a finite state machine for this message, creating a
             * new FSM if necessary. Pass the message to the finite
             * state machine and update statistics.
             */
            I_MessageStoreFSM finiteStateMachine = get (uniqueId);
            boolean isNewMessage = (finiteStateMachine == null);

			if (debug)
				System.out.println("MessageStore: processMessage: is it a new message? "+ isNewMessage);

            if (isNewMessage)
            {
				if (debug)
					System.out.println("MessageStore: processMessage: new message, create FSM.");

				//finiteStateMachine = createFSM (fsmid);

				finiteStateMachine = fsmPool.getFSMInstance(fsmid);
				if (finiteStateMachine == null)
				{
					if (debug)
						System.out.println("MessageStore.ProcessMessage: no predefined instance, create a new one.");
					finiteStateMachine = createFSM (fsmid);

                                        if (finiteStateMachine == null)
                                            return; // service not started
				}

                put (uniqueId, finiteStateMachine);
            }

			if (debug)
				System.out.println("MessageStore: processMessage: deliveryMode="+appmsg.getDeliveryMode());

            if (appmsg.getFirstExtensionByType (Extension.RAW_DATA) != null)
            {
				if (debug)
					System.out.println("MessageStore: processMessage: pass message (having payload) to FSM instance.");

                finiteStateMachine.receiveDataMessage (appmsg, isNewMessage);
                //dataBytesReceived += appmsg.getFirstExtensionByType (Extension.RAW_DATA).getADUPayloadLength();
				dataBytesReceived += appmsg.getFirstExtensionByType (Extension.RAW_DATA).getSize();
                ++dataMessagesReceived;
            }
            else 
            {
				if (debug)
					System.out.println("MessageStore: processMessage: call finiteStateMachine.receiveControlMessage.");

                finiteStateMachine.receiveControlMessage (appmsg, isNewMessage);
                ++controlMessagesReceived;
            }
        }
    }

    /**
     * Store the message in the hashtable
     * @param messageid UniqueID
     * @param fsm I_MessageStoreFSM     
     */
    public synchronized void put (byte[] messageid, I_MessageStoreFSM fsm)
    {
        store.put (new String (messageid), fsm);
    }

    /**
     * Get the message in the hashtable
     * @param messageid UniqueID
     * @return I_MessageStoreFSM     
     */
    public synchronized I_MessageStoreFSM get (byte[] messageid)
    {
        return (I_MessageStoreFSM)store.get(new String(messageid));
    }

    /**
     * Remove the message in the hashtable
     * @param   messageid UniqueID
     */
    public synchronized void remove (byte[] messageid)
    {
        store.remove(new String(messageid));
    }

    /**
     * Set  the timer 
     * @param fsm I_MessageStoreFSM     
     * @param index timestampID
     * @param delay_ms time of delay
     */
    public void setTimer (I_MessageStoreFSM fsm, int index, long delay_ms)
    {
        socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new Timer_ID(fsm,index)),delay_ms);
    }

    /**
     * Set  the timer 
     * @param fsm I_MessageStoreFSM     
     * @param obj Object that will be returned when the timer expires
     * @param delay_ms time of delay
     */
    public void setTimer (I_MessageStoreFSM fsm, Object obj, long delay_ms)
    {
        socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new Timer_ID(fsm,obj)),delay_ms);
    }

    /**
     * Clear  the timer 
     * @param   fsm I_MessageStoreFSM     
     * @param   index timestampID
     */
    public void clearTimer (I_MessageStoreFSM fsm, int index)
    {
        socket.socketAdapter.clearTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new Timer_ID(fsm,index)));
    }

    /**
     * Clear  the timer 
     * @param   fsm I_MessageStoreFSM     
     * @param   obj Object attached to the timer
     */
    public void clearTimer (I_MessageStoreFSM fsm, Object obj)
    {
        socket.socketAdapter.clearTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new Timer_ID(fsm,obj)));
    }

    /**
     * Get  the timer 
     * @param fsm I_MessageStoreFSM     
     * @param index timestampID
     * returns the delay
     */
    public long getTimer (I_MessageStoreFSM fsm, int index)
    {
        return socket.socketAdapter.getTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore,new Timer_ID(fsm,index)));
    }

    /** Sync the message store periodically
     * @param  timestamp integer
     */
    public void timerExpired (Object timestamp)
    {
        /*
         * The timer id used in all finite state machines are Timer_ID object.
         * To distinguish, the messagestore itself uses integer as timer id.
         * The timer expired event is demultiplexed here based on the timer id
         */
        if (timestamp instanceof QueryTimer)
        {
            try
            {
                startQueryTime = System.currentTimeMillis();
                MessageStoreFSM_Sync.sendQuery (store, socket);
            }
            catch (Exception ignored) {}
            /*
             * set the timer periodically synchronize
             */
            socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfMessageStore, new QueryTimer()),
                                          MessageStoreFSM_Sync.synchronizationPeriod);
        }
        else
        {
            Timer_ID id = (Timer_ID) timestamp;
            try 
            {
                I_MessageStoreFSM fsm = id.getFSM();
                Object obj = id.getObject();
                int stamp = id.getID();

				//if (debug)
				//	System.out.println("MessageStore:timerExpired: fsm is null?"+(fsm==null)+", store.contains(fsm)?"+store.contains(fsm));

                if (fsm!=null && store.contains(fsm))
                { 
					//if (debug)
					//	System.out.println("MessageStore:timerExpired: call fsm.timerExpired(stamp)");

                if (obj != null)
                    fsm.timerExpired(obj);
                else
                    fsm.timerExpired(stamp);
                }
            }
            catch (Exception e) 
            {   
                System.out.println("Encountered exception \""+e.getMessage() + "\" in timerExpired()");
            }
        }
    }


	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
        statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, false/* writable flag */);
		
		statsPro.addStatsElement("HashtableCurrentSize"/* name */, new HashtableSize()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("SendQueryTime", new sendqueryTimePeriod()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("FSM-Data", new FSMData()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("FSM-Data-Bytes", new FSMDataBytes()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("FSM-Control", new FSMControl()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("NumberOfRetransmissions", new RetransmissionTimes()/* I_Stats object */, 1, 1);
	}


	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}


	 /** @see I_Stats#getStats */
	 public Element[] getStats (Document doc, XPath xpath) throws HyperCastStatsException
	 {
		 //System.out.println("MessageStore:getStats: xpath string is "+xpath.getPatternString());
		 //System.out.println("MessageStore:getStats: call statsPro.getStatsResult(doc, xpath)");
		 return statsPro.getStatsResult(doc, xpath);
	 }
	
	 /** @see I_Stats#setStats */
	 public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	 {
		 //System.out.println("MessageStore:setStats: xpath string is "+xpath.getPatternString());
		 //System.out.println("MessageStore:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		 return statsPro.setStatsResult(doc, xpath, newValue);
	 }
	
	 /** @see I_Stats#getReadSchema */
	 public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	 {
		 //System.out.println("MessageStore:getReadSchema: xpath string is "+xpath.getPatternString());
		 //System.out.println("MessageStore:getReadSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		 return statsPro.getReadSchemaResult(doc, xpath);
	 }
	
	 /** 
	  * Return the schema element which represents the root of the sub-tree, specified
	  * by the given xpath, in write schema tree. 
	  * 
	  * @param doc   Document used for create new elements or nodes.
	  * @param xpath XPath instance representing the statistics which is the root
	  * 	of the sub-tree to be queried.
	  * @see I_Stats#getWriteSchema(Document, XPath)
	  */
	 public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	 {
		 return statsPro.getWriteSchemaResult(doc, xpath);
	 }
	 

	/**
	 * This class manages simple statistic "HashtableCurrentSize".
	 */
	class HashtableSize extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+store.size());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "SendQueryTime".
	 */
	class sendqueryTimePeriod extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+startQueryTime);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "FSM-Data".
	 */
	class FSMData extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+dataMessagesReceived);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "FSM-Data-Bytes".
	 */
	class FSMDataBytes extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+dataBytesReceived);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "FSM-Control".
	 */
	class FSMControl extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String("" + controlMessagesReceived);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "NumberOfRetransmissions".
	 */
	class RetransmissionTimes extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String("" + numberOfRetransmissions);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

}



class Message_Processor implements Runnable
{
    /**
     * The MessageStore for which this processor will do work.
     */
    private MessageStore messageStore;

    /**
     * The thread in which this processor runs.
     */
    private Thread processorThread;

    /**
     * True if the function die() has been called.
     */
    private boolean dying;

    /**
     * True if <code>processorThread</code> has returned from run().
     */
    private boolean died;

    /**
     * True while <code>processorThread</code> is waiting to receive.
     */
    private boolean waiting;

    /**
     * Guard used to protect <code>waiting</code> when it is changed.
     */
    private Object waitingGuard;	
	
    /**
     * Contructs the thread with a MessageStore object. 
     */
    Message_Processor (MessageStore messageStore)
    {
        this.messageStore = messageStore;
		
        this.dying = false;	
        this.died = false;	

        this.waiting = false;
        this.waitingGuard = new Object();
    }
	
    private int message_count = 0;

    public void run() 
    {	    		
        processorThread = Thread.currentThread();
        I_Message i_message = null;
		
        while (true)
        {
            try 
            {
                try 
                {
                    try 
                    {
                        synchronized (waitingGuard) 
                        {
                            waiting = true;
                        }
                        if (dying) break;
                        i_message = messageStore.buffer.read(); // may block
//                        messageStore.config.log.println ("Message Store read message  " + (++message_count));
                    }
                    finally 
                    {
                        synchronized (waitingGuard) 
                        { 
                            waiting = false;
                        }
                    }
                }
                finally
                {
                    // what purpose does this serve?
                    if (Thread.interrupted()) 
                    {
                        Thread.sleep(1000);
                    }
                }   

                if (dying) break;

                try 
                {
                    OL_Message ol_message = (OL_Message) i_message;
                    messageStore.processMessage (ol_message);
                }
                catch (Exception e) 
                {   
                    messageStore.config.log.println ("Encountered exception while processing message " + e);
                    e.printStackTrace();
                }
            }          
            catch (final InterruptedException ie)
            {
                // This message is elided here because it means that
                // the read was interrupted, maybe due to shutdown,
                // this is normal. What interruptions can there be
                // other than shutdown?

                //messageStore.config.log.println ("Encountered Interruption exception while waiting to read message: " + ie);
            }						
          
        }	    
	
        synchronized (messageStore.deathGuard) 
        {
            died = true;
            messageStore.deathGuard.notify();
        }
    } 
	
    /** This function causes <code>processorThread</code> to die eventually. */
    public void die()
    {
        synchronized (waitingGuard)
        {
            dying = true;
            if (waiting) processorThread.interrupt();  // in the case that it is blocking on the wait call.
        }
    }
	
    /** Returns true if <code>processorThread</code> will never call an I_AdapterCallback method again. */
    public boolean isDead() 
    {
        return (died || waiting);
    }
}
