/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import java.util.Vector;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.HyperCastStatsException;
import hypercast.I_LogicalAddress;
import hypercast.I_OverlayMessage;
import hypercast.I_OverlaySocket;
import hypercast.HyperCastConfig;
import hypercast.HyperCastWarningRuntimeException;
import hypercast.I_Stats;
import hypercast.StatsElement;
import hypercast.StatsProcessor;
import hypercast.util.XmlUtil;
import hypercast.util.XmlUtil;
import java.text.ParseException;

/**
 * Receives, parses, and handles MC messages. Internal class, not exposed to applications.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
/* package */ class MCSocket
    implements Runnable
{
    public static final boolean debug = false;
	
    /**
     * <code>true</code> if this socket is open. Calls to sendTo and
     * sendToAll succeed only when the socket is open.
     */
    private boolean open = false;

    protected HyperCastConfig config;
	
    private I_OverlaySocket olSocket;

    /** The transcript file output stream */
    private OutputStream transcript;
    
    private boolean logSend;
    private boolean logReceive;
    private boolean append;

    /**
     * Property name used to specify a filename that will contain XML messages sent and received over the <code>MCSocket</code>.
     */
    private static final String PROP_NAME_TRANSCRIPT_SENDS    = "Transcript/RecordOutgoing";
    private static final String PROP_NAME_TRANSCRIPT_RECEIVES = "Transcript/RecordIncoming";
    private static final String PROP_NAME_TRANSCRIPT_APPEND   = "Transcript/Append";
	
    /** the callback handling the incoming message */
    private I_MCRaw_Callback callback;
	
    /**
     * Starts a message receiver thread.
     *
     * @param propertiesPrefix Currently ignored because there are no configuration options for MCSocket.
     */
    public MCSocket (final HyperCastConfig config, final String propertiesPrefix, final I_MCRaw_Callback callback)
    {
        if (config == null)
            throw new NullPointerException();
		
        this.config = config;
        this.olSocket = config.createOverlaySocket (null);
        this.callback = callback;
        
        this.logSend    = config.getBooleanAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_TRANSCRIPT_SENDS));
        this.logReceive = config.getBooleanAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_TRANSCRIPT_RECEIVES));
        this.append     = config.getBooleanAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_TRANSCRIPT_APPEND));

        new Thread (this, "MCSocket").start();
    }

    public I_OverlaySocket getOverlaySocket()
    {
        return olSocket;
    }

    /**
     * Creates an advertise message
     *
     * @see #createQueryMessage(int)
     */
    public MCAdvertiseMessage createAdvertiseMessage (final int messageType)
    {
        switch (messageType)
        {
            case MCRawMessage.ADVERTISE_MONITOR:
            case MCRawMessage.ADVERTISE_PORTAL:
                break;

            default:
                throw new IllegalArgumentException (MCRawMessage.intToTypeString [messageType] + " is not a type of advertise message");
        }

        final MCAdvertiseMessage message = new MCAdvertiseMessage (messageType, getLogicalAddress());

        return message;
    }

    /**
     * Create an data query message.
     *
     * The source is set here, presumably, because the source logical
     * address can change between the time that a message is created
     * and the time when a message is sent. There could be situations
     * when a caller wants to know the logical address at the time of
     * message creation?
     *
     * @param messageType A query message type
     */
    public MonMessage createQueryMessage (final int messageType)
    {
        switch (messageType)
        {
            case MCRawMessage.GET_QUERY:
            case MCRawMessage.READ_SCHEMA_QUERY:
            case MCRawMessage.SET_QUERY:
            case MCRawMessage.WRITE_SCHEMA_QUERY:
                break;

            default:
                throw new IllegalArgumentException (MCRawMessage.intToTypeString [messageType] + " is not a type of query message");
        }

        final MonMessage message = new MonMessage (messageType, getLogicalAddress());

        return message;
    }
	
    /**
     * Send message to one destination; non-blocking.
     *
     * @param message Message that is sent to a member of the overlay.
     *
     * @throws NullPointerException if message or destination is null
     * @throws HyperCastWarningRuntimeException if the socket is not open
     */
    public void sendTo (final MCRawMessage message, final I_LogicalAddress destination)
    {
        if (message == null)
            throw new NullPointerException ("message cannot be null");

        if (destination == null)
            throw new NullPointerException ("destination logical address cannot be null");

        if (! open)
            throw new HyperCastWarningRuntimeException ("Attempt to send message on unopened Monitor and Control Socket.");

        message.setSender (getLogicalAddress());
        message.setDestination (destination);
        message.setTimeStamp (System.currentTimeMillis());

        if (logSend)
            updateTranscript (message);

        final I_OverlayMessage olMessage = olSocket.createMessage (toByteArray (message.getDocument()));
        olSocket.sendToNode (olMessage, destination);
    }

    /**
     * Send message to all destinations by multicast; non-blocking.
     *
     * @param message Message that is sent to all members of the overlay.
     *
     * @throws NullPointerException if message is null
     * @throws HyperCastWarningRuntimeException if the socket is not open
     */
    public void sendToAll (final MCRawMessage message)
    {
        if (message == null)
            throw new NullPointerException ("message cannot be null");

        if (! open)
            throw new HyperCastWarningRuntimeException ("Attempt to send message on unopened Monitor and Control Socket.");

        message.setSender (getLogicalAddress());
        message.setDestination (MCRawMessage.MULTICAST_DESTINATION);
        message.setTimeStamp (System.currentTimeMillis());

        if (logSend)
            updateTranscript (message);

        final I_OverlayMessage olMessage = olSocket.createMessage (toByteArray (message.getDocument()));
        olSocket.sendToAll (olMessage);
    }

    /** 
     * @see I_Portal#startTranscript
     */
    public synchronized void startTranscript (final String filename)
        throws IllegalStateException, FileNotFoundException, HyperCastWarningRuntimeException
    {
        if (! logSend
            &&
            ! logReceive)
        {
            throw new HyperCastWarningRuntimeException ("Neither sending nor receiving chosen for logging in MCSocket!");
        }

        if (transcript != null)
        {
            throw new IllegalStateException ("Cannot start transcript while another is already open; must call stopTranscript first.");
        }

        transcript = new FileOutputStream (filename, append);

        try
        {
            final String xmlHeader = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<Transcript>\n";
            final byte [] headerBytes = xmlHeader.getBytes();
            transcript.write (headerBytes);
        }
        catch (final IOException ioe)
        {
            try
            {
                transcript.close();                
                throw new HyperCastWarningRuntimeException ("Could not write header to transcript file: " + ioe);
            }
            catch (final IOException ioe2)
            {
                throw new HyperCastWarningRuntimeException ("Could not write header to transcript file or close file: " + ioe + "\n" + ioe2);
            }
            finally
            {
                transcript = null;
            }
        }

        System.out.println ("Opened transcript file \"" + filename + "\" send(" + logSend + ") receive(" + logReceive + ") append(" + append + ")");
    }

    /**
     * Stops transcripting 
     */
    public synchronized void stopTranscript()
        throws IllegalStateException, HyperCastWarningRuntimeException
    {
        if (transcript == null)
        {
            throw new IllegalStateException ("Cannot stop transcript if none is active; must call startTranscript first.");
        }

        String cannotWriteString = "";
        try
        {
            final String xmlTrailer = "</Transcript>\n";
            final byte [] trailerBytes = xmlTrailer.getBytes();
            transcript.write (trailerBytes);
        }
        catch (final IOException ioeWriting)
        {
            cannotWriteString = "Could not write trailer to transcript file; file closed without trailer: " + ioeWriting;
            throw new HyperCastWarningRuntimeException (cannotWriteString);
        }
        finally
        {
            try
            {
                transcript.close();
            }
            catch (final IOException ioeClosing)
            {
                throw new HyperCastWarningRuntimeException ("Could not write trailer or close transcript file: " +
                                                            cannotWriteString + "\n" + ioeClosing);
            }
            finally
            {
                transcript = null;
            }
        }
    }

    protected synchronized void updateTranscript (final MCRawMessage message)
    {
        if (transcript != null)
        {
            try
            {
            	Document doc = message.getDocument();
            	/* remove xml tag ?*/
            	XmlUtil.writeXml (doc, transcript, true);
                transcript.flush();
            }
            catch (final IOException ioe)
            {
                config.err.println ("Error writing transcript file: " + ioe);
            }
        }
    }

    /** Returns the Logical Address. */
    public I_LogicalAddress getLogicalAddress()
    {
        return olSocket.getLogicalAddress();
    }

    public synchronized void open()
    {
        olSocket.joinOverlay();
        open = true;
    }

    public synchronized void close()
    {
        open = false;
        olSocket.leaveOverlay();
    }

    public boolean isOpen()
    {
        return open;
    }

    /**
     * parse a LogicalAddress from string
     */
    public I_LogicalAddress createLogicalAddress (final String logicalAddressString)
    {
        return olSocket.createLogicalAddress (logicalAddressString);
    }

    public void run()
    {
        final String threadName = Thread.currentThread().getName();
        while (true)
        {
            try
            {
                final I_OverlayMessage olMessage;
                final Document document;
			
                try
                {
                    olMessage = olSocket.receive();  // blocks
                    final byte [] payload = olMessage.getPayload();
                    document = XmlUtil.createDocument (payload);
                }
                catch (final ParseException pe)
                {
                    config.err.println ("Parse exception: could not parse payload of overlay message, continuing: " + pe);
                    continue;
                }
                catch (final HyperCastWarningRuntimeException hcwre)
                {
                    config.err.println (threadName + ": interrupted in receive, continuing: " + hcwre);
                    continue;
                }

                final MCRawMessage message = MCRawMessage.restoreMessage (this, document);

                if (logReceive)
                    updateTranscript (message);
            
                /* send to the receiving function */
                if (callback != null)
                    callback.receiveMCRawMessage (message);            
            }
            catch (final Exception e) // Catch all exceptions here - we do not want this thread to stop.
            {
            	e.printStackTrace();
                System.err.println (Thread.currentThread().getName() + ": Caught exception, continuing with MCSocket: " + e);
            }
        }
    }

    /**
     * Converts an XML document into a byte array.    
     * @throws HyperCastWarningRuntimeException if any problem occured while writing XML document to ByteArrayOutputStream.
     */
    protected byte [] toByteArray (final Document doc)
    {
        if (doc == null)
            throw new NullPointerException ("doc cannot be null");

        try
        {
            final ByteArrayOutputStream outstream = new ByteArrayOutputStream();
            XmlUtil.writeXml (doc, outstream);
            return outstream.toByteArray();
        }
        catch (IOException ioe)
        {
            throw new HyperCastWarningRuntimeException ("IO Exception in MCSocket.toByteArray():" + ioe.toString());
        }
    }
}

