/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import hypercast.I_Stats;
import hypercast.SimpleStats;
import hypercast.I_LogicalAddress;
import hypercast.HyperCastStatsException;
import hypercast.StatsProcessor;
import hypercast.StatsElement;
import hypercast.HyperCastConfig;

import hypercast.HyperCastWarningRuntimeException;
import hypercast.HyperCastFatalRuntimeException;
import hypercast.util.XmlUtil;

import java.io.FileNotFoundException;

import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.XPath;

import javax.xml.transform.TransformerException;

/**
 * A Portal runs an overlay socket such that the Monitor can establish
 * communication with the Portal through the overlay network.
 * 
 * An advertising protocol is run between Monitor and Portal for them to 
 * discover each other. 
 * 
 * A Portal provides the way to access an local I_Stats object from the remote
 * Monitor. Portal accepts query message from the Monitor and replies. 
 * The query message coming from a Monitor can have one of the following 
 * purposes:
 *
 * 1. Read a list of statistics values of the I_Stats object.
 * 2. Set a list of statistics values of the I_Stats object.
 * 3. Get the schema of a list of readable statistics of the I_Stats object.
 * 4. Get the schema of a list of writable statistics of the I_Stats object.
 *  
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 * @author Jianping Wang
 */
public class Portal
    implements I_Portal, I_MCRaw_Callback
{
    private final static boolean debug = false;

    /** Properties prefix */
    private static final String PORTAL_CONFIG_PREFIX = "/Public/MonitorAndControl/Portal/";
	
    /**
     * The StatsAccessor for this Portal. It receives MCMessages (in
     * XML format) and sends back a reply to a Monitor.
     */
    private StatsAccessor statsAccessor;
    
    /**
     * The PortalAdvertiser for this Portal. It periodically sends
     * AdvertisePortal messages (beacons) to Monitors.
     */
    private PortalAdvertiser portalAdvertiser;

    /**
     * A StatsProcessor instance which handles the statistics in this object.
     */
    private StatsProcessor statsProcessor;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;

	private HyperCastConfig config;

	/**
	 * The statistics name of the I_Stats instance queried by this object.
	 */
	private String clientStatsName;


	/** The I_Stats object to which queries are sent. */
	private I_Stats clientStats;
	
    /** the MC socket */
    private MCSocket mcSocket;
	
    /**
     * @param config Used to initialized various components of the Portal.
     * @param applicationStats statistics object for statistics of the Portal application
     * @param applicationStatsRootName name of the root of the application statistics, ex. "Appl" or "Socket"
     *
     *        The user of the Monitor trying to access the statistics on the Portal should write the XPath query
     *        String like: "/Stats/Appl/...".
     *
     * To access NumSockets statistic in RunServer use this XPath string: 
     *      "/Stats/Appl/NumOfSockets
     *
     * To access the logical address statistic of the second Socket statistic in RunServer use this XPath string:
     *      "/Stats/Appl/Socket[2]/Node/LogicalAddress
     *
     * To access the logical address statistic of a Portal's Socket use this XPath string:
     *      "/Stats/Portal/Socket/Node/LogicalAddress
     *
     * Assuming that the RunServer used "Appl" as the name of the root
     * of its stats. "Stats" is fixed by the implementation.
     */
    public Portal (final HyperCastConfig config, final I_Stats applicationStats, final String applicationStatsRootName)
    {
        this.config = config;
        this.mcSocket = new MCSocket (config, PORTAL_CONFIG_PREFIX, this);

 		clientStats = applicationStats;
 		
 		if(applicationStatsRootName !=null){
 			clientStatsName = applicationStatsRootName;
 			applicationStats.setStatsName(clientStatsName);
 		}else
 			clientStatsName = applicationStats.getStatsName();

		statisticsName = "Portal";
		
        this.portalAdvertiser = new PortalAdvertiser (config, mcSocket, PORTAL_CONFIG_PREFIX);
        this.statsAccessor = new StatsAccessor (this);
        
        new Thread (portalAdvertiser, "Portal Advertiser").start();

        /*
         * Get the statistics name defined in the configuration file for this I_Stats instance.
         */
        statisticsName = config.getTextAttribute (XmlUtil.createXPath (PORTAL_CONFIG_PREFIX + "StatName"));

        initStatisticsStructure();
    }
    
    /**
     * @see I_Portal#startTranscript(String)
     */
    public void startTranscript (final String filename)
        throws FileNotFoundException, IllegalStateException, HyperCastWarningRuntimeException
    {
        mcSocket.startTranscript (filename);
    }

    /**
     * @see I_Portal#stopTranscript()
     */
    public void stopTranscript()
        throws IllegalStateException, HyperCastWarningRuntimeException
    {
        mcSocket.stopTranscript();
    }

    /**
     * @see I_Portal#openPortal()
     */
    public void openPortal()
    {
        if (! mcSocket.isOpen())
        {
            mcSocket.open();
            config.log.println("Opening Portal at Logical Address: "+ mcSocket.getLogicalAddress());
            portalAdvertiser.signalOpenPortal();
        }
        else {
            config.log.println("Attempt to open a Portal that is already open.");
        }
    }

    /**
     * @see I_Portal#closePortal()
     */
    public void closePortal()
    {
        if (mcSocket.isOpen())
        {
            config.log.println("Closing Portal.");

            try
            {
                stopTranscript();
            }
            catch (final IllegalStateException ignored)
            {
                // if no transcript was started there will be an
                // exception, but that is ok: we want to ensure that
                // if a transcript was opened, it will be closed.
            }

            portalAdvertiser.signalClosePortal();
            mcSocket.close();
        }
        else {
            config.log.println("Attempt to close a Portal that is already closed.");
        }
    }
	
    /**
     * @see I_Portal#setSilentOff
     */
    public void setSilentOn()
    {
        portalAdvertiser.setSilentOn();
    }

    /**
     * @see I_Portal#setSilentOff
     */
    public void setSilentOff()
    {
        portalAdvertiser.setSilentOff();
    }

    /* (non-Javadoc)
     * @see hypercast.MonitorAndControl.I_MCRaw_Callback#receiveMCRawMessage(hypercast.MonitorAndControl.MCRawMessage)
     */
    public void receiveMCRawMessage (final MCRawMessage message)
    {
        /* demultiplexing messages */
        switch (message.getTypeNum())
        {
            case MCRawMessage.ADVERTISE_MONITOR:
                portalAdvertiser.receiveMCRawMessage (message);
                break;

            case MCRawMessage.ADVERTISE_PORTAL:
                // we hear our own advertise or from another portal: do nothing
                break;

            default:
                statsAccessor.receiveMCRawMessage (message);
        }
    }

    private void initStatisticsStructure()
    {
        statsProcessor = new StatsProcessor(this, true, true);

        statsProcessor.addStatsElement ("PortalSocket", this.mcSocket.getOverlaySocket(), 1, 1);
        statsProcessor.addStatsElement ("SilentMode", new SilentModeOperator() ,1, 1);
        statsProcessor.addStatsElement (clientStatsName, clientStats, 1, 1);
    }

	/** @see I_Stats#getStatsName */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

    /** I_Stats#getStats */
    public Element[] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }

    /** @see I_Stats#setStats */
    public Element[] setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }

    /** I_Stats#getReadSchema */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }

    /** I_Stats#getWriteSchema */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }

	/**
	 * This class manages simple statistic "SilentMode".
	 */
	class SilentModeOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return ""+ (portalAdvertiser.getSilentMode());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			final boolean mode = newValue.trim().equalsIgnoreCase ("true");
                        if (mode)
                        {
                            portalAdvertiser.setSilentOn();
                        }
                        else
                        {
                            portalAdvertiser.setSilentOff();
                        }
			return ""+ (mode);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

    /**	
     * @author Greg Mattes, January 2005
     * @author Guangyu Dong, January 2005
     * @author Jianping Wang
     * @author Michael Nahas
     */
    private class StatsAccessor
        implements I_MCRaw_Callback
    {
        /** The I_Stats object to which queries are sent. */
        private I_Stats server;
	
        /**
         * @param server
         */
        public StatsAccessor (final I_Stats server)
        {
            this.server = server;
        }

        /**
         * Reuses the message that is passed to it as a response. Simply
         * changes the type of message. Note that the message id stays the
         * same.
         *
         * The get read schema and get write schema functions are merely
         * for convenience. They return the respective schemas but are not
         * necessary to statistics proper, i.e. get and set.
         *
         * @param rawMessage The query message.
         *
         * @see hypercast.MonitorAndControl.I_ReceiveCallback#receiveMessage(hypercast.MonitorAndControl.MonMessage)
         */
        public void receiveMCRawMessage (final MCRawMessage rawMessage)
        {
            if (Portal.debug)
                System.err.println ("StatsAccessor: Received MCRawMessage: " + rawMessage.getType());
	
            final int replyType;

            switch (rawMessage.getTypeNum())
            {
                case MCRawMessage.GET_QUERY:          replyType = MCRawMessage.GET_REPLY;          break;
                case MCRawMessage.SET_QUERY:          replyType = MCRawMessage.SET_REPLY;          break;
                case MCRawMessage.READ_SCHEMA_QUERY:  replyType = MCRawMessage.READ_SCHEMA_REPLY;  break;
                case MCRawMessage.WRITE_SCHEMA_QUERY: replyType = MCRawMessage.WRITE_SCHEMA_REPLY; break;

                default:
                    throw new IllegalArgumentException ("Portal Callback Manager: Unsupported MC message type: " + rawMessage.getType());
            }

            final MonMessage message = (MonMessage) rawMessage;
		
            final int originalType = message.getTypeNum();
            message.setTypeNum (replyType);
		
            for (int i = 0; i < message.getStatsCount(); ++i)
            {
                try
                {
                    try
                    {
                        final String xpath = message.getStatXPath (i).getPatternString();
                        final XPath xpathObj = new XPath (xpath, null, null, XPath.MATCH);                
                        final Element [] value;
                        final Document doc = message.getDocument();
                
                        switch (originalType)
                        {
                            case MCRawMessage.GET_QUERY:
                                value = server.getStats (doc, xpathObj);
                                break;

                            case MCRawMessage.SET_QUERY:
                            {
                                final Element []setValue = message.getStatValue (i);
                                if(setValue.length==0){
                                	message.setStatError(i, "There is no value to set statistics: "+xpath);
                                	value = new Element [0];
                                }else
                                	value = server.setStats (doc, xpathObj, setValue[0]);
                            }
                            break;

                            case MCRawMessage.READ_SCHEMA_QUERY: 
                                value = server.getReadSchema (doc, xpathObj);
                                break;
                        
                            case MCRawMessage.WRITE_SCHEMA_QUERY:
                                value = server.getWriteSchema (doc, xpathObj);
                                break;

                            default:
                                throw new IllegalArgumentException ("Portal Callback Manager: Unsupported MC message type: " + message.getType());
                        }
                        //TODO: here value may return an array of Element instances, changed the operation on message.
                        if(value.length == 0)
                        	message.setStatError(i, "Statistics not found for: "+xpath);
                        else	
                        	message.setStatValue (i, value);
                    }
                    catch (final HyperCastStatsException hcse)
                    {
                        message.setStatError (i, hcse.toString());
                    }
                    catch (final TransformerException te)
                    {
                        message.setStatError (i, te.toString());
                    }
                }
                catch (final HyperCastStatsException hcse)
                {
                    throw new HyperCastFatalRuntimeException ("stats exception should never happen here");
                }
            }

            final I_LogicalAddress monitorAddress =  (message.getSender());
            Portal.this.mcSocket.sendTo (message, monitorAddress);
        }
    }        
}
