/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NAMING_EVENT;

import hypercast.util.XmlUtil;

import java.io.PrintWriter;

import java.util.Vector;

import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;

/**
 * <p>This class implements the APIs of the naming service. Logically,
 * this code is part of the <code>OL_Socket</code>. This class
 * implements a proxy, used by instances of <code>OL_Socket</code>,
 * for the naming APIs. This is done to reduce the complexity
 * (i.e. bloat) of the <code>OL_Socket</code> code.</p>
 *
 * <p>This class implements the API side of the naming service. The
 * other side of the naming service is the Naming FSM in the Message
 * Store.</p>
 *
 * <p>As per the HyperCast system design, there are no "conventional"
 * API calls that can be made to the Message Store. Rather, the
 * interface that is exported by the Message Store consists of passing
 * a message to the message store (input) and receiving a message from
 * the message store (output). The system is designed in this way
 * because the Message Store is designed to interact with other
 * Message Stores, not the application/programmer.</p>
 *
 * <p>The Message Store handles two types of messages application and
 * control. Application messages contain data that is being
 * transmitted to and from applications. Control, or non-application
 * payload, messages are internal to the Message Store, generated and
 * consumed by communicating Message Stores. The Naming Service needs
 * to be able to send and receive non-application payload information
 * to and from the message store. All Naming Service messages are
 * control messages.</p>
 *
 * <p>To communicate beteen the Naming Service API and the Message
 * Store a "dummy" message mechanism is used. In this scheme
 * messages are sent from a socket to its own Message Store. These
 * messages are not designed to travel "on-the-wire" (implemented by
 * setting the hop limit field to zero). These dummy message are
 * formatted as application payload messages at the socket/API
 * level. Application payload messages are used because control
 * messages cannot be created at the API level. When the dummy
 * message with application payload formatting are received by the
 * Message Store Naming Service FSM, they are reformatted as control
 * messages.</p>
 *
 * <p>The idiom that is used to send dummy messages to the message
 * store is:</p>
 *
 * <pre>
 * // Inject the dummy message into the message store. Ignore
 * // the return value because this isn't the forwarding engine.
 *
 * (void) messageStore.receiveMessage (message);
 * </pre>
 *
 * The <code>(void)</code> syntax preceeding the method invocation is
 * not valid Java code; it is an idiom borrowed from the C programming
 * language that means "ignore the return value."
 *
 * Results from the Message Store are received via the asynchonous
 * notificate system of HyperCast rather than by the method call
 * return mechanism. Naming Service API calls pass information to the
 * Message Store via the dummy message mechanism causing the Naming
 * Service FSM to do work on behalf of the API (create a binding,
 * perform a query, etc.). When the Naming Service FSM is finished
 * with the work it notifies the application via an asynchonous
 * message.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version Spring 2005
 */
public class Naming
{
    /**
     * <p>Controls debugging output. Hardcoded to allow debugging code to
     * be eliminated by the compiler when debugging is not wanted.</p>
     */
    private static final boolean debug = false; // non-final allows runtime choice for debugging

    /**
     * <p>The overlay socket associated with this instance of
     * <code>Naming</code>. Cannot be <code>null</code>.</p>
     */
    private final OL_Socket socket;

    /**
     * <p>The log output of this overlay socket.</p>
     */
    private final PrintWriter log;

    /**
     * <p>The message store object associated with the overlay socket
     * associated with this instance of <code>Naming</code>. Cannot be
     * null.</p>
     */
    private final MessageStore messageStore;

    /**
     * <p>The notification handler object associated with the overlay
     * socket (and message store) associated with this instance of
     * <code>Naming</code>. Cannot be <code>null</code>.</p>
     */
    private final NotificationHandler notificationHandler;

    /*
     * <p>Constants used when creating "dummy" messages. These
     * constants are used as the first byte of the payload of the
     * "dummy" messages that are sent to the local message store to
     * initiate a naming operation.</p>
     */
    public static final byte MESSAGE_TYPE_SET_NAME              = 's';
    public static final byte MESSAGE_TYPE_SET_CERT              = 'S';
    public static final byte MESSAGE_TYPE_UNSET_NAME            = 'u';
    public static final byte MESSAGE_TYPE_UNSET_CERT            = 'U';
    public static final byte MESSAGE_TYPE_NAME_QUERY            = 'N';
    public static final byte MESSAGE_TYPE_LOCAL_NAME_QUERY      = 'n';
    public static final byte MESSAGE_TYPE_LOGICAL_ADDRESS_QUERY = 'L';
    public static final byte MESSAGE_TYPE_INSTALL_CERT          = 'I';
    public static final byte MESSAGE_TYPE_PUSH                  = 'P';
    public static final byte MESSAGE_TYPE_INVALIDATE            = 'V';
    public static final byte MESSAGE_TYPE_SET_NAME_DELEGATE     = 'd';
    public static final byte MESSAGE_TYPE_SET_CERT_DELEGATE     = 'D';

    /**
     * <p>This object is used to separate the policy (i.e. socket
     * configuration) from the mechanism in the naming service.</p>
     *
     * @see NamingPolicy
     */
    private final NamingPolicy policy;

    /** Constant used when there is no serial number to be packed in a dummy message. Only query messages have serial numbers. */
    private static final int NO_SERIAL_NUMBER = 0;

    /** Constant used with query APIs to denote that an unlimited amount of time will be spent waiting for a query response. */
    public static final int WAIT_INDEFINITELY = 0;

    /** Constant used with query APIs to denote that an unlimited number of responses will be allowed. */
    public static final int NO_RESPONSE_LIMIT = 0;

    /** Used in a dummy message to signal that the configured query radius should not be overridden. */
    public static final short USE_CONFIGURED_QUERY_RADIUS = (short) 0;

    /**
     * Each query is assigned a unique serial number. Responses contain serial number presented in query. These numbers must
     * match if configured to do so.
     *
     * Note that name and LA queries use same serial number space, i.e. there is no such thing as name query serial number 4
     * *and* LA query serial number 4. Serial number 4 is either a name query or an LA query. This is not an arbitrary
     * decision; they *must* use the same space since any response could satisfy either a name or LA query. To see this
     * consider an example that illustrates the potential ambiguity: LA query for name "foo" assigned serial number 4. Prior
     * name query for LA "123" also assigned serial number 4. Response with serial number 4 received with binding <"foo",
     * "123"> - which query does it satisfy? You can't tell. Name and LA queries must share the same serial number space.
     *
     * Note also that when the policy specifies that serial numbers of responses must match those of queries there is one
     * important exception to this rule: push messages do not carry serial numbers, but the information in push message may
     * satisfy any query (i.e. any serial number).
     *
     * 0 is a distinguished value that is not a valid serial number
     *
     * Interpreted as an unsigned 4 byte quantity.
     *
     * @see #NO_SERIAL_NUMBER
     */
    private int querySerialNumber = 1;

    private synchronized int getNextSerialNumber()
    {
        final int result = querySerialNumber;

        ++querySerialNumber;

        if (querySerialNumber == NO_SERIAL_NUMBER)
            ++querySerialNumber;
            
        return result;
    }

    /**
     * Constructs a naming object and associates it with the specified
     * socket. Each socket may be associated with a single naming
     * object; each naming object must be associated with a socket.
     *
     * @param socket the overlay socket that this naming object will be associated with.
     *
     * @throws HyperCastFatalRuntimeException if the socket argument is null or it does not have a message store.
     * @throws HyperCastConfigException if the socket does not have a message store.
     */
    public Naming (final OL_Socket socket)
    {
        if (socket == null)
            throw new HyperCastFatalRuntimeException (new NullPointerException ("socket argument cannot be null"));

        if (socket.messagestore == null)
            throw new HyperCastConfigException (new NullPointerException ("In order to use naming there must be a message store."));

        if (socket.messagestore.notificationHandler == null)
            throw new HyperCastConfigException (new NullPointerException ("In order to use naming there must be a notification handler."));

        this.socket = socket;
        this.log = socket.config.log;
        this.messageStore = socket.messagestore;
        this.notificationHandler = socket.getNotificationHandler();
        this.policy = NamingPolicy.getInstance (socket.config);

        //this.debug = policy.getDebug();

        if (debug)
            log.println ("Naming System Activated");
    }

    /**
     * Associates a name with the local socket.
     *
     * @param name The name that will be associated with the local socket
     *
     * @throws HyperCastWarningRuntimeException if the parameter <code>name</code> is <code>null<code>
     */
    public void setName (final String name)
    {
        if (debug)
            log.println ("Naming.setName(String)");

        if (name == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Name must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_SET_NAME, name);
    }

    public void setName (final String name, final I_LogicalAddress logicalAddress)
    {
        if (debug)
            log.println ("Naming.setName(String,I_LogicalAddress)");

        if (name == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Name must be non-null"));

        if (logicalAddress == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Logical Address must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_SET_NAME_DELEGATE, name, logicalAddress);
    }

    /**
     * Associates the common name of the subject of the certificate
     * with the local socket. The private key allows the Naming
     * Service FSM in the Message Store to add digitial signatures to
     * the Naming Bindings that it generates for this name.
     *
     * @param certificate The common name of the subject of this
     * certificate will be associated to this socket.
     * @param privateKey This private key, that matches the public key
     * in <code>certificate</code> will be used to digitally sign
     * bindings corresponding to the common name of the subject
     * contained in <code>certificate</code>.
     */
    public void setName (final Certificate certificate, final PrivateKey privateKey)
    {
        if (debug)
            log.println ("Naming.setName(Certificate,PrivateKey)");

        if (certificate == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Certificate must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_SET_CERT, certificate, privateKey);
    }

    public void setName (final Certificate certificate, final PrivateKey privateKey, final String name, final I_LogicalAddress logicalAddress)
    {
        if (debug)
            log.println ("Naming.setName(Certificate,PrivateKey,String,I_LogicalAddress)");

        if (certificate == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Certificate must be non-null"));

        if (privateKey == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Private key must be non-null"));

        if (name == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Name must be non-null"));

        if (logicalAddress == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Logical Address must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_SET_CERT_DELEGATE, name, logicalAddress, certificate, privateKey);
    }

    /**
     * Disassociates a name with this socket. If the name was
     * associated using a certificate and a private key, this
     * operation will not succeed.
     */
    public void unsetName (final String name)
    {
        if (debug)
            log.println ("Naming.unsetName(String)");

        if (name == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Name must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_UNSET_NAME, name);
    }

    /**
     * Disassociates the common name of the subject of the certificate
     * from the local socket. The private key matches the public key
     * of the certificate.
     *
     * @param certificate The common name of the subject of this
     * certificate will be disassociated from this socket.
     * @param privateKey This private key, that matches the public key
     * in <code>certificate</code>.
     */
    public void unsetName (final Certificate certificate, final PrivateKey privateKey)
    {
        if (debug)
            log.println ("Naming.unsetName(Certificate,PrivateKey)");

        if (certificate == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Certificate must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_UNSET_CERT, certificate, privateKey);
    }

    /**
     * Installs a certificate as a trust anchor. When a certificate
     * chain is being built it need only reach a trust anchor to be
     * trusted.
     *
     * @param certificate The certificate of the trust anchor
     */
    public void installTrustedNamingCertificate (final Certificate certificate)
    {
        if (debug)
            log.println ("Naming.installTrustedNamingCertificate()");

        if (certificate == null)
            throw new HyperCastWarningRuntimeException (new NullPointerException ("Certificate must be non-null"));

        sendDummyMessage (MESSAGE_TYPE_INSTALL_CERT, certificate);
    }

    /**
     * Returns the names that are associated with this socket.
     *
     * @return The names associated with this overlay socket. This method never returns <code>null</code>.
     */
    public String [] getNames()
    {
        if (debug)
            log.println ("getNames()");

        final Vector result = new Vector();

        long now = System.currentTimeMillis();

        final long timeout = 2 * 1000; // 2 seconds
        final long deadline = timeout + now;

        final int serialNumber = getNextSerialNumber();

        sendDummyMessage (MESSAGE_TYPE_LOCAL_NAME_QUERY, serialNumber);

        while (now < deadline)
        {
            final NAMING_EVENT response = waitForResponse (deadline, serialNumber, true);

            if (response == null) // timeout
                break;

            if (debug)
                log.println ("getNames (local) - response: " + response);

            final ImmutableNamingBinding [] bindings = response.getLocalBindings();

            if (bindings.length == 0) {
                now = System.currentTimeMillis();
            }
            else
            {
                for (int i = 0; i < bindings.length; ++i)
                    result.add (bindings[i].getName());

                break;
            }
        }

        return (String []) result.toArray (new String [0]);
    }

    /**
     * @see #getNames(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public String [] getNames (final I_LogicalAddress peer) {
        return getNames (peer,
                         policy.getTrustedResponsesOnly(),
                         policy.getBlockingAPITimeout(),
                         policy.getRequireBindingMessageSerialNumberMatch(),
                         policy.getAuthoritativeResponsesOnly(),
                         policy.getMaximumResponseBindingAge(),
                         policy.getMaximumResponsesPerQuery(),
                         policy.getMinimumQueryHopCount(),
                         USE_CONFIGURED_QUERY_RADIUS);
    }

    /**
     * <p>Calls <code>getNamesNonBlocking()</code> to inject a
     * dummy message into the message store to query the names of
     * the given logical address; blocks waiting for responses to the
     * query.</p>
     *
     * @see #getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     *
     * @param peer The logical address for which a name lookup is being performed
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     *
     * @return An array of the names associated with the given logical address.
     */
    public String [] getNames (final I_LogicalAddress peer,
                               final boolean          requireTrusted,
                               final long             timeout,
                               final boolean          requireSerialNumberMatch,
                               final boolean          requireAuthoritative,
                               final int              maxAge,
                               final int              maxResponses,
                               final short            minHopCount,
                               final short            maxHopCount)
    {
        if (maxResponses == NO_RESPONSE_LIMIT
            &&
            timeout == WAIT_INDEFINITELY)
            throw new HyperCastWarningRuntimeException ("Cannot have both NO_RESPONSE_LIMIT and WAIT_INDEFINITELY: would never return!");

        final String logicalAddressString = peer.toString();

        final int serialNumber = getNextSerialNumber();

        getNamesNonBlocking (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber);

        final long deadline = (timeout == WAIT_INDEFINITELY)
            ? Long.MAX_VALUE
            : timeout + System.currentTimeMillis();

        int responseCount = 0;
        final Vector result = new Vector();

        while (maxResponses == NO_RESPONSE_LIMIT
               ||
               responseCount < maxResponses)
        {
            final NAMING_EVENT response = waitForResponse (null, logicalAddressString, deadline, serialNumber, requireSerialNumberMatch);

            if (response == null)
                break;

            if (debug)
                log.println ("getNames - response: " + response);

            final ImmutableNamingBinding [] bindings = response.getBindingsWithLogicalAddress (logicalAddressString);
            responseCount += bindings.length;
            for (int i = 0; i < bindings.length; ++i)
            {
                final String resultName = bindings[i].getName();
                result.add (resultName);
            }
        }

        return (String []) result.toArray (new String [0]);
    }

    /**
     * @see #getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer) {
        return getNamesNonBlocking (peer,
                                    policy.getTrustedResponsesOnly(),
                                    policy.getBlockingAPITimeout(),
                                    policy.getRequireBindingMessageSerialNumberMatch(),
                                    policy.getAuthoritativeResponsesOnly(),
                                    policy.getMaximumResponseBindingAge(),
                                    policy.getMaximumResponsesPerQuery(),
                                    policy.getMinimumQueryHopCount(),
                                    USE_CONFIGURED_QUERY_RADIUS);
    }

    /**
     * Performs a query on the given logical address to get the
     * names of the socket having that logical address.
     *
     * @param peer The logical address for which a name lookup is being performed
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     *
     * @return serial number of query
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer,
                                    final boolean          requireTrusted,
                                    final long             timeout,
                                    final boolean          requireSerialNumberMatch,
                                    final boolean          requireAuthoritative,
                                    final int              maxAge,
                                    final int              maxResponses,
                                    final short            minHopCount,
                                    final short            maxHopCount)
    {
        final int serialNumber = getNextSerialNumber();
        getNamesNonBlocking (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber);
        return serialNumber;
    }

    /**
     * Implementation of getNamesNonBlocking
     */
    private void getNamesNonBlocking (final I_LogicalAddress peer,
                                      final boolean          requireTrusted,
                                      final long             timeout,
                                      final boolean          requireSerialNumberMatch,
                                      final boolean          requireAuthoritative,
                                      final int              maxAge,
                                      final int              maxResponses,
                                      final short            minHopCount,
                                      final short            maxHopCount,
                                      final int              serialNumber)
    {
        if (debug)
            log.println ("getNames ("               +
                         "peer"                     + " -> " + peer                     + ", " +
                         "requireTrusted"           + " -> " + requireTrusted           + ", " +
                         "timeout"                  + " -> " + timeout                  + ", " +
                         "requireSerialNumberMatch" + " -> " + requireSerialNumberMatch + ", " +
                         "requireAuthoritative"     + " -> " + requireAuthoritative     + ", " +
                         "maxAge"                   + " -> " + maxAge                   + ", " +
                         "maxResponses"             + " -> " + maxResponses             + ", " +
                         "minHopCount"              + " -> " + minHopCount              + ", " +
                         "maxHopCount"              + " -> " + maxHopCount              + ", " +
                         "serialNumber"             + " -> " + serialNumber             +
                         ")...");

        final String logicalAddressString = peer.toString();
        sendDummyMessage (MESSAGE_TYPE_NAME_QUERY, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber, logicalAddressString);
    }

    /**
     * @see #getLogicalAddressByName(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String name) {
        return getLogicalAddressByName (name,
                                        policy.getTrustedResponsesOnly(),
                                        policy.getBlockingAPITimeout(),
                                        policy.getRequireBindingMessageSerialNumberMatch(),
                                        policy.getAuthoritativeResponsesOnly(),
                                        policy.getMaximumResponseBindingAge(),
                                        policy.getMaximumResponsesPerQuery(),
                                        policy.getMinimumQueryHopCount(),
                                        USE_CONFIGURED_QUERY_RADIUS);
    }

    /**
     * Calls <code>getLogicalAddressByNameNonBlocking()</code> to inject a
     * dummy message into the message store to query the logical
     * address of the name; blocks waiting for a response to the
     * query.
     *
     * @see #getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     *
     * @param name The logical address for which a name lookup is being performed.
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return A array of the logical addresses associated with the given name.
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String  name,
                                                        final boolean requireTrusted,
                                                        final long    timeout,
                                                        final boolean requireSerialNumberMatch,
                                                        final boolean requireAuthoritative,
                                                        final int     maxAge,
                                                        final int     maxResponses,
                                                        final short   minHopCount,
                                                        final short   maxHopCount)
    {
        if (maxResponses == NO_RESPONSE_LIMIT
            &&
            timeout == WAIT_INDEFINITELY)
            throw new HyperCastWarningRuntimeException ("Cannot have both NO_RESPONSE_LIMIT and WAIT_INDEFINITELY: would never return!");

        final int serialNumber = getNextSerialNumber();
        getLogicalAddressByNameNonBlocking (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber);

        final long deadline = (timeout == WAIT_INDEFINITELY)
            ? Long.MAX_VALUE
            : timeout + System.currentTimeMillis();

        int responseCount = 0;
        final Vector result = new Vector();

        while (maxResponses == NO_RESPONSE_LIMIT
               ||
               responseCount < maxResponses)
        {
            if (debug)
                log.println ("getLogicalAddressByName - waiting for response, already have " + responseCount);

            final NAMING_EVENT response = waitForResponse (name, null, deadline, serialNumber, requireSerialNumberMatch);

            if (response == null)
                break;

            if (debug)
                log.println ("getLogicalAddressByName - response: " + response);

            final ImmutableNamingBinding [] bindings = response.getBindingsWithName (name);
            responseCount += bindings.length;
            for (int i = 0; i < bindings.length; ++i)
            {
                final String logicalAddressString = bindings[i].getLogicalAddressString();
                final I_LogicalAddress resultLogicalAddress = socket.createLogicalAddress (logicalAddressString);
                result.add (resultLogicalAddress);
            }
        }

        return (I_LogicalAddress []) result.toArray (new I_LogicalAddress [0]);
    }

    /**
     * @see #getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getLogicalAddressByNameNonBlocking (final String name) {
        return getLogicalAddressByNameNonBlocking (name,
                                                   policy.getTrustedResponsesOnly(),
                                                   policy.getBlockingAPITimeout(),
                                                   policy.getRequireBindingMessageSerialNumberMatch(),
                                                   policy.getAuthoritativeResponsesOnly(),
                                                   policy.getMaximumResponseBindingAge(),
                                                   policy.getMaximumResponsesPerQuery(),
                                                   policy.getMinimumQueryHopCount(),
                                                   USE_CONFIGURED_QUERY_RADIUS);
    }

    /**
     * Performs a query on the given name to get the logical
     * addresses of the sockets having that name.
     *
     * @param name The name for which a logical address lookup is being performed
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return serial number of query
     */
    public int getLogicalAddressByNameNonBlocking (final String  name,
                                                   final boolean requireTrusted,
                                                   final long    timeout,
                                                   final boolean requireSerialNumberMatch,
                                                   final boolean requireAuthoritative,
                                                   final int     maxAge,
                                                   final int     maxResponses,
                                                   final short   minHopCount,
                                                   final short   maxHopCount)
    {
        final int serialNumber = getNextSerialNumber();
        getLogicalAddressByNameNonBlocking (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber);
        return serialNumber;
    }

    /**
     * Implementation of getLogicalAddressByNameNonBlocking
     */
    private void getLogicalAddressByNameNonBlocking (final String  name,
                                                     final boolean requireTrusted,
                                                     final long    timeout,
                                                     final boolean requireSerialNumberMatch,
                                                     final boolean requireAuthoritative,
                                                     final int     maxAge,
                                                     final int     maxResponses,
                                                     final short   minHopCount,
                                                     final short   maxHopCount,
                                                     final int     serialNumber)
    {
        if (debug)
            log.println ("getLogicalAddressByName ("    +
                         "name"                     + " -> " + name                     + ", " +
                         "requireTrusted"           + " -> " + requireTrusted           + ", " +
                         "timeout"                  + " -> " + timeout                  + ", " +
                         "requireSerialNumberMatch" + " -> " + requireSerialNumberMatch + ", " +
                         "requireAuthoritative"     + " -> " + requireAuthoritative     + ", " +
                         "maxAge"                   + " -> " + maxAge                   + ", " +
                         "maxResponses"             + " -> " + maxResponses             + ", " +
                         "minHopCount"              + " -> " + minHopCount              + ", " +
                         "maxHopCount"              + " -> " + maxHopCount              + ", " +
                         "serialNumber"             + " -> " + serialNumber             +
                         ")...");

        sendDummyMessage (MESSAGE_TYPE_LOGICAL_ADDRESS_QUERY, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber, name);
    }

    /**
     * <p>Called by <code>OL_Socket.joinOverlay()</code> to trigger a
     * name push, depending on configured policy for naming
     * service.</p>
     *
     * @see OL_Socket#joinOverlay()
     */
    /* package */ void joinOverlay()
    {
        if (debug)
            log.println ("naming.joinOverlay() called");

        if (policy.doPushOnJoinOverlay())
            sendDummyMessage (MESSAGE_TYPE_PUSH);
    }

    /**
     * <p>Called by <code>OL_Socket.leaveOverlay()</code>. Included
     * for completeness.
     *
     * @see OL_Socket#leaveOverlay()
     */
    /* package */ void leaveOverlay()
    {
        if (debug)
            log.println ("naming.leaveOverlay() called");

        if (policy.doInvalidateOnLeaveOverlay())
            sendDummyMessage (MESSAGE_TYPE_INVALIDATE);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type) {
        sendDummyMessage (type, NO_SERIAL_NUMBER, null, null, null, null);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final int serialNumber) {
        sendDummyMessage (type, serialNumber, null, null, null, null);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final String string) {
        sendDummyMessage (type, string, null, null, null);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final String string, final I_LogicalAddress logicalAddress) {
        sendDummyMessage (type, string, logicalAddress, null, null);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final Certificate certificate) {
        sendDummyMessage (type, null, null, certificate, null);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final Certificate certificate, final PrivateKey privateKey) {
        sendDummyMessage (type, null, null, certificate, privateKey);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final String string, final I_LogicalAddress logicalAddress, final Certificate certificate, final PrivateKey privateKey) {
        sendDummyMessage (type, NO_SERIAL_NUMBER, string, logicalAddress, certificate, privateKey);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte type, final int serialNumber, final String string, final I_LogicalAddress logicalAddress, final Certificate certificate, final PrivateKey privateKey)
        throws HyperCastWarningRuntimeException
    {
        sendDummyMessage (type,
                          policy.getTrustedResponsesOnly(),
                          policy.getBlockingAPITimeout(),
                          policy.getRequireBindingMessageSerialNumberMatch(),
                          policy.getAuthoritativeResponsesOnly(),
                          policy.getMaximumResponseBindingAge(),
                          policy.getMaximumResponsesPerQuery(),
                          policy.getMinimumQueryHopCount(),
                          USE_CONFIGURED_QUERY_RADIUS,
                          serialNumber,
                          string,
                          logicalAddress,
                          certificate,
                          privateKey);
    }

    /** @see #sendDummyMessage(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)*/
    private void sendDummyMessage (final byte        type,
                                   final boolean     requiresTrusted,
                                   final long        timeout,
                                   final boolean     requireSerialNumberMatch,
                                   final boolean     requiresAuthoritative,
                                   final int         maxAge,
                                   final int         maxResponses,
                                   final short       minHopCount,
                                   final short       maxHopCount,
                                   final int         serialNumber,
                                   final String      string)
        throws HyperCastWarningRuntimeException
    {
        sendDummyMessage (type, requiresTrusted, timeout, requireSerialNumberMatch, requiresAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount, serialNumber, string, null, null, null);
    }

    /**
     * Delivers a dummy message to the message store. There is only
     * one format for dummy messages, it contains the union of all
     * information that is passed by the various Naming Service APIs.
     *
     * @param type Specifies the type of dummy message, decoded in Naming Service Message Store FSM
     * @param requiresTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requiresAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     * @param serialNumber Query serial number; 0 if not query.
     * @param string The string for which a query is being performed, either a logical address or a name. Also, the string of the name being use for a binding.
     * @param certificate The common name of the subject of this certificate will be used as a name for this socket.
     * @param privateKey This private key, that matches the public key in <code>certificate</code>.
     * @param logicalAddress The logical address that will be used in a new binding or null if local address is to be used.
     */
    private void sendDummyMessage (final byte             type,
                                   final boolean          requiresTrusted,
                                   final long             timeout,
                                   final boolean          requireSerialNumberMatch,
                                   final boolean          requiresAuthoritative,
                                   final int              maxAge,
                                   final int              maxResponses,
                                   final short            minHopCount,
                                   final short            maxHopCount,
                                   final int              serialNumber,
                                   final String           string,
                                   final I_LogicalAddress logicalAddress,
                                   final Certificate      certificate,
                                   final PrivateKey       privateKey)
        throws HyperCastWarningRuntimeException
    {
        try
        {
            final long deadline = (timeout == WAIT_INDEFINITELY)
                ? Long.MAX_VALUE
                : timeout + System.currentTimeMillis();

            final byte [] payloadData = packDummyData (type,
                                                       requiresTrusted,
                                                       deadline,
                                                       requireSerialNumberMatch,
                                                       requiresAuthoritative,
                                                       maxAge,
                                                       maxResponses,
                                                       minHopCount,
                                                       maxHopCount,
                                                       serialNumber,
                                                       string,
                                                       logicalAddress,
                                                       certificate,
                                                       privateKey);
            final byte [] streamID = FSM_Extension.NAMING_STREAM_ID;
            final OL_Message message = (OL_Message) socket.createMessage (payloadData, streamID); // note: blind cast
            message.setHopLimit ((short) 0);

            messageStore.receiveMessage (message);
        }
        catch (final CertificateEncodingException cee)
        {
            throw new HyperCastWarningRuntimeException ("Could not encode certificate", cee);
        }
    }

    //
    // !NOTE! These constants are need in the Naming Service FSM to
    // decode dummy messages. They are *copied* there (not
    // referenced from here because the FSM is not supposed to be able
    // to access this portion of the system). THE COPIES OF THESE
    // CONSTANTS MUST BE KEPT IN SYNC.
    //

    // byte the identifies the type of message
    private static final int MESSAGE_TYPE_LENGTH        = 1;

    // of local interest only
    private static final int TRUSTED_LENGTH             = 1;
    private static final int DEADLINE_LENGTH            = 8;
    private static final int SERIAL_NUMBER_MATCH_LENGTH = 1;

    // sent on wire (depending on message type)
    private static final int AUTHORITATIVE_LENGTH       = 1;
    private static final int MAXIMUM_AGE_LENGTH         = 4;
    private static final int MAX_RESPONSES_LENGTH       = 4;
    private static final int MINIMUM_HOP_COUNT_LENGTH   = 2;
    private static final int MAXIMUM_HOP_COUNT_LENGTH   = 2;
    private static final int SERIAL_NUMBER_LENGTH       = 4;

    // end of fixed length fields

    // sent on wire (depending on message type)
    private static final int STRING_LENGTH_BYTES        = 2;
    private static final int CERTIFICATE_LENGTH_BYTES   = 2;

    // of local interest only
    private static final int LOGICAL_ADDRESS_LENGTH_BYTES = 2;
    private static final int PRIVATE_KEY_LENGTH_BYTES   = 2;

    // length of fields that are the prefix of query messages sent from peer to peer
    private static final int PEER_QUERY_PREFIX_LENGTH =
    AUTHORITATIVE_LENGTH     +
    MAXIMUM_AGE_LENGTH       +
    MAX_RESPONSES_LENGTH     +
    MINIMUM_HOP_COUNT_LENGTH +
    SERIAL_NUMBER_LENGTH;

    // length of fields that are the prefix of query messages sent from API to FSM
    private static final int LOCAL_QUERY_PREFIX_LENGTH =
    MESSAGE_TYPE_LENGTH        +
    TRUSTED_LENGTH             +
    DEADLINE_LENGTH            +
    SERIAL_NUMBER_MATCH_LENGTH +
    MAXIMUM_HOP_COUNT_LENGTH   +
    PEER_QUERY_PREFIX_LENGTH;

    /**
     * type (1 byte) | SN (4 bytes) | trusted (1 byte) | deadline (8 bytes) | matchSN (1 byte) | auth (1 byte) | max age (4) | max responses (4 bytes) | min hops (2 bytes) |
     * string length (2 bytes) | string (0 or more bytes) | cert length (2 bytes) | cert (0 or more bytes) | priv key length (2 bytes) | priv key (0 or more bytes)
     *
     * @param string can be null
     * @param certificate can be null
     */
    private byte [] packDummyData (final byte        type,
                                   final boolean     requiresTrusted,
                                   final long        deadline,
                                   final boolean     requireSerialNumberMatch,
                                   final boolean     requiresAuthoritative,
                                   final int         maxAge,
                                   final int         maxResponses,
                                   final short       minHopCount,
                                   final short       maxHopCount,
                                   final int         serialNumber,
                                   final String      string,
                                   final I_LogicalAddress logicalAddress,
                                   final Certificate certificate,
                                   final PrivateKey  privateKey)
        throws CertificateEncodingException
    {
        final byte [] stringBytes         = (string         != null) ? string.getBytes()                    : new byte [0];
        final byte [] logicalAddressBytes = (logicalAddress != null) ? logicalAddress.toString().getBytes() : new byte [0];
        final byte [] certificateBytes    = (certificate    != null) ? certificate.getEncoded()             : new byte [0];
        final byte [] privateKeyBytes     = (privateKey     != null) ? privateKey.getEncoded()              : new byte [0];

        final int resultLength =
            MESSAGE_TYPE_LENGTH        +
            SERIAL_NUMBER_LENGTH       +
            TRUSTED_LENGTH             +
            DEADLINE_LENGTH            +
            SERIAL_NUMBER_MATCH_LENGTH +
            AUTHORITATIVE_LENGTH       +
            MAXIMUM_AGE_LENGTH         +
            MAX_RESPONSES_LENGTH       +
            MINIMUM_HOP_COUNT_LENGTH   +
            MAXIMUM_HOP_COUNT_LENGTH   +
            STRING_LENGTH_BYTES        + stringBytes.length      +
            CERTIFICATE_LENGTH_BYTES   + certificateBytes.length +
            PRIVATE_KEY_LENGTH_BYTES   + privateKeyBytes.length  +
            LOGICAL_ADDRESS_LENGTH_BYTES + logicalAddressBytes.length;

        final byte [] result = new byte [resultLength];
        int index = 0;

        result [index] = type;
        index += MESSAGE_TYPE_LENGTH;

        // serialNumber
        result [index + 0] = (byte) ((serialNumber >>> (8*3)) & 0xFF);
        result [index + 1] = (byte) ((serialNumber >>> (8*2)) & 0xFF);
        result [index + 2] = (byte) ((serialNumber >>> (8*1)) & 0xFF);
        result [index + 3] = (byte) ((serialNumber >>> (8*0)) & 0xFF);
        index += SERIAL_NUMBER_LENGTH;

        // trusted response required
        result [index + 0] = (byte) ((requiresTrusted ? 1 : 0) & 0xFF);
        index += TRUSTED_LENGTH;

        // deadline
        result [index + 0] = (byte) ((deadline >>> (8*7)) & 0xFF);
        result [index + 1] = (byte) ((deadline >>> (8*6)) & 0xFF);
        result [index + 2] = (byte) ((deadline >>> (8*5)) & 0xFF);
        result [index + 3] = (byte) ((deadline >>> (8*4)) & 0xFF);
        result [index + 4] = (byte) ((deadline >>> (8*3)) & 0xFF);
        result [index + 5] = (byte) ((deadline >>> (8*2)) & 0xFF);
        result [index + 6] = (byte) ((deadline >>> (8*1)) & 0xFF);
        result [index + 7] = (byte) ((deadline >>> (8*0)) & 0xFF);
        index += DEADLINE_LENGTH;

        // serial number match required
        result [index + 0] = (byte) ((requireSerialNumberMatch ? 1 : 0) & 0xFF);
        index += SERIAL_NUMBER_MATCH_LENGTH;

        // authoritative response required
        result [index + 0] = (byte) ((requiresAuthoritative ? 1 : 0) & 0xFF);
        index += AUTHORITATIVE_LENGTH;
        
        // maximum allowable age of a response
        result [index + 0] = (byte) ((maxAge >>> (8*3)) & 0xFF);
        result [index + 1] = (byte) ((maxAge >>> (8*2)) & 0xFF);
        result [index + 2] = (byte) ((maxAge >>> (8*1)) & 0xFF);
        result [index + 3] = (byte) ((maxAge >>> (8*0)) & 0xFF);
        index += MAXIMUM_AGE_LENGTH;

        // maxResponses
        result [index + 0] = (byte) ((maxResponses >>> (8*3)) & 0xFF);
        result [index + 1] = (byte) ((maxResponses >>> (8*2)) & 0xFF);
        result [index + 2] = (byte) ((maxResponses >>> (8*1)) & 0xFF);
        result [index + 3] = (byte) ((maxResponses >>> (8*0)) & 0xFF);
        index += MAX_RESPONSES_LENGTH;

        // minHopCount
        result [index + 0] = (byte) ((minHopCount >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((minHopCount >>> (8*0)) & 0xFF);
        index += MINIMUM_HOP_COUNT_LENGTH;

        // maxHopCount
        result [index + 0] = (byte) ((maxHopCount >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((maxHopCount >>> (8*0)) & 0xFF);
        index += MAXIMUM_HOP_COUNT_LENGTH;

        // string length
        result [index + 0] = (byte) ((stringBytes.length >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((stringBytes.length >>> (8*0)) & 0xFF);
        index += STRING_LENGTH_BYTES;

        // string data
        System.arraycopy (stringBytes, 0, result, index, stringBytes.length);
        index += stringBytes.length;

        // certificate length
        result [index + 0] = (byte) ((certificateBytes.length >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((certificateBytes.length >>> (8*0)) & 0xFF);
        index += CERTIFICATE_LENGTH_BYTES;

        // certificate data
        System.arraycopy (certificateBytes, 0, result, index, certificateBytes.length);
        index += certificateBytes.length;

        // private key length
        result [index + 0] = (byte) ((privateKeyBytes.length >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((privateKeyBytes.length >>> (8*0)) & 0xFF);
        index += PRIVATE_KEY_LENGTH_BYTES;

        // private key data
        System.arraycopy (privateKeyBytes, 0, result, index, privateKeyBytes.length);
        index += privateKeyBytes.length;

        // logical address length
        result [index + 0] = (byte) ((logicalAddressBytes.length >>> (8*1)) & 0xFF);
        result [index + 1] = (byte) ((logicalAddressBytes.length >>> (8*0)) & 0xFF);
        index += LOGICAL_ADDRESS_LENGTH_BYTES;

        // logical address data
        System.arraycopy (logicalAddressBytes, 0, result, index, logicalAddressBytes.length);
        index += logicalAddressBytes.length;

        return result;
    }

    /**
     * @see #waitForResponse(String, String, long, int, boolean)
     */
    private NAMING_EVENT waitForResponse (final long deadline, final int serialNumber, final boolean requireSerialNumberMatch) {
        return waitForResponse (null, null, deadline, serialNumber, requireSerialNumberMatch);  // null/null means return any event
    }

    /**
     * <p>Waits for a matching <code>NAMING_EVENT</code> to arrive via
     * the notification system. The parameters of this method
     * constitute a template that is used in the matching of
     * events. This call can return before a match is found if the
     * a timeout occurs.
     *
     * <p>If both parameters are <code>null</code> then any event will be returned.
     *
     * @param name If not null, a name that must be matched in a response.
     * @param logicalAddressString If not null, a logical address that must be matched in a response.
     * @param deadline The absolute time at which waiting should cease.
     * @param serialNumber Unique serial number of a query
     * @param requireSerialNumberMatch If true, serial number match must occur
     *
     * @return The <code>NAMING_EVENT</code> that matched the specified template, or null if no match was found before a timeout.
     *
     * @throws HyperCastFatalRuntimeException If the template event contains no information for use in matching.
     */
    private NAMING_EVENT waitForResponse (final String name, final String logicalAddressString, final long deadline, final int serialNumber, final boolean requireSerialNumberMatch)
    {
        final String debugParamString = "waitForResponse (" + name + ", " + logicalAddressString + ", " + deadline + "," + serialNumber + ", " + requireSerialNumberMatch + ")";

        if (debug)
            log.println (debugParamString);

        final long start = System.currentTimeMillis();
        final long timeout = deadline - start;
        long elapsed = 0; // ms

        NAMING_EVENT event = null;

        while ((event == null  // nothing has been received
                ||
                // we require serial numbers of responses to match those of queries and the serial number is not a match
                (requireSerialNumberMatch && (event.getSerialNumber() != serialNumber))
                ||
                // we are trying to match a name, but the name of the received event is not a match
                ((name != null && ! event.containsName (name))
                 ||
                 // we are trying to match a logical address, but the logical address of the received event is not a match
                 (logicalAddressString != null && ! event.containsLogicalAddress (logicalAddressString))))
               &&
               (deadline == WAIT_INDEFINITELY || elapsed < timeout))
        {
            event = notificationHandler.waitUntil_NAMING_EVENT (timeout - elapsed);

            final long now = System.currentTimeMillis();
            elapsed = now - start;

            if (event == null)
            {
                if (debug)
                    log.println (debugParamString + " encounterd timeout: (timeout: " + timeout + " ms, elapsed: " + elapsed + " ms)");
            }
            else
            {
                if (debug)
                    log.println (debugParamString + " received event: (" + event + "), (timeout: " + timeout + " ms, elapsed: " + elapsed + " ms)");
            }
        }

        if (debug)
            log.println (debugParamString + " returning: " + event + " (timeout: " + timeout + " ms , elapsed: " + elapsed + " ms)");

        return event;
    }
}

/**
 * <p>This class encapsulates all policy information regarding the
 * naming service. NamingPolicy information, obtained from reading a
 * configuration file, or altered at run-time via the monitor and
 * control mechanism through the statistics interface, is used to
 * make policy decisions. An example of a policy decision is:
 * should a name/logical address binding push operation occur when
 * a new name is set for a socket to manage? Should it occur when
 * the socket joins an overlay network? Should it occur
 * periodically?</p>
 *
 * <p>These choices are not "hardcoded" into the logic that
 * implements the mechanisms of the naming service, rather policy
 * and mechanism are separated by confining policy information in
 * this class.</p>
 */
class NamingPolicy
{
    private static boolean debugPolicy = false;

    private static NamingPolicy singleton = null;

    public static final long NAMING_BLOCK_INDEFINITELY = 0;

    private int     bindingTimeout;               // ms
    private long    timeoutAPI;                   // ms; 0 -> indefinitely
    private short   pushRadius;                   // hops
    private short   queryRadius;                  // hops
    private boolean pushOnSetName;
    private boolean invalidateOnUnsetName;
    private boolean pushOnJoinOverlay;
    private boolean invalidateOnLeaveOverlay;
    private boolean pushOnChangeOfLogicalAddress;
    private int     pushPeriod;                   // ms; 0 -> no periodic push
    private short   invalidateRadius;
    private String  hashAlgorithm;
    private boolean trustedResponses;
    private boolean authoritativeResponses;
    private int     maxResponseAge;               // ms
    private int     maxResponseCount;
    private short   minHopCount;
//    private boolean debug;
    private long    certificateRequestTimeout;
    private int     peerBindingCachePushMaximumSize;
    private int     peerBindingCachePullMaximumSize;
    private boolean snoopBindings;
    private boolean cachePulledBindings;
    private boolean snoopCertificates;
    private String  statsName;
    private boolean requireBindingSerialNumberMatch;
    private boolean requireTrustSerialNumberMatch;
    private boolean tracingQueries;
    private int     queryTracingModulo;
    private boolean preallocateFSM;
    private boolean preallocateCrypto;
    private boolean bypassPeerBindingsCacheOnQuery;
    private boolean certificateCaching;
    private boolean writeCertificatesToBackingStore;
//    private boolean bindingsStickyTrust;

    // TODO(EXPERIMENTS): remove
    private long queryFrequency;
    private int socketsInNetwork;
    private boolean doNamingQueries;
    private int experimentQueryCount;

    /**
     * <p>The log output of this overlay socket.</p>
     */
    private final PrintWriter log;

    /**
     * Ensures that there is only a single NamingPolicy object.
     */
    public static synchronized NamingPolicy getInstance (final HyperCastConfig configuration) {
        if (singleton == null)
            singleton = new NamingPolicy (configuration);

        return singleton;
    }

    private static final String ATTRIBUTE_PREFIX = "/Public/NetworkService";

    /*
     * Constructs a <code>Policy</code> object based on the given
     * <code>HyperCastConfig</code>.
     *
     * @see HyperCastConfig
     *
     * @param configuration A <code>HyperCastConfig</code> object that is used to initial the <code>Policy</code> object.
     * @throws HyperCastConfigException If a configuration value cannot be parsed.
     */
    private NamingPolicy (final HyperCastConfig configuration)
    {
        this.log = configuration.log;

        /*
         * Debugging. Comment out this code and change the debug
         * variable in Naming to have modifier "final" for a
         * compile time setting of debug.
         */
/*
        final String debugString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/Debug"));
        this.debug = configuration.booleanFromYesNoString (debugString);
        if (debugPolicy)
            log.println ("Naming configuration - debug: " + debug);
*/

        /*
         * Timeout for a binding
         */
        this.bindingTimeout = configuration.getIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/BindingTimeout"));
        if (debugPolicy)
            log.println ("Naming configuration - BindingTimeout: " + bindingTimeout);

        /*
         * Timeout for a blocking api call
         */
        this.timeoutAPI = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/BlockingAPITimeout"));
        if (debugPolicy) log.println ("Naming configuration - timeoutAPI: " + timeoutAPI);

        /*
         * TODO: Invalidate on move? motivation: can only
         * invalidate a particular LA when you have that LA, therefore must invalidate before changing
         */

        /*
         * Push Radius
         */
        this.pushRadius = configuration.getNonNegativeShortAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PushRadius"));
        if (debugPolicy)
            log.println ("Naming configuration - pushRadius: " + pushRadius);

        /*
         * Query Radius
         */
        this.queryRadius = configuration.getNonNegativeShortAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/QueryRadius"));
        if (debugPolicy)
            log.println ("Naming configuration - queryRadius: " + queryRadius);

        /*
         * Push on Set Name
         */
        final String pushOnSetNameString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PushOnSetName"));
        this.pushOnSetName = configuration.booleanFromYesNoString (pushOnSetNameString);
        if (debugPolicy)
            log.println ("Naming configuration - pushOnSetName: " + pushOnSetName);

        /*
         * Invalidate on Unset Name
         */
        final String invalidateOnUnsetNameString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/InvalidateOnUnsetName"));
        this.invalidateOnUnsetName = configuration.booleanFromYesNoString (invalidateOnUnsetNameString);
        if (debugPolicy)
            log.println ("Naming configuration - invalidateOnUnsetName: " + invalidateOnUnsetName);

        /*
         * Push on Join Overlay
         */
        final String pushOnJoinOverlayString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PushOnJoinOverlay"));
        this.pushOnJoinOverlay = configuration.booleanFromYesNoString (pushOnJoinOverlayString);
        if (debugPolicy)
            log.println ("Naming configuration - pushOnJoinOverlay: " + pushOnJoinOverlay);

        /*
         * Invalidate on Leave Overlay
         */
        final String invalidateOnLeaveOverlayString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/InvalidateOnLeaveOverlay"));
        this.invalidateOnLeaveOverlay = configuration.booleanFromYesNoString (invalidateOnLeaveOverlayString);
        if (debugPolicy)
            log.println ("Naming configuration - invalidateOnLeaveOverlay: " + invalidateOnLeaveOverlay);

        /*
         * Push on Change of Logical Address
         */
        final String pushOnChangeOfLogicalAddressString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PushOnChangeOfLogicalAddress"));
        this.pushOnChangeOfLogicalAddress = configuration.booleanFromYesNoString (pushOnChangeOfLogicalAddressString);
        if (debugPolicy)
            log.println ("Naming configuration - pushOnChangeOfLogicalAddress: " + pushOnChangeOfLogicalAddress);

        /*
         * Push Period
         */
        this.pushPeriod = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PushPeriod"));
        if (debugPolicy)
            log.println ("Naming configuration - pushPeriod: " + pushPeriod);

        /*
         * Invalidate Radius
         */
        this.invalidateRadius = configuration.getNonNegativeShortAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/InvalidateRadius"));
        if (debugPolicy)
            log.println ("Naming configuration - invalidateRadius: " + invalidateRadius);

        /*
         * DigitalSignatureAlgorithm
         */
        this.hashAlgorithm = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/DigitalSignatureAlgorithm"));
        if (debugPolicy)
            log.println ("Naming configuration - hashAlgorithm: " + hashAlgorithm);

        /*
         * TrustedResponsesOnly
         */
        final String trustedResponsesString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/TrustedResponsesOnly"));
        this.trustedResponses = configuration.booleanFromYesNoString (trustedResponsesString);
        if (debugPolicy)
            log.println ("Naming configuration - trustedResponses: " + trustedResponses);

        /*
         * AuthoritativeResponsesOnly
         */
        final String authoritativeResponsesString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/AuthoritativeResponsesOnly"));
        this.authoritativeResponses = configuration.booleanFromYesNoString (authoritativeResponsesString);
        if (debugPolicy)
            log.println ("Naming configuration - authoritativeResponses: " + authoritativeResponses);

        /*
         * MaximumResponseBindingAge
         */
        this.maxResponseAge = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/MaximumResponseBindingAge"));
        if (debugPolicy)
            log.println ("Naming configuration - maxResponseAge: " + maxResponseAge);

        /*
         * MaximumResponsesPerQuery
         */
        this.maxResponseCount = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/MaximumResponsesPerQuery"));
        if (debugPolicy)
            log.println ("Naming configuration - maxResponseCount: " + maxResponseCount);

        /*
         * MinimumQueryRadius
         */
        this.minHopCount = configuration.getNonNegativeShortAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/MinimumQueryRadius"));
        if (debugPolicy)
            log.println ("Naming configuration - minHopCount: " + minHopCount);

        /*
         * CertificateRequestTimeout
         */
        this.certificateRequestTimeout = configuration.getNonNegativeShortAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/CertificateRequestTimeout"));
        if (debugPolicy)
            log.println ("Naming configuration - certificateRequestTimeout: " + certificateRequestTimeout);

        /*
         * PeerBindingCachePushMaximumSize
         */
        this.peerBindingCachePushMaximumSize = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PeerBindingCachePushMaximumSize"));
        if (debugPolicy)
            log.println ("Naming configuration - peerBindingCachePushMaximumSize: " + peerBindingCachePushMaximumSize);

        /*
         * PeerBindingCachePullMaximumSize
         */
        this.peerBindingCachePullMaximumSize = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PeerBindingCachePullMaximumSize"));
        if (debugPolicy)
            log.println ("Naming configuration - peerBindingCachePullMaximumSize: " + peerBindingCachePullMaximumSize);

        /*
         * StatsName
         */
        this.statsName = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/StatName"));
        if (debugPolicy)
            log.println ("Naming configuration - statsName: " + statsName);

        /*
         * SnoopBindings
         */
        final String snoopBindingsString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/SnoopBindings"));
        this.snoopBindings = configuration.booleanFromYesNoString (snoopBindingsString);
        if (debugPolicy)
            log.println ("Naming configuration - snoopBindings: " + snoopBindings);

        /*
         * 
         */
        final String cachePulledBindingsString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/CachePulledBindings"));
        this.cachePulledBindings = configuration.booleanFromYesNoString (cachePulledBindingsString);
        if (debugPolicy)
            log.println ("Naming configuration - cachePulledBindings: " + cachePulledBindings);

        /*
         * SnoopCertificates
         */
        final String snoopCertificatesString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/SnoopCertificates"));
        this.snoopCertificates = configuration.booleanFromYesNoString (snoopCertificatesString);
        if (debugPolicy)
            log.println ("Naming configuration - snoopCertificates: " + snoopCertificates);

        /*
         * RequireBindingSerialNumberMatch
         */
        final String requireBindingSerialNumberMatchString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/RequireBindingMessageSerialNumberMatch"));
        this.requireBindingSerialNumberMatch = configuration.booleanFromYesNoString (requireBindingSerialNumberMatchString);
        if (debugPolicy)
            log.println ("Naming configuration - requireBindingSerialNumberMatch: " + requireBindingSerialNumberMatch);

        /*
         * RequireTrustSerialNumberMatch
         */
        final String requireTrustSerialNumberMatchString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/RequireTrustMessageSerialNumberMatch"));
        this.requireTrustSerialNumberMatch = configuration.booleanFromYesNoString (requireTrustSerialNumberMatchString);
        if (debugPolicy)
            log.println ("Naming configuration - requireTrustSerialNumberMatch: " + requireTrustSerialNumberMatch);

        /*
         * tracingQueries
         */
        final String tracingQueriesString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/QueryTracing"));
        this.tracingQueries = configuration.booleanFromYesNoString (tracingQueriesString);
        if (debugPolicy)
            log.println ("Naming configuration - tracingQueries: " + tracingQueries);

        /*
         * queryTracingModulo
         */
        this.queryTracingModulo = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/QueryTracingModulus"));
        if (debugPolicy)
            log.println ("Naming configuration - queryTracingModulo: " + queryTracingModulo);

        /*
         * preallocateFSM
         */
        final String preallocateFSMString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PreallocateFSM"));
        this.preallocateFSM = configuration.booleanFromYesNoString (preallocateFSMString);
        if (debugPolicy)
            log.println ("Naming configuration - preallocateFSM: " + preallocateFSM);

        /*
         * preallocateCrypto
         */
        final String preallocateCryptoString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/PreallocateCrypto"));
        this.preallocateCrypto = configuration.booleanFromYesNoString (preallocateCryptoString);
        if (debugPolicy)
            log.println ("Naming configuration - preallocateCrypto: " + preallocateCrypto);

        /*
         * bypassPeerBindingsCacheOnQuery
         */
        final String bypassPeerBindingsCacheOnQueryString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/BypassPeerBindingsCacheOnQuery"));
        this.bypassPeerBindingsCacheOnQuery = configuration.booleanFromYesNoString (bypassPeerBindingsCacheOnQueryString);
        if (debugPolicy)
            log.println ("Naming configuration - bypassPeerBindingsCacheOnQuery: " + bypassPeerBindingsCacheOnQuery);

        /*
         * certificateCaching
         */
        final String certificateCachingString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/CertificateCaching"));
        this.certificateCaching = configuration.booleanFromYesNoString (certificateCachingString);
        if (debugPolicy)
            log.println ("Naming configuration - certificateCaching: " + certificateCaching);

        /*
         * writeCertificatesToBackingStore
         */
        final String writeCertificatesToBackingStoreString  = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/WriteCertificatesToBackingStore"));
        this.writeCertificatesToBackingStore = configuration.booleanFromYesNoString (writeCertificatesToBackingStoreString);
        if (debugPolicy)
            log.println ("Naming configuration - writeCertificatesToBackingStore: " + writeCertificatesToBackingStore);

        /*
         * bindingsStickyTrust
         */
/*
        final String bindingsStickyTrustString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/BindingsStickyTrust"));
        this.bindingsStickyTrust = configuration.booleanFromYesNoString (bindingsStickyTrustString);
        if (debugPolicy)
            log.println ("Naming configuration - bindingsStickyTrust: " + bindingsStickyTrust);
*/

        // TODO(EXPERIMENTS): remove

        /*
         * queryFrequency
         */
        this.queryFrequency = configuration.getNonNegativeLongAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/QueryFrequency"));
        if (debugPolicy)
            log.println ("Naming configuration - queryFrequency: " + queryFrequency);

        /*
         * socketsInNetwork
         */
        this.socketsInNetwork = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/SocketsInNetwork"));
        if (debugPolicy)
            log.println ("Naming configuration - socketsInNetwork: " + socketsInNetwork);

        /*
         * doNamingQueries
         */
        final String doNamingQueriesString = configuration.getTextAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/DoNamingQueries"));
        this.doNamingQueries = configuration.booleanFromYesNoString (doNamingQueriesString);
        if (debugPolicy)
            log.println ("Naming configuration - doNamingQueries: " + doNamingQueries);

        /*
         * experimentQueryCount
         */
        this.experimentQueryCount = configuration.getNonNegativeIntAttribute (XmlUtil.createXPath (ATTRIBUTE_PREFIX + "/Naming/QueryCount"));
        if (debugPolicy)
            log.println ("Naming configuration - experimentQueryCount: " + experimentQueryCount);
    }

    /** @return Whether or not user configured debugging for naming service. */
/*
    public boolean getDebug() { return debug; }
    public void setDebug (final boolean value) { debug = value; }
*/
    
    /** @return Timeout (in milliseconds) on calls to the message store. */
    public long getBlockingAPITimeout() { return timeoutAPI; }
    public void setBlockingAPITimeout (final long value) { timeoutAPI = value; }

    /** @return Timeout (in milliseconds) for bindings. */
    public int getBindingTimeout() { return bindingTimeout; }
    public void setBindingTimeout (final int value) { bindingTimeout = value; }

    /** @return Number of hops that a name mapping travel. */
    public short getPushRadius() { return pushRadius; }
    public void setPushRadius (final short value) { pushRadius = value; }

    /** @return Number of hops that an invalidate name mapping will travel. */
    public short getInvalidateRadius() { return invalidateRadius; }
    public void setInvalidateRadius (final short value) { invalidateRadius = value; }

    /** @return Number of hops that a query message will travel from the query originator. */
    public short getQueryRadius() { return queryRadius; }
    public void setQueryRadius (final short value) { queryRadius = value; }

    /** @return Whether or not a push operation will occur when the name of this socket is set. */
    public boolean doPushOnSetName() { return pushOnSetName; }
    public void setPushOnSetName (final boolean value) { pushOnSetName = value; }

    /** @return Whether or not an invalidate operation will occur when a name is unset at this socket. */
    public boolean doInvalidateOnUnsetName() { return invalidateOnUnsetName; }
    public void setInvalidateOnUnsetName (final boolean value) { invalidateOnUnsetName = value; }

    /** @return Whether or not a push operation will occur when this socket joins an overlay. */
    public boolean doPushOnJoinOverlay() { return pushOnJoinOverlay; }
    public void setPushOnJoinOverlay (final boolean value) { pushOnJoinOverlay = value; }

    /** @return Whether or not an invalidation operation will occur when this socket leaves an overlay. */
    public boolean doInvalidateOnLeaveOverlay() { return invalidateOnLeaveOverlay; }
    public void setInvalidateOnLeaveOverlay (final boolean value) { invalidateOnLeaveOverlay = value; }

    /** @return Whether or not a push operation will occur when the logical address of this socket changes. */
    public boolean doPushOnChangeOfLogicalAddress() { return pushOnChangeOfLogicalAddress; }
    public void setPushOnChangeOfLogicalAddress (final boolean value) { pushOnChangeOfLogicalAddress = value; }

    /** @return The number of milliseconds between name mapping pushes, 0 means no periodic push. */
    public int getPushPeriod() { return pushPeriod; }
    public void setPushPeriod (final int value) { pushPeriod = value; }

    /** @return the digital signature algorithm used to compute and verify signatures */
    public String getDigitalSignatureAlgorithm() { return hashAlgorithm; }
    public void setDigitalSignatureAlgorithm (final String value) { hashAlgorithm = value; }

    /** @return whether or not only trusted responses will be accepted. */
    public boolean getTrustedResponsesOnly() { return trustedResponses; }
    public void setTrustedResponsesOnly (final boolean value) { trustedResponses = value; }

    /** @return whether or not only authoritative responses will be accepted. */
    public boolean getAuthoritativeResponsesOnly() { return authoritativeResponses; }
    public void setAuthoritativeResponsesOnly (final boolean value) { authoritativeResponses = value; }

    /** @return the maximum age of an acceptable binding. */
    public int getMaximumResponseBindingAge() { return maxResponseAge; }
    public void setMaximumResponseBindingAge (final int value) { maxResponseAge = value; }

    /** @return The maximum number of responses per query. */
    public int getMaximumResponsesPerQuery() { return maxResponseCount; }
    public void setMaximumResponsesPerQuery (final int value) { maxResponseCount = value; }

    /** @return The minimum number of hops that a query should make. */
    public short getMinimumQueryHopCount() { return minHopCount; }
    public void setMinimumQueryHopCount (final short value) { minHopCount = value; }

    /** @return The timeout in millisecond to wait for a response to a certificate request. */
    public long getCertificateRequestTimeout() { return certificateRequestTimeout; }
    public void setCertificateRequestTimeout (final long value) { certificateRequestTimeout = value; }

    /** @return The maximum number of bindings permitted in the peer binding cache via push. */
    public int getPeerBindingCachePushMaximumSize() { return peerBindingCachePushMaximumSize; }
    public void setPeerBindingCachePushMaximumSize (final int value) { peerBindingCachePushMaximumSize = value; }
    
    /** @return The maximum number of bindings permitted in the peer binding cache via pull. */
    public int getPeerBindingCachePullMaximumSize() { return peerBindingCachePullMaximumSize; }
    public void setPeerBindingCachePullMaximumSize (final int value) { peerBindingCachePullMaximumSize = value; }
    
    /** @return The name by which the Naming FSM will be known. */
    public String getStatName() { return statsName; }
    public void setStatName (final String value) { statsName = value; }
    
    /** @return true if a pull message destined for another node can be used to satisfy queries for an intermediate node. */
    public boolean snoopBindings() { return snoopBindings; }
    public void setSnoopBindings (final boolean value) { snoopBindings = value; }
    
    /** @return if true, only bindings that are received in a push message are subject to caching. */
    public boolean cachePulledBindings() { return cachePulledBindings; }
    public void setCachePulledBindings (final boolean value) { cachePulledBindings = value; }
    
    /** @return true if a certificate response message destined for another node can be used populate the certificate cache of an intermediate node. */
    public boolean snoopCertificates() { return snoopCertificates; }
    public void setSnoopCertificates (final boolean value) { snoopCertificates = value; }
    
    /** @return If true, binding response serial numbers must match binding query serial numbers. */
    public boolean getRequireBindingMessageSerialNumberMatch() { return requireBindingSerialNumberMatch; }
    public void setRequireBindingMessageSerialNumberMatch (final boolean value) { requireBindingSerialNumberMatch = value; }

    /** @return If true, trust response serial numbers must match trust query serial numbers. */
    public boolean getRequireTrustMessageSerialNumberMatch() { return requireTrustSerialNumberMatch; }
    public void setRequireTrustMessageSerialNumberMatch (final boolean value) { requireTrustSerialNumberMatch = value; }

    /** Determines if any queries are traced. */
    public boolean getTracingQueries() { return tracingQueries; }
    public void setTracingQueries (final boolean value) { tracingQueries = value; }

    /** Determines which queries are traced. Example: set to 500 to have every 500th query traced. */
    public int getQueryTracingModulo() { return queryTracingModulo; }
    public void setQueryTracingModulo (final int value) { queryTracingModulo = value; }

    /** Determines if the naming FSM is allocated before receiving a naming message. */
    public boolean getPreallocateFSM() { return preallocateFSM; }
    public void setPreallocateFSM (final boolean value) { preallocateFSM = value; }

    /** Determines if the naming FSM crypto structures are allocated before receiving a naming message requiring a crypto opertion. */
    public boolean getPreallocateCrypto() { return preallocateCrypto; }
    public void setPreallocateCrypto (final boolean value) { preallocateCrypto = value; }

    /**
     * Determines if the peer binding cache is inspected for a query. Similar in concept to authoritative responses, but note
     * that authoritative responses can be cached as authoritative responses.
     */
    public boolean getBypassPeerBindingsCacheOnQuery() { return bypassPeerBindingsCacheOnQuery; }
    public void setBypassPeerBindingsCacheOnQuery (final boolean value) { bypassPeerBindingsCacheOnQuery = value; }

    /**
     * Determines if certificate exchange operations are forced for each step of a trust chain or if certs already in the
     * cert cache are considered. This really doesn't turn off cert caching, it determines when it is used.
     *
     * Trust chains can be extended by either using information (i.e. certificates) in the certificate cache or by
     * information requested from network peers. If this contant is false then all information used to extend trust chains
     * must come from a network peer - effectively this disables certificate caching. (not really though, certificates are
     * still cached, otherwise trust chains could not be built. The idea is that the "next step" of any particular chain of
     * trust must come from a new network operation, not from the cert cache.)
     */
    public boolean getCertificateCaching() { return certificateCaching; }
    public void setCertificateCaching (final boolean value) { certificateCaching = value; }

    /**
     * Determines if certificates are written to a backing store. If false, only an in-memory copy will be saved. If true,
     * <code>SoftReferences</code> are used to control memory usage.
     */
    public boolean getWriteCertificatesToBackingStore() { return writeCertificatesToBackingStore; }
    public void setWriteCertificatesToBackingStore (final boolean value) { writeCertificatesToBackingStore = value; }

    /**
     * Sticky trust means that a binding that is determined to be trusted will remain trusted without any further crypto
     * operation. If this constant is false, the trust of a binding will be recomputed each time it is needed.
     */
/*
    public boolean getBindingsStickyTrust() { return bindingsStickyTrust; }
    public void setBindingsStickyTrust (final boolean value) { bindingsStickyTrust = value; }
*/

    // TODO(EXPERIMENTS): remove

    /** @return queries per second in the experiment - aggregate, not per socket. */
    public long getEXPERIMENTqueryFrequency() { return queryFrequency; }
    public void setEXPERIMENTqueryFrequency (final long value) { queryFrequency = value; }
    
    /**
     * @return the number of sockets an experimental network - this
     * determines the names that are chosen randomly, e.g. 0000-0009
     * if the network is of size 10.
     */
    public int getEXPERIMENTsocketsInNetwork() { return socketsInNetwork; }
    public void setEXPERIMENTsocketsInNetwork (final int value) { socketsInNetwork = value; }

    /** @return whether or not real naming traffic is generated, if false only "Hello" broadcasts will be sent. */
    public boolean getEXPERIMENTdoNamingQueries() { return doNamingQueries; }
    public void setEXPERIMENTdoNamingQueries (final boolean value) { doNamingQueries = value; }

    /** Determines the number of queries for an experiment */
    public int getEXPERIMENTQueryCount() { return experimentQueryCount; }
    public void setEXPERIMENTQueryCount (final int value) { experimentQueryCount = value; }
}
