/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.I_PhysicalAddress;
import hypercast.I_LogicalAddress;
import hypercast.I_AddressPair;
import hypercast.IP_Address;
import hypercast.Pastry.Pastry_Node_BuddyList;

import hypercast.Pastry.rice.pastry.messaging.MessageReceiver;
import hypercast.Pastry.rice.pastry.hc.HCNodeHandle;
import hypercast.Pastry.rice.pastry.NodeHandle;
import hypercast.Pastry.rice.pastry.PastryNode;
import hypercast.Pastry.rice.pastry.messaging.Message;
import hypercast.Pastry.rice.pastry.Id;
import hypercast.Pastry.rice.pastry.NodeId;
import hypercast.Pastry.rice.pastry.leafset.LeafSet;
import hypercast.Pastry.rice.pastry.routing.RouteSet;
import hypercast.Pastry.rice.pastry.routing.RoutingTable;

import java.util.Vector;
import java.util.HashMap;
import java.io.IOException;

/**
 *  This class implements an interface between the native Pastry (Rice
 *  University) network mechanism and the HyperCast network mechanism.
 *
 *  @author HyperCast Team
 *  @author Greg Mattes
 *  @version May 2004
 */
public class HyperCastNodeAdapter
{
    private final static boolean debug = false;

    private HCNodeHandle nodeHandle;
    private Pastry_Node_BuddyList node;

    public HyperCastNodeAdapter (final Pastry_Node_BuddyList node) {
        this.node = node;
    }

    public void setMessageReceiver (final MessageReceiver nodeHandle) {
        this.nodeHandle = (HCNodeHandle) nodeHandle;
    }

    public I_PhysicalAddress getPhysicalAddress() {
        return node.createPhysicalAddress();
    }

    public I_PhysicalAddress createPhysicalAddress (final String physicalAddressString) {
        return node.createPhysicalAddress (physicalAddressString);
    }

    private Pastry_AddressPair makeAddressPair (final I_PhysicalAddress physical, final NodeId logical)
    {
        final Pastry_LogicalAddress hc_logical = new Pastry_LogicalAddress (logical);
        final Pastry_AddressPair pair = new Pastry_AddressPair (physical, hc_logical);
        return pair;
    }

    public void nodeIsReady()
    {
        node.nodeIsReady();
    }

    public long getCurrentTime() {
        return node.getCurrentTime();
    }

    /**
     * Called by the "black-box" when it needs to send a message to another node
     */
    public void sendUnicastMessage (final HyperCastPhysicalAddress dstPhysicalHC,
                                    final NodeId                   dstLogical,
                                    final HyperCastPhysicalAddress srcPhysicalHC,
                                    final NodeId                   srcLogical,
                                    final Message                  msg)
    {
        final I_PhysicalAddress dstPhysical = dstPhysicalHC.getPhysicalAddress (this);
        final I_PhysicalAddress srcPhysical = srcPhysicalHC.getPhysicalAddress (this);

        if (debug)
            System.err.println ("PASTRY DEBUG: HyperCastNodeAdapter.sendUnicastMessage():" +
                                "\n  dstPhysical: " + dstPhysical +
                                "\n  dstLogical: " + dstLogical +
                                "\n  srcPhysical: "+ srcPhysical +
                                "\n  srcLogical: " + srcLogical +
                                "\n  msg: " + msg);

        final Pastry_AddressPair dstPair = makeAddressPair (dstPhysical, dstLogical);
        final Pastry_AddressPair srcPair = makeAddressPair (srcPhysical, srcLogical);

        final HyperCastMessage hc_msg = new HyperCastMessage (msg);

        node.sendUnicastMessage (dstPair, srcPair, hc_msg);
    }

    /**
     * Called to ping a remote node
     */
    public void ping (final HyperCastPhysicalAddress dstPhysicalHC, final NodeId dstLogical) {
        node.ping (dstPhysicalHC, dstLogical);
    }

    /**
     * Called when a remote node does respond to a ping
     */
    public void markAliveWithProximity (final NodeId otherPastryLogicalAddress, final int oneWayDelayApprox)
    {
        final PastryNode pn = nodeHandle.getLocalNode();

        final LeafSet leafSet = pn.getLeafSet();
        final HCNodeHandle lsnh = (HCNodeHandle) leafSet.get (otherPastryLogicalAddress);
        if (lsnh != null)
        {
            if (oneWayDelayApprox > 0)
                lsnh.setProximity (oneWayDelayApprox);

            lsnh.markAlive();
        }

        final RoutingTable routeTable = pn.getRoutingTable();
        final HCNodeHandle rtnh = (HCNodeHandle) routeTable.get (otherPastryLogicalAddress);
        if (rtnh != null)
        {
            if (oneWayDelayApprox > 0)
                rtnh.setProximity (oneWayDelayApprox);

            rtnh.markAlive();
        }
    }

    /**
     * Called when a message is received from a remote node
     */
    public void markAlive (final NodeId otherPastryLogicalAddress) {
        markAliveWithProximity (otherPastryLogicalAddress, -1);
    }

    public synchronized void receiveMessage (final HyperCastMessage msg)
    {
        Message m = msg.getMessage();
        nodeHandle.receiveMessage (m);
    }

  public synchronized byte [] getSerializedRouteRow (int row)
  {
      try
      {
          PastryNode pn = nodeHandle.getLocalNode();
          RoutingTable routeTable = pn.getRoutingTable();
        
          int rows = routeTable.numRows();
          if (row < 0 || row >= rows)
          {
              return null;
          }

          RouteSet [] routerow = routeTable.getRow (row);
          if (routerow == null)
          {
              throw new RuntimeException();
          }
          return HyperCastMessage.serialize (routerow);
      }
      catch (IOException ioe)
      {
          System.err.println ("Caught io exception while trying to serialize leaf set: " + ioe);
          return null;
      }
  }

    public synchronized byte [] getSerializedLeafSet()
    {
        try
        {
            PastryNode pn = nodeHandle.getLocalNode();
            LeafSet leafSet = pn.getLeafSet();
            return HyperCastMessage.serialize (leafSet);
        }
        catch (IOException ioe)
        {
            System.err.println ("Caught io exception while trying to serialize leaf set: " + ioe);
            return null;
        }
    }

    private static final int GET_ALL_NEIGHBORS = -1; // must be less than zero
    private static final int GET_NEIGHBOR_COUNT = -2; // must be less than zero

    public synchronized I_AddressPair [] getAllNeighbors()
    {
        return getNeighbors (GET_ALL_NEIGHBORS, null);
    }

    public synchronized I_AddressPair getNthNeighbor (final int n)
    {
        final I_AddressPair [] neighborArray = getNeighbors (n, null);
        return neighborArray [0];
    }

    public synchronized int getNeighborCount()
    {
        int [] count = new int [1];
        getNeighbors (GET_NEIGHBOR_COUNT, count);
        return count [0];
    }

    private I_AddressPair [] getNeighbors (final int n, final int [] count)
    {
        int neighborCount = 0;
        PastryNode pn = nodeHandle.getLocalNode();

        Vector addresses = new Vector();

        /*
         *  Get all of the neighbors in the leaf set
         */
        LeafSet leafSet = pn.getLeafSet();

    int cwSize = leafSet.cwSize();
    int ccwSize = leafSet.ccwSize();
    
        if (count != null)
        {
            neighborCount += cwSize + ccwSize;
        }
        else
        {
            for (int i=-ccwSize; i<=cwSize; i++) {
                if (i == 0) continue;
        
                final HCNodeHandle nh = (HCNodeHandle) leafSet.get(i);
                if (nh == null || nh.isAlive() == false) continue;
        
                final I_PhysicalAddress pa = nh.getAddress().getPhysicalAddress (this);
                final NodeId nid = nh.getNodeId();

                final Pastry_AddressPair pair = makeAddressPair (pa, nid);

                if (! addresses.contains (pair))
                {
                    if (n == GET_ALL_NEIGHBORS || neighborCount == n)
                    {
                        if (debug)
                            System.err.println ("get neighbors, adding leaf: " + pair);

                        addresses.add (pair);
                    }
                    ++neighborCount;
                }
            }
        }

        /*
         *  Get all of the neighbors in the routing table
         */
        RoutingTable routeTable = pn.getRoutingTable();
        
        int rows = routeTable.numRows();
        int cols = routeTable.numColumns();

        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < cols; c++)
            {
                RouteSet rs = routeTable.getRouteSet (r, c);
                if (rs != null)
                {
                    int size = rs.size();
                    for (int s = 0; s < size; s++)
                    {
                        HCNodeHandle nh = (HCNodeHandle) rs.get (s);

                        I_PhysicalAddress pa = nh.getAddress().getPhysicalAddress (this);
                        NodeId nid = nh.getNodeId();

                        Pastry_AddressPair pair = makeAddressPair (pa, nid);

                        if (! addresses.contains (pair))
                        {
                            if (n == GET_ALL_NEIGHBORS || neighborCount == n)
                            {
                                if (debug)
                                    System.err.println ("get neighbors, adding route table entry: " + pair);

                                addresses.add (pair);
                            }
                            ++neighborCount;
                        }
                    }
                }
            }
        }

        if (count != null)
        {
            count [0] = neighborCount;
            return new I_AddressPair [0];
        }
        else
        {
            /*
             *  Convert the vector of addresses to an array for I_AddressPairs
             */
            int arrayLength = addresses.size();
        
            if (arrayLength == 0)
            {
                return new I_AddressPair [0];
            }
            else
            {
                I_AddressPair [] pairs = new I_AddressPair [arrayLength];

                for (int i = 0; i < arrayLength; ++i)
                {
                    pairs [i] = (Pastry_AddressPair) addresses.get (i);
                }
                return pairs;
            }
        }
    }

    public synchronized I_AddressPair [] getParent (final I_LogicalAddress root)
    {
        if (! (root instanceof Pastry_LogicalAddress))
            throw new RuntimeException ("in getParent, root is not a Pastry_LogicalAddress");

        final Pastry_LogicalAddress pla = (Pastry_LogicalAddress) root;
        final Id rootId = pla.getNodeId();

        return new I_AddressPair [] { computeNextHop (rootId) };
    }

    /**
     *  This code is from rice/pastry/standard/StandardRouter.java
     */
    private I_AddressPair computeNextHop (Id target)
    {
    if (target == null)
        {
            throw new RuntimeException ("Cannot computer next path to null!");

            // StandardRouter
            //target = localId;
        }

        // The purpose of this method is to compute these two values
        I_PhysicalAddress nextHopPhysical = null;
        NodeId nextHopLogical = null;

        PastryNode pn = nodeHandle.getLocalNode();
        LeafSet leafSet = pn.getLeafSet();
        RoutingTable routeTable = pn.getRoutingTable();

    int cwSize = leafSet.cwSize();
    int ccwSize = leafSet.ccwSize();

    int lsPos = leafSet.mostSimilar(target);

    if (lsPos == 0) // message is for the local node so deliver it
        {
            // StandardRouter
        //msg.nextHop = localHandle;

            nextHopPhysical = nodeHandle.getAddress().getPhysicalAddress (this);
            nextHopLogical = nodeHandle.getNodeId();
        }
    else if ( (lsPos>0 && (lsPos<cwSize || !leafSet.get(lsPos).getNodeId().clockwise(target))) ||
          (lsPos<0 && (-lsPos<ccwSize || leafSet.get(lsPos).getNodeId().clockwise(target))) )
        // the target is within range of the leafset, deliver it directly 
        {
        HCNodeHandle handle = (HCNodeHandle) leafSet.get(lsPos);

        if (handle.isAlive() == false) {   // node is dead - get rid of it and try again
            leafSet.remove(handle.getNodeId());

                    // StandardRouter
            //receiveRouteMessage(msg);
            return computeNextHop (target);
        } else {
                    // StandardRouter
            //msg.nextHop = handle;

                    nextHopPhysical = handle.getAddress().getPhysicalAddress (this);
                    nextHopLogical = handle.getNodeId();
        }
        }
    else {
        // use the routing table
        RouteSet rs = routeTable.getBestEntry(target);
        HCNodeHandle handle = null;

        if (rs == null ||
        // get the closest alive node
        (handle = (HCNodeHandle) rs.closestNode()) == null) {
        
        // no live routing table entry matching the next digit
        // get best alternate RT entry

        handle = (HCNodeHandle) routeTable.bestAlternateRoute(target);

        if (handle == null) {
            // no alternate in RT, take leaf set
            handle = (HCNodeHandle) leafSet.get(lsPos);

            if (handle.isAlive() == false) {
            leafSet.remove(handle.getNodeId());

                        // StandardRouter
            //receiveRouteMessage(msg);
                        return computeNextHop (target);
            }
        } else {
            NodeId.Distance altDist = handle.getNodeId().distance(target);
            NodeId.Distance lsDist = leafSet.get(lsPos).getNodeId().distance(target);

            if (lsDist.compareTo(altDist) < 0) {
            // closest leaf set member is closer
            //System.out.println("forw to edge leaf set member, alt=" + handle.getNodeId() + 
            //" lsm=" + leafSet.get(lsPos).getNodeId());
            handle = (HCNodeHandle) leafSet.get(lsPos);

            if (handle.isAlive() == false) {
                leafSet.remove(handle.getNodeId());

                            // StandardRouter
                // receiveRouteMessage(msg);
                            return computeNextHop (target);
            }
            }
        }
        }
        else {
        // we found an appropriate RT entry, check for RT holes at previous node

                /*
                 * Can't do this here: in the Pastry StandardRouter
                 * code, there is a mechanism that allows the routing
                 * table of the node that send a message to be updated
                 * if the receiving node determines that it has
                 * information that can patch the routing table at the
                 * previous node. In this case we are not processing a
                 * route message, we are simply doing a lookup, so
                 * there is no "previous node" in this context
                 */

                // StandardRouter 
        //checkForRouteTableHole(msg, handle);
        }

            // StandardRouter
        // msg.nextHop = handle;

            nextHopPhysical = handle.getAddress().getPhysicalAddress (this);
            nextHopLogical = handle.getNodeId();
    }

        // StandardRouter
    //msg.setPrevNode(localHandle);
    //localHandle.receiveMessage(msg);

        return makeAddressPair (nextHopPhysical, nextHopLogical);
    }

    public HCNodeHandle [] constructNodeHandleArrayFromSerializedLeafSet (byte [] serializedLeafSet)
    {
        try
        {
            LeafSet leafSet = (LeafSet) HyperCastMessage.deserialize (serializedLeafSet);

            Vector handles = new Vector();

            for (int i = 1; i <= leafSet.cwSize() ; i++)
                handles.add(leafSet.get(i));

            for (int i = -leafSet.ccwSize(); i < 0; i++)
                handles.add(leafSet.get(i));

            return (HCNodeHandle []) handles.toArray(new HCNodeHandle [0]);
        }
        catch (IOException ioe)
        {
            System.err.println ("Caught io exception while trying to deserialize leaf set: " + ioe);
            return null;
        }
    }

    public HCNodeHandle [] constructNodeHandleArrayFromSerializedRouteRow (byte [] serializedRouteRow)
    {
        try
        {
            RouteSet [] routeSets = (RouteSet []) HyperCastMessage.deserialize (serializedRouteRow);

            Vector handles = new Vector();

            for (int i=0 ; i<routeSets.length ; i++) {
                RouteSet set = routeSets[i];

                if (set != null) {
                    for (int j=0; j<set.size(); j++)
                        handles.add(set.get(j));
                }
            }

            return (HCNodeHandle[]) handles.toArray(new HCNodeHandle[0]);
        }
        catch (IOException ioe)
        {
            System.err.println ("Caught io exception while trying to deserialize route row: " + ioe);
            return null;
        }
    }
}
