/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.I_UnicastAdapter;
import hypercast.Pastry.HyperCastScheduledMessage;

/**
 *  This class implements an interface between the native Pastry (Rice
 *  University) timer mechanism and the HyperCast timer mechanism found
 *  in the node adapter.
 *
 *  Pastry uses java.util.Timer (Java 1.3), whereas HyperCast
 *  implements its own Timer class (in the adapter directory). Both do
 *  essentially the same thing, but there are at least three major
 *  differences:
 *
 *  (1) HyperCast's "Timer" class, found in adapters/Timer.java,
 *  contains no API that allows a programmer to specify that an event
 *  should continue to occur periodically, rather the burden is on the
 *  programmer, when handling a timer event via the timerExpired
 *  method, to reschedule that event manually if it needs
 *  rescheduling. In contrast, Pastry schedules periodic events by
 *  calling the java.util.Timer APIs "schedule" and
 *  "scheduleAtFixedRate." These two APIs do slightly different
 *  things, one is "fixed-delay" and the other is "fixed-rate" (See
 *  the Javadoc for java.util.Timer).
 *
 *  The HyperCastTimer class contains two APIs that support
 *  "fixed-delay" and "fixed-rate" event scheduling. There is also a
 *  timerExpired method that knows how to reschedule previously
 *  scheduled HyperCastScheduledMessages. This timerExpired method
 *  should be called when the timerExpired method of the I_Node
 *  interface that is implemented by all node types in HyperCast is
 *  called with a HyperCastScheduledMessage. Making such a call
 *  ensures that the event is rescheduled properly if it was defined
 *  by Pastry to be periodic.
 *
 *  (2) The HyperCastTimer does not run in its own thread as does the
 *  Pastry timer (since Pastry uses java.util.Timer). Instead, the
 *  HyperCastTimer class acts as an interface to the timer thread that
 *  runs in the node adapter. The HyperCastTimer class ensures that
 *  Pastry timer events occur in the node adapter timer thread by
 *  delegating method calls to the node adapter.
 *
 *  (3) The Java Timer class (used by Pastry) works with Java
 *  TimerTask objects rather than general Objects, as is the case with
 *  HyperCast's Timer class. TimerTask objects have associated with
 *  them a run method (this does not imply that TimerTask is derived
 *  from Runnable, i.e. a TimerTask does not have its own thread,
 *  rather it uses the java.util.Timer thread). This run method must
 *  be invoked in order for the TimerTask registered with the Timer to
 *  do its work.
 *
 *  In HyperCast no such run method is attached to the objects that
 *  are registered with the setTimer method (although a programmer can
 *  implement such a mechanism as has been done in
 *  Pastry_Node_BuddyList.java). Instead, the programmer writes a
 *  handler that does whatever is appropriate with the Object that is
 *  passed to the timerExpired method of the I_AdapterCallback
 *  interface. Said another way: the code run by a HyperCast timerID
 *  object is not usually stored with the object, but rather with a
 *  separate handler method, whereas a TimerTask has a built in
 *  handler method called "run."  The HyperCastTimer method
 *  timerExpired takes care of invoking the run method of a Pastry
 *  TimerTask (Pastry extends "TimerTask" to be a "ScheduledMessage,"
 *  there is another class called "HyperCastScheduledMessage" that is
 *  extended from "ScheduledMessage" which is used in this interface.
 *
 * @see java.util.Timer
 *
 *  @author HyperCast Team
 *  @author Greg Mattes
 *  @version May 2004
 */

public class HyperCastTimer
{
    /**
     * HyperCast adapter that has a timer thread that we will use.
     */
    private I_UnicastAdapter unicastAdapter;

    private boolean allTimersCancelled = false;

    public HyperCastTimer (final I_UnicastAdapter unicastAdapter) {
        this.unicastAdapter = unicastAdapter;
    }

    public long getCurrentTime() {
        return unicastAdapter.getCurrentTime();
    }

    public synchronized void cancel() {
        allTimersCancelled = true;
    }

    /**
     * Schedule a Pastry message to occur after a timeout, possibly at a regular interval.
     *
     * @see java.util.Timer
     */
    public void scheduleAtFixedRate (final HyperCastScheduledMessage msg)
    {
        msg.setScheduledFixedRate (this);
        unicastAdapter.setTimer (msg, msg.getDelay());
    }

    /**
     * Schedule a Pastry message to occur after a timeout, possibly at a regular interval.
     *
     * @see java.util.Timer
     */
    public void schedule (final HyperCastScheduledMessage msg)
    {
        msg.setScheduled (this);
        unicastAdapter.setTimer (msg, msg.getDelay());
    }

    /**
     * Unschedule a previously scheduled message.
     */
    public void clearTimer (final HyperCastScheduledMessage msg) {
        unicastAdapter.clearTimer (msg);
    }

    /**
     * Must be called from implementation of timerExpired of I_AdapterCallback object.
     */
    public synchronized void timerExpired (final HyperCastScheduledMessage msg)
    {
        if (msg.isCancelled()
            ||
            allTimersCancelled)
            return;

        if (! msg.isFixedRate())
            msg.run();

        if (msg.shouldReschedule())
            unicastAdapter.setTimer (msg, msg.getPeriod());

        if (msg.isFixedRate())
            msg.run();
    }
}
