/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast.Pastry;

import hypercast.I_Message;
import hypercast.I_PhysicalAddress;
import hypercast.ByteArrayUtility;
import hypercast.I_UnicastAdapter;
import hypercast.Pastry.Pastry_AddressPair;
import hypercast.Pastry.Pastry_LogicalAddress;

import java.io.IOException;

/** This class encapsulates a Pastry protocol message.
 * The message contains a type (byte), a Overlay hash(int) and 4 AddressPairs. 
 * The Src and Dst AddressPairs can not both be null.
 *<p>
 * It is immutable.
 *
 * 
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 *
 * @author Greg Mattes
 * @version May 2004 
 */

class Pastry_Message
    implements I_Message
{
    static final boolean debug = false;

    /*
     *  In the file rice/pastry/wire/WirePastryNodeFactory.java there
     *  are several "command messages" that are exchanged. The code in
     *  that file is rather straight-forward since it uses
     *  bi-directional communications channels. HyperCast does not
     *  have support for such channels and so the mechanisms for
     *  exchanging the information of those messages must be
     *  reimplemented here and in Pastry_Node_Buddylist.java.
     */
    public static final byte NODE_ID_REQUEST     = 1;
    public static final byte NODE_ID_RESPONSE    = 2;
    public static final byte LEAFSET_REQUEST     = 3;
    public static final byte LEAFSET_RESPONSE    = 4;
    public static final byte ROUTEROW_REQUEST    = 5;
    public static final byte ROUTEROW_RESPONSE   = 6;
    public static final byte PASTRY_ENCAPSULATED = 7;
    
    public static String getStringOfType (int x)
    {
        switch (x)
        {
            case NODE_ID_REQUEST:     return "NODE_ID_REQUEST";
            case NODE_ID_RESPONSE:    return "NODE_ID_RESPONSE";
            case LEAFSET_REQUEST:     return "LEAFSET_REQUEST";
            case LEAFSET_RESPONSE:    return "LEAFSET_RESPONSE";
            case ROUTEROW_REQUEST:    return "ROUTEROW_REQUEST";
            case ROUTEROW_RESPONSE:   return "ROUTEROW_RESPONSE";
            case PASTRY_ENCAPSULATED: return "PASTRY_ENCAPSULATED";
            default: return "UNKNOWN TYPE " + x;
        }
    }

    /** Protocol Number value in the Protocol field of a Pastry3.0 protocol message */
    public static final byte PROTONUM_Pastry = 4; // HC:1 DT:2 SPT:3

    protected byte Proto_num;
    protected byte Proto_sub;

    /** The protocol message type. */
    private byte type;

    /** A hash of Overlay properties.  It is used to detect either
      conflicting Overlays or detect bad packets. */
    private int overlayHash;

    /** Logical and Physical address of the sender. */
    private Pastry_AddressPair Src;

    /** Logical and Physical address of the destination. */
    private Pastry_AddressPair Dst;

    /** encapsulated Pastry protocol message */
    private byte [] payload;

    private HyperCastMessage hcMessage;

    public Pastry_Message (byte proto_sub,
                           int type, 
                           int overlayhash,
                           Pastry_AddressPair Src,
                           Pastry_AddressPair Dst,
                           HyperCastMessage hcMessage)
    {
        this.hcMessage = hcMessage;
        init (proto_sub, type, overlayhash, Src, Dst, hcMessage.toByteArray());
    }

    public Pastry_Message (byte proto_sub,
                           int type, 
                           int overlayhash,
                           Pastry_AddressPair Src,
                           Pastry_AddressPair Dst,
                           byte [] payload)
    {
        init (proto_sub, type, overlayhash, Src, Dst, payload);
    }

    private void init (byte proto_sub,
                       int type, 
                       int overlayhash,
                       Pastry_AddressPair Src,
                       Pastry_AddressPair Dst,
                       byte [] payload)
    {
        this.Proto_num = PROTONUM_Pastry;
        this.Proto_sub = proto_sub;
        this.type = (byte) type;
        this.overlayHash = overlayhash;

        this.Src = Src;
        this.Dst = Dst;
        if (Src == null && Dst == null)
        {
            throw new IllegalArgumentException ("Both Src and Dst were null in Pastry_Message constructor!");
        }

        this.payload = payload;

        if (type == PASTRY_ENCAPSULATED)
        {
            try {
                this.hcMessage = new HyperCastMessage (payload);
            }
            catch (final IOException ioe) {
                System.err.println ("Could not reconstruct message in Pastry_Message constructor - will drop!");
            }
        }
    }

    public HyperCastMessage getHyperCastMessage() {
        return hcMessage;
    }

    /**
     * Creates a message, if possible, from a byte array.
     * If a message is read, the value of beginValidBytes is incremented to the index
     * of the first byte not in the message.
     * int [] beginValidBytes is not really an array, just an int passed by reference
     * @param adapter used to create new PhysicalAddress from the byte array.
     */
    public static Pastry_Message restoreMessage (byte [] array,
                                                 int [] beginValidBytes,
                                                 int endValidBytes,
                                                 I_UnicastAdapter adapter,
                                                 int thisOverlayHash)
    {
        Pastry_Message result = null;

        final int paSize = adapter.getPhysicalAddressSize();
        final int laSize = Pastry_LogicalAddress.getSizeStatic();
        final int pairSize = paSize + laSize;

        int index = beginValidBytes[0];

        if (endValidBytes - beginValidBytes[0] < 4)
        {
            return null;
        }

        try
        {
            byte proto_num = (byte) ((array[index] >> 4) & 0xf);
            byte proto_sub = (byte) (array[index] & 0xf);
            index += 1;

            if (proto_num != PROTONUM_Pastry)
            {
                throw new IllegalArgumentException("Pastry_Message.restoreMessage: protocol umber not PROTONUM_Pastry(" + PROTONUM_Pastry + ") instead it is: " + proto_num);
            }

            short length = ByteArrayUtility.toShort (array, index);
            index += 2;

            if (endValidBytes - beginValidBytes[0] < length)
            {
                return null;
            }

            int type = array[index];
            index += 1;

            int incomingOverlayHash = ByteArrayUtility.toInteger (array, index);
            index += 4;
        
            if (incomingOverlayHash != thisOverlayHash)
            {
                throw new IllegalArgumentException("Message is not of this Overlay(" + thisOverlayHash + ") instead: " + incomingOverlayHash);
            }

            Pastry_AddressPair src = deserializeAddressPair (array, index, paSize, laSize, adapter);
            index += pairSize;
            Pastry_AddressPair dst = deserializeAddressPair (array, index, paSize, laSize, adapter);
            index += pairSize;

            int payloadSize = ByteArrayUtility.toInteger (array, index);
            index += 4;

            byte [] payload = null;
            if (payloadSize > 0)
              {
                payload = new byte [payloadSize];
                System.arraycopy (array, index, payload, 0, payloadSize);
                String payloadString = new String (payload);
              }
            index += payloadSize;

            beginValidBytes[0] = index;
            result = new Pastry_Message (proto_sub, type, incomingOverlayHash, src, dst, payload);
        }
        catch (IndexOutOfBoundsException iobe)
        {
            System.err.println ("IndexOutOfBoundsException when restoring Pastry Message");
            result = null;
        }
        catch ( ArrayStoreException  ase)
        {
            System.err.println ("ArrayStoreException when restoring Pastry Message");
            result = null;
        }
        return result;
    }

    /**
     * Converts the message into a byte array for transmission.
     */
    public byte [] toByteArray () 
    {
        final int paSize = (Src == null) ? Dst.getPhysicalAddress().getSize() : Src.getPhysicalAddress().getSize();
        final int laSize = Pastry_LogicalAddress.getSizeStatic();
        final int pairSize = paSize + laSize;
        final int payloadLength = (payload == null) ? 0 : payload.length;
        final int resultSize =  1 + // protocol
                                2 + // length
                                1 + // type
                                4 + // overlay hash
                         pairSize + // src address pair
                         pairSize + // dst address pair
                                4 + // payload length
                    payloadLength;  // payload

        byte [] Result = new byte [resultSize];
        int index = 0;
        
        try
        {
            byte protocol = (byte) ((Proto_num << 4) | (Proto_sub & 0xf));
            Result[index] = protocol;
            index += 1;
    
            short length = (short) (1 + 4 + (2 * pairSize) + payloadLength);
            System.arraycopy(ByteArrayUtility.toByteArray(length), 0, Result, index, 2);
            index += 2;

            Result[index] = type;
            index += 1;
        
            System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
            index += 4;

            index += serializeAddressPair (Src, Result, index, paSize, laSize);
            index += serializeAddressPair (Dst, Result, index, paSize, laSize);
        
            System.arraycopy (ByteArrayUtility.toByteArray (payloadLength), 0, Result, index, 4);
            index += 4;

            if (payload != null)
            {
                System.arraycopy (payload, 0, Result, index, payloadLength);
            }
        }
        catch (IndexOutOfBoundsException iobe)
        {
            System.err.println (iobe);
            Result = null;
        }
        catch (ArrayStoreException ase)
        {
            System.err.println (ase);
            Result = null;
        }
        return Result;
    }

    /**
     * Writes the bytes representing a physical/logical address pair
     * into the array Result.
     *
     * Returns the number of bytes written to array Result
     */
    private int serializeAddressPair (Pastry_AddressPair pair, byte [] Result, int offset, int paSize, int laSize)
    {
        int index = offset;
        if (pair != null) 
        {
            System.arraycopy (pair.getPhysicalAddress().toByteArray(), 0, Result, index, paSize);
            index += paSize;
            Pastry_LogicalAddress pairLogical = pair.getPastryLogicalAddress();
            if (pairLogical == null)
            {
                ByteArrayUtility.setBytesZero (Result, index, laSize);
                index += laSize;
            }
            else
            {
                byte [] logicalAddressBytes = pairLogical.toByteArray();
                if (logicalAddressBytes.length != laSize)
                {
                    throw new RuntimeException ("logicalAddressBytes.length != expectedSize");
                }
                System.arraycopy (logicalAddressBytes, 0, Result, index, laSize);
                index += laSize;
            }
        }
        else
        {
            ByteArrayUtility.setBytesZero (Result, index, laSize + paSize);
            index += laSize + paSize;
        }
        return index - offset;
    }

    private static Pastry_AddressPair deserializeAddressPair (byte [] array, int index, final int paSize, final int laSize, I_UnicastAdapter adapter)
    {
        Pastry_AddressPair result = null;
        if (! ByteArrayUtility.isBytesZero (array, index, paSize + laSize))
        {
            I_PhysicalAddress pa = adapter.createPhysicalAddress (array, index);
            byte [] logicalAddressBytes = new byte [laSize];
            System.arraycopy (array, index + paSize, logicalAddressBytes, 0, laSize);
            Pastry_LogicalAddress la = new Pastry_LogicalAddress (logicalAddressBytes);
            result = new Pastry_AddressPair (pa, la);
        }
        return result;
    }
        
    public String getLabel() {
        return getStringOfType (type);
    }

    public byte getProtoNum() { return Proto_num; }
    public byte getProtoSub() { return Proto_sub; }
    public byte getType() { return type; }
    public int getOverlayHash() { return overlayHash; }
    public byte [] getPayload() { return payload; }

    public I_PhysicalAddress getSrcPA () {
        return (Src == null) ? null : Src.getPhysicalAddress();
    }
  
    public I_PhysicalAddress getDstPA () {
        return (Dst == null) ? null : Dst.getPhysicalAddress();
    }
  
    public Pastry_LogicalAddress getSrcLA() {
        return (Src == null) ? null : Src.getPastryLogicalAddress();
    }
  
    public Pastry_LogicalAddress getDstLA () {
        return (Dst == null) ? null : Dst.getPastryLogicalAddress();
    }

    public Pastry_AddressPair getSrc() { return Src; }
    public Pastry_AddressPair getDst() { return Dst; }

    public String toString()
    {
        if (type == PASTRY_ENCAPSULATED)
        {
            if (payload != null)
            {
                if (hcMessage == null)
                {
                    try {
                        hcMessage = new HyperCastMessage (payload);
                    }
                    catch (final IOException ioe) {
                    }
                }

                if (hcMessage == null)
                    return getStringOfType (type);

                final String fullName = hcMessage.getClass().getName();
                return fullName.substring (fullName.indexOf ('$') + 1);
            }
            else
                return getStringOfType (type);
        }
        else
        {
            return getStringOfType (type);
        }
        //return hypercast.HTTP_ServerUtility.toHexString (this.toByteArray());
    }
}
