/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast;

import java.io.*;

//import java.math.*;
import java.security.*;
import java.security.cert.*;

//import java.security.interfaces.*;
import javax.crypto.*;
import javax.crypto.spec.*;


/**
 * This class defines the format for security information exchange 
 * messages, including key request, ket reply, certificate request
 * and certificate reply. It also provides a wrap message for plaintext
 * protocol messages so that they can be exchanged between different
 * overlay nodes when security policy (privacy or integrity) is applied.
 * The format of a SecInfoExchange_Message instance in byte array format is:
 * 
 * -----------------------------------------------------------------------------
 * |Proto number|Payload Length |   Type     | Sequence Num  | OverlayID hash  |
 * |  8 bits    |    16 bits    |  8 bits    |   32 bits     |    32 bits      |
 * -----------------------------------------------------------------------------
 * |  Src Addr Pair  /  Dest Addr Pair /
 * | LA size+PA size / LA size+PA size /
 * -------------------------------------
 * (only for a wrap message, i.e. type is ProtoMsg)
 * ----------------------------------------------------------------------
 * /Payload Mac Len/  Payload MacType   / Sender key timestamp  /
 * /    16 bits    /       variable     /       64 bits         /
 * --------------------------------------------------------------
 * (except key request message)
 * -------------------
 * /     Payload     /
 * /                 /
 * -------------------
 * The payload is a certificate for certificate request or reply message, 
 * a key timestamp and key for key update message, or the byte array of 
 * plaintext protocol message for a ProtoMsg message.
 * 
 * @author HyperCast Team
 * @author Josh Zaritsky
 * @author Jianping Wang
 * @version 3.0, Sep. 07, 2004
 */
public class SecInfoExchange_Message implements I_Message
{
	/**
	 * Constants for the type of a SecInfoExchange_Message instance.
	 */
	public static final byte CertRequest = 1;
	public static final byte CertReply = 2;
	public static final byte KeyRequest = 3;
	public static final byte KeyUpdate = 4;
	public static final byte ProtoMsg = 5;

	/**
	 * Debug flag.
	 */
	protected boolean debug = false;

	/**
	 * Protocol Number value in Protocol field for SecInfoExchange message
	 */
	public static final byte PROTONUM_SIE30 = (byte) 0xf0;
	
	/**
	 * The certificate carried in this message.
	 */
	protected X509Certificate cert;
	
	/**
	 * The key carried in a key update message.
	 */
	protected SecretKeySpec key;
	
	/**
	 * Timestamp when the key of this message's sender is crerated.
	 * if keyTimestamp==0 then it's using shared key
	 */
	protected long keyTimestamp;
	
	/**
	 * The plaintext protocol message that is wraped by this message.
	 */
	I_Message protoMsg;
	
	/**
	 * The length of payload carried in this message.
	 */
	int payloadLength;
	
	/**
	 * The byte array of the plaintext protocol message that is wraped by this message.
	 */
	byte[] protoMsgBytes;
	
	/**
	 * security processor.
	 */
	protected SecurityProcessor secProc;
	
	/**
	 * overlay node.
	 */
	protected I_Node node;
	
	/**
	 * Key vault object.
	 */
	protected KeyVault vault;
	
	/**
	 * source address pair.
	 */
	protected I_AddressPair src = null;
	
	/**
	 * destination address pair.
	 */
	protected I_AddressPair dst = null;
	
	/**
	 * Type of this message.
	 */
	protected byte Type;
	
	/**
	 * Overlay ID hash value 
	 */
	private int overlayHash;

	/**
	 * Sequence number
	 */
	private int sequenceNum = -1;

	/**
	 * Constructor.
	 * 
	 * @param type	type of this message.
	 * @param gh	overlay ID hash value.
	 * @param _src	source address of this message.
	 * @param _dst	destination address of this message.
	 * @param sp	security processor which handles secure message sending/
	 * 	receiving issues in this overlay socket.
	 */
	
	public SecInfoExchange_Message(int type, int gh, I_AddressPair _src, I_AddressPair _dst, SecurityProcessor sp)
	{
		Type = (byte) type;

		overlayHash = gh;

		secProc=sp;

		/*
		 * Get the overlay node stored in the security processor.
		 */
		node = sp.getNode();
		
		/*
		 * Get the key vault object stored in the security processor.
		 */
		vault = secProc.getKeyVault();
		cert = vault.getMyCertificate();
		/*
		 * Get source address pair.
		 */
		src = _src;
		
		/*
		 * Get destination address pair.
		 */
		dst = _dst;

		/*
		 * In a SecInfoExchange_Message, the source address cannot
		 * be 0.
		 */
		if (src == null) {
			throw new IllegalArgumentException("The source addresses is null when creating a SecInfoExchange_Message instance!");
		}

		/*
		 *  Get the timestamp of the sender's key.
		 */
		//keyTimestamp = vault.getMyKeyTimestamp();
		keyTimestamp = secProc.getMyKeyTimestampFor(dst, src);
		/*
		 * There is no need to set the personal key (variable key) here. Only for KeyUpdate message, 
		 * the key is needed, which is set explicitly from outside by SecInfoExchange_Message.setKey() call.
		 */
	}


	/**
	 *Constructor. It constructs a SecInfoExchange_Message instance from byte
	 *array.
	 *
	 * @param A 	packet, i.e., a SecInfoExchange_Message message in byte array format.
	 * @param beginValidBytes 	start point for reading available data in the packet.
	 * @param adapter 				node adapter.
	 * @param sp 					security processor.
	 * 
	 */
	public SecInfoExchange_Message(byte[] A, int[] beginValidBytes,
								   I_UnicastAdapter adapter, SecurityProcessor sp)
	{
		
		
		
		/*
		 * Get the start position to read data in the packet.
		 */
		int index = beginValidBytes[0];
		
		/*
		 * Set the security processor variable to the given one.
		 */
		secProc=sp;

		/*
		 * Get the overlay node stored in the security processor.
		 */
		node = sp.getNode();

		/*
		 * Get the key vault object stored in the security processor.
		 */
		KeyVault vault = secProc.getKeyVault();
		
		/*
		 * Get the logical address size.
		 */
		int LAsize = node.getAddressPair().getLogicalAddress().getSize();

		/*
		 * Get the physical address size from node adapter. There is no LA size 
		 * or PA size field in the packet.
		 */
		int PAsize = node.getAddressPair().getPhysicalAddress().getSize();

		I_PhysicalAddress pa;
		I_LogicalAddress la;
		
		int headerSize = 1 + 2 + 1 + 4 + 4 + (2 * LAsize) + (2 * PAsize);
		byte []headerBytes = new byte[headerSize];
		System.arraycopy(A, index, headerBytes, 0, headerSize);

		/*
		 * Check protocol number field.
		 */
		byte protocolnum = A[index];
  		if (protocolnum != PROTONUM_SIE30)
  		{
			throw new IllegalArgumentException("The Protocol Number of SecInfoExchange_Message message is not PROTONUM_SIE30.");
  		}
		index += 1;

		/*
		 * Get payload length.
		 */
		byte[] lengthBytes = new byte[4];
		System.arraycopy(A, index, lengthBytes, 2, 2);
		payloadLength = ByteArrayUtility.toInteger(lengthBytes, 0);
		index += 2;

		/*
		 * Get the type of this message.
		 */
		Type = A[index];
		index++;
		
		/* 
		 * Get sequence number.
		 */
		byte[] sequenceNumBytes = new byte[4];
		System.arraycopy(A, index, sequenceNumBytes, 0, 4);
		sequenceNum = ByteArrayUtility.toInteger(sequenceNumBytes, 0);
		index += 4;
		
		
		
		/* 
		 *  Get overlay ID hash.
		 */
		overlayHash = ByteArrayUtility.toInteger(A, index);
		index += 4;

		/* 
		 * Get source address pair which now is an instance of G_AddressPair.
		 * The adapter here is the normal node adapter.
		 */
		if (!ByteArrayUtility.isBytesZero(A, index, PAsize + LAsize)) {
			pa = adapter.createPhysicalAddress(A, index);
			la = node.createLogicalAddress(A, index + PAsize);
			src = secProc.createAddressPair(pa, la);
		}
		index += (PAsize + LAsize);

		/* 
		 * Get destination address pair which now is an instance of G_AddressPair.
		 */
		if (!ByteArrayUtility.isBytesZero(A, index, PAsize + LAsize)) {
			pa = adapter.createPhysicalAddress(A, index);
			la = node.createLogicalAddress(A, index + PAsize);
			dst = secProc.createAddressPair(pa, la);
		}
		index += (PAsize + LAsize);

		/*
		 * If this is a wrap message and security policy for protocol messages
		 * is set to Integrity, check the payload mac.
		 */
		byte[] mac;

		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();
		
		
		
		if (vault.getProtoSecPolicy().equals("integrity") && (Type == ProtoMsg))
		{
			/*
			 * Get payload mac length.
			 */
			byte[] macLengthBytes = new byte[4];
			System.arraycopy(A, index, macLengthBytes, 2, 2);
			int macLength = ByteArrayUtility.toInteger(macLengthBytes, 0);
			index += 2;

			/*
			 * Get payload mac.
			 */
			byte[] msgMac = new byte[macLength];
			System.arraycopy(A, index, msgMac, 0, macLength);
			index += macLength;

			/*
			 * Get the keyTimestamp of the sender.
			 */
			byte[] keyTimestampBytes = new byte[8];
			
			System.arraycopy(A, index, keyTimestampBytes, 0, 8);
			keyTimestamp = ByteArrayUtility.toLong(keyTimestampBytes, 0);
			
			index += 8;

			Mac macCompute;
			Key neighborKey;

			/*
			 * Get the payload byte array.
			 */
			byte[] payloadBytes = new byte[payloadLength];
			System.arraycopy(A, index, payloadBytes, 0, payloadLength);
			
			byte[] bytesForMac = ByteArrayUtility.concatenate(headerBytes, keyTimestampBytes);
			bytesForMac = ByteArrayUtility.concatenate(bytesForMac, payloadBytes);
			//System.out.println("receive mac=" + HTTP_ServerUtility.toHexString(bytesForMac));
			
			/*
			 * If this node has the key of the sender, calculate the payload
			 * mac using the sender's key and compare it with the mac carried
			 * in the packet.
			 */
			
			if (vault.haveCert(srcLA) && (vault.getKeyMode().equals("GroupKeys") || vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY) || vault.haveKey(srcLA)))
			{
				try
				{
					
					/*
					 * Get the key for calculating the payload mac. 
					 */
					if(vault.getKeyMode().equals("GroupKeys") || vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY)){
						neighborKey = vault.getGroupKey();
					}
					else if ((srcLA.equals(node.getAddressPair().getLogicalAddress())) &&
						(srcPA.equals(node.getAddressPair().getPhysicalAddress())))
					{
						neighborKey = vault.getMyPersonalKey();
					}
					else
					{
						neighborKey = vault.getKey(srcLA);
					}
					//System.err.println("Use key to check MAC: " + HTTP_ServerUtility.toHexString(neighborKey.getEncoded()));

					/*
					 * Recalculate the payload mac using sender's key.
					 */
					macCompute = Mac.getInstance(vault.getIntegrityAlgorithm());
					macCompute.init(neighborKey);
					//mac = macCompute.doFinal(payloadBytes);
					mac = macCompute.doFinal(bytesForMac);
					
				} catch (NoSuchAlgorithmException nsae) {
					throw new IllegalArgumentException("Required integrity algorithm is unavailable");
				} catch (InvalidKeyException ike) {
					throw new IllegalArgumentException("The personal key is invalid");
				}

				/*
				 * Compare the recalculated mac with the mac in the packet. If two macs are 
				 * different, check if the key of the sender is out of date and try to get
				 * new key if needed.
				 */
				if (!(HTTP_ServerUtility.toHexString(mac).equals(HTTP_ServerUtility
																 .toHexString(msgMac))))
				{
					if (debug)
					{
						System.out.println("The maclength is:     " + macLength);
						System.out.println("The mac is:           "
										   + HTTP_ServerUtility.toHexString(msgMac));
						System.out.println("The computed mac is:  "
										   + HTTP_ServerUtility.toHexString(mac));
						System.out.println("The sender key is:    "
										   + HTTP_ServerUtility.toHexString(neighborKey
																			.getEncoded()));
						System.out.println("The key timestamp is: "
										   + vault.getKeyTimestamp(srcLA));
						System.out.println("Timestamp in msg is:  "
										   + vault.getKeyTimestamp(srcLA));
					}

					if(!vault.getKeyMode().equals("GroupKeys") && !vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY)){
						/*
						 * Set the key as invalid
						 */
						vault.setKeyInvalid(srcLA, true);
						/*
						 * Triger the key request action.
						 */
						sp.requestKeyFor(srcLA);
						
					}
					
					throw new IllegalArgumentException("Protocol message integrity check failed!");
					
					//System.out.println("Protocol message integrity check failed!");
					
					
				}
				/*
				 * If two macs are the same, make the key valid.
				 */
				else
				{
					vault.setKeyInvalid(srcLA, false);
				}
			}
			/*
			 * If this node doesn't have the key of the sender, do nothing.
			 */
			else
			{
				// not have key
				//throw new IllegalArgumentException("This neighbor has not been authenticated!");
				//System.err.println("Neighbor "+src+" has not been authenticated!");
				//TODO should we do this??
				//sp.keyOutOfDate(srcLA);
			}
		}

		/*
		 * After below update, beginValidBytes[0] points to the
		 * payload of the packet.
		 */
		beginValidBytes[0] = index;
	}
	
	

	/**
	 * Re-construct a SecInfoExchange_Message instance from byte array.
	 *
	 * @param A 	packet, i.e., a SecInfoExchange_Message message in byte array format.
	 * @param beginValidBytes	start point for reading available data in the packet.
	 * @param endValidBytes			end point of available data in the packet.
	 * @param adapter 				node adapter.
	 * @param thisOverlayHash		overlay ID hash.
	 *
	 * @return a SecInfoExchange_Message instance.
	 *
	 * @throws IllegalArgumentException when an error occurs in creating or
	 * 		   decoding security parameters
	 */
	public static SecInfoExchange_Message restoreMessage(byte[] A,
														 int[] beginValidBytes,
														 int endValidBytes,
														 I_UnicastAdapter adapter,
														 int thisOverlayHash,
														 SecurityProcessor sp)
	{
														 	
		//System.out.println("restore message bytesize="+A.length+" from "+beginValidBytes[0]+" to "+endValidBytes);														 	
														 	
		/*
		 * Get the overlay node ketp in the security processor.
		 */
		I_Node node=sp.getNode();

		/*
		 * Get the key vault kept in the security processor.
		 */
		KeyVault vault = sp.getKeyVault();

		/*
		 * Construct a SecInfoExchange_Message from byte array. The message header 
		 * part is reconstructed in the below call.
		 */
		SecInfoExchange_Message msg;
		try{
			msg =
				new SecInfoExchange_Message(A, beginValidBytes, adapter, sp);
		}catch(Exception e){
			e.printStackTrace();
			return null;
		}
			
		/*
		 * beginValidBytes[0] is changed in above call. It points to the payload now.
		 */

		/*
		 * If this is a certificate request or certificate reply message, get the certificate
		 * from the packet.
		 */
		if ((msg.getType() == CertRequest) || (msg.getType() == CertReply))
		{
			
			ByteArrayInputStream rawCertStream =
				new ByteArrayInputStream(A, beginValidBytes[0],	endValidBytes - beginValidBytes[0]);

			X509Certificate cert;
			CertificateFactory certFactory;

			try
			{
				certFactory = CertificateFactory.getInstance("X509");
				cert =
					(X509Certificate) certFactory.generateCertificate(rawCertStream);

				cert.checkValidity();

				cert.verify(vault.getCACertificate().getPublicKey());
				
				/*
				 * Set the certificate variable in the created message.
				 */
				msg.setCertificate(cert);
			}
			catch (CertificateExpiredException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Certificate has expired.");
			}
			catch (CertificateNotYetValidException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Certificate not yet valid.");
			}
			catch (NoSuchProviderException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Specified provider does not exist.");
			}
			catch (SignatureException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Certificate has invalid signature.");
			}
			catch (NoSuchAlgorithmException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Required algorithm does not exist.");
			}
			catch (InvalidKeyException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Invalid key.");
			}
			catch (CertificateException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Invalid certificate.");
			}

			beginValidBytes[0] += msg.getPayloadLength();

			return msg;
		}
		/*
		 * If this is a key update message, get the sender key timestamp and key from
		 * the packet.
		 */
		else if (msg.getType() == KeyUpdate)
		{
			int keyLength = vault.getKeySize();

			int index = beginValidBytes[0];

			/*
			 * Get the sender key timestamp.
			 */
			byte[] keyTimestampBytes = new byte[8];
			System.arraycopy(A, index, keyTimestampBytes, 0, 8);
			msg.setKeyTimestamp(ByteArrayUtility.toLong(keyTimestampBytes, 0));
			index += 8;

			int wrappedKeyLength = msg.getPayloadLength() - 8; //endValidBytes - index;

			byte[] keyBytes = new byte[wrappedKeyLength];
			System.arraycopy(A, index, keyBytes, 0, wrappedKeyLength);

			Cipher unwrapper;
			SecretKeySpec key;
			
			/*
			 * Get the personal key of the message sender. Here my private key is used.
			 * When a key update message is sent out, the key is wraped using the public
			 * key of the destination node.
			 */
			try {
				unwrapper = Cipher.getInstance("RSA", "BC");
				unwrapper.init(Cipher.UNWRAP_MODE, vault.getMyPrivateKey());

				key = (SecretKeySpec) unwrapper.unwrap(keyBytes,
													   vault
													   .getConfidentialityAlgorithm(),
													   Cipher.SECRET_KEY);
			}
			catch (NoSuchAlgorithmException e)
			{
				throw new IllegalArgumentException("Required encryption algorithm is unavailable");
			}
			catch (NoSuchPaddingException e)
			{
				throw new IllegalArgumentException("Required padding algorithm is unavailable");
			}
			catch (InvalidKeyException e)
			{
				throw new IllegalArgumentException("The private key is not valid");
			}
			catch (NoSuchProviderException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("The provider is not valid");
			}
			catch (IllegalStateException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("The private key is not valid");
			}

			/*
			 * Set the key variable in the created message.
			 */
			msg.setKey(key);

			beginValidBytes[0] = index + wrappedKeyLength;

			return msg;
		}
		/*
		 * If this message is a wrap message, get the payload byte array 
		 * , which represents the packet of a plaintext protocol message,
		 * and store it in this message. The plaintext protocol message is
		 * not re-constructed here.
		 */
		else if (msg.getType() == ProtoMsg)
		{
			int index = beginValidBytes[0];
			int payloadLength = msg.getPayloadLength();

			byte[] msgBytes = new byte[payloadLength];
			System.arraycopy(A, index, msgBytes, 0, payloadLength);
			msg.setProtoMsgBytes(msgBytes);

			beginValidBytes[0] = index + payloadLength;

			return msg;
		}
		/*
		 * Otherwise it is a key request message, update the start point in the 
		 * receive buffer.
		 */
		else
		{
			beginValidBytes[0] += msg.getPayloadLength();

			return msg;
		}
	}

	/**
	 * Converts this object to byte array.
	 * @see hypercast.I_Message#toByteArray()
	 */
	public byte[] toByteArray()
	{
		
		/*
		 * Set the sequence number
		 */
		this.setSequenceNum(vault.getNextSequenceNum());
		
		/*
		 * Byte array for entire SecInfoExchange_Message packet.
		 */
		byte[] msgBytes;

		/*
		 * Payload byte array. Payload is the part following the header of 
		 * SecInfoExchange_Message. Different types of SecInfoExchange_Message
		 * have different payload.
		 */
		byte[] payloadBytes;

		byte[] mac = new byte[0];

		int index = 0;
		
		/*
		 * Get the logical address size.
		 */
		int LAsize = node.getAddressPair().getLogicalAddress().getSize();

		/*
		 * Get the physical address size.
		 */
		int PAsize = node.getAddressPair().getPhysicalAddress().getSize();

		/*
		 * The length of message header is calculated as below:
		 *  length = protocol(1 byte) + length(2 bytes) + type(1 byte) + sequenceNum(4 bytes) 
		 * 			 overlayHash(4 bytes) + SrcLA(PAsize+LAsize) + DestLA(PAsize+LAsize)
		 */
		msgBytes = new byte[1 + 2 + 1 + 4 + 4 + (2 * LAsize) + (2 * PAsize)];

		/*
		 * Fill in protocol number.
		 */
		msgBytes[index] = PROTONUM_SIE30;
		index++;

		/*
		 * If this message is certificate request or certificate reply,
		 * the certificate of local user is put into the packet as payload.
		 */
		if ((Type == CertRequest) || (Type == CertReply))
		{
			cert = vault.getMyCertificate();

			/*
			 * Convert the local user certificate into byte array.
			 */
			try
			{
				payloadBytes = cert.getEncoded();
			}
			catch (CertificateEncodingException cee)
			{
				payloadBytes = new byte[0];
			}
		}
		/*
		 * If this message is key update, the key of local user is put 
		 * into the packet as payload.
		 */
		else if (Type == KeyUpdate)
		{
			int keySize = vault.getKeySize();

			Cipher wrapper;
			byte[] wrappedKeyBytes;

			if (dst == null)
			{
				throw new IllegalArgumentException("Destination address is null, the key cannot be wraped with the public key of the destination node.");
			}

			try {
				wrapper = Cipher.getInstance("RSA", "BC");
				wrapper.init(Cipher.WRAP_MODE,
							 vault.getCertificate(dst.getLogicalAddress())
								  .getPublicKey());

				wrappedKeyBytes = wrapper.wrap(key);

				if (debug)
					System.out.println("Length of wrapped key is: "
								   + wrappedKeyBytes.length);
			} catch (NoSuchAlgorithmException e) {
				throw new IllegalArgumentException("Required encryption algorithm is unavailable");
			} catch (NoSuchPaddingException e) {
				throw new IllegalArgumentException("Required padding algorithm is unavailable");
			} catch (IllegalBlockSizeException e) {
				e.printStackTrace();
				throw new IllegalArgumentException("Invalid block size.");
			} catch (InvalidKeyException e) {
				throw new IllegalArgumentException("The private key is not valid");
			} catch (NoSuchProviderException e) {
				e.printStackTrace();
				throw new IllegalArgumentException("The provider is not valid");
			} catch (IllegalStateException e) {
				e.printStackTrace();
				throw new IllegalArgumentException("The private key is not valid");
			}

			/*
			 * keyTimestamp is long number, so the byte array created below is 8 bytes.
			 */
			
			byte[] keyTimestampBytes =
				ByteArrayUtility.toByteArray(keyTimestamp);
				
			/*
			 * The payload contains timestamp of key and key (in byte array).
			 */
			payloadBytes =
				ByteArrayUtility.concatenate(keyTimestampBytes, wrappedKeyBytes);
		}
		/*
		 * If this message is a wrap message, the plaintext protocol message is 
		 * put into the packet as payload.
		 */
		else if (Type == ProtoMsg)
		{
			payloadBytes = protoMsg.toByteArray();

			/*
			 * If the security policy for protocol messages is set to integrity,
			 * calculate the MAC on the payload with personal key of local node.
			 */
			/*
			if (vault.getProtoSecPolicy().equals("integrity"))
			{
				Mac macCompute;

				try
				{
					macCompute = Mac.getInstance(vault.getIntegrityAlgorithm());
					//macCompute.init(vault.getMyPersonalKey());
					macCompute.init(this.secProc.getMyKeyFor(dst, src));
					mac = macCompute.doFinal(payloadBytes);
				} catch (NoSuchAlgorithmException nsae) {
					throw new IllegalArgumentException("Required integrity algorithm is unavailable");
				} catch (InvalidKeyException ike) {
					throw new IllegalArgumentException("My personal key is invalid");
				}
			}
			*/
		}
		/*
		 * Otherwise it is a key request message. No payload.
		 */
		else
		{
			payloadBytes = new byte[0];
		}

		/*
		 * Fill in payload length.
		 */
		int length = payloadBytes.length;
		byte[] lengthBytes;
		lengthBytes = ByteArrayUtility.toByteArray(length);
		System.arraycopy(lengthBytes, 2, msgBytes, index, 2);
		index += 2;

		/*
		 * Fill in message type.
		 */
		msgBytes[index] = Type;
		index++;
		
		/*
		 * Fill in sequence number
		 */
		System.arraycopy(ByteArrayUtility.toByteArray(sequenceNum), 0,
				 msgBytes, index, 4);
		index += 4;

		/*
		 * Fill in overlay ID hash.
		 */
		System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0,
						 msgBytes, index, 4);
		index += 4;

		/*
		 *  If source address pair is not null, convert it to byte array
		 * and put into the packet.
		 */
		if (src != null)
		{
			System.arraycopy(src.getPhysicalAddress().toByteArray(), 0,
							 msgBytes, index, PAsize);

			if (src.getLogicalAddress() != null)
			{
				System.arraycopy(src.getLogicalAddress().toByteArray(), 0,
								 msgBytes, index + PAsize, LAsize);
			}
			else
			{
				ByteArrayUtility.setBytesZero(msgBytes, index + PAsize, LAsize);
			}
		}
		/*
		 * Otherwise, set the field for source address pair to all-zero.
		 */
		else
		{
			ByteArrayUtility.setBytesZero(msgBytes, index, LAsize + PAsize);
		}

		index += (LAsize + PAsize);

		/*
		 *  If destination address pair is not null, convert it to byte array
		 * and put into the packet.
		 */
		if (dst != null) {
			if(dst.getPhysicalAddress()!=null)
			{
				System.arraycopy(dst.getPhysicalAddress().toByteArray(), 0,
							 msgBytes, index, PAsize);
			}
			else
			{ 
				ByteArrayUtility.setBytesZero(msgBytes, index , PAsize);				
			}

			if (dst.getLogicalAddress() != null)
			{
				System.arraycopy(dst.getLogicalAddress().toByteArray(), 0,
								 msgBytes, index + PAsize, LAsize);
			}
			else
			{
				ByteArrayUtility.setBytesZero(msgBytes, index + PAsize, LAsize);
			}
		}
		/*
		 * Otherwise, set the field for source address pair to all-zero.
		 */
		else
		{
			ByteArrayUtility.setBytesZero(msgBytes, index, PAsize + LAsize);
		}

		index += (LAsize + PAsize);

//TODO: Find better way to to this.
		/*
		 * If this is a wrap message, the payload mac, payload mac length 
		 * and sender key timestamp are put into the packet.
		 */
		if ((Type == ProtoMsg)
			&& (vault.getProtoSecPolicy().equals("integrity")))
		{
			

			byte[] keyTimestampBytes =
				ByteArrayUtility.toByteArray(keyTimestamp);
			
			/* compute mac */
			
			/*
			 * Thye MAC is computed on al bytes of the message except the MAC length and MAC fields.
			 */
			byte [] bytesToComputeMac = ByteArrayUtility.concatenate(msgBytes, keyTimestampBytes);
			bytesToComputeMac = ByteArrayUtility.concatenate(bytesToComputeMac, payloadBytes);
				
			Mac macCompute;

			try
			{
				macCompute = Mac.getInstance(vault.getIntegrityAlgorithm());
				//macCompute.init(vault.getMyPersonalKey());
				macCompute.init(this.secProc.getMyKeyFor(dst, src));
				mac = macCompute.doFinal(bytesToComputeMac);
				//System.out.println("send mac=" + HTTP_ServerUtility.toHexString(bytesToComputeMac));
			} catch (NoSuchAlgorithmException nsae) {
				throw new IllegalArgumentException("Required integrity algorithm is unavailable");
			} catch (InvalidKeyException ike) {
				throw new IllegalArgumentException("My personal key is invalid");
			}
			
			/* end of computing mac */
			
			int macLength = mac.length;
			byte[] macLengthBytes = new byte[2];
			byte[] tempBytes = ByteArrayUtility.toByteArray(macLength);
			macLengthBytes[0] = tempBytes[2];
			macLengthBytes[1] = tempBytes[3];
			
			msgBytes = ByteArrayUtility.concatenate(msgBytes, macLengthBytes);
			msgBytes = ByteArrayUtility.concatenate(msgBytes, mac);
			msgBytes =
				ByteArrayUtility.concatenate(msgBytes, keyTimestampBytes);
		}

		byte[] entireMsg = ByteArrayUtility.concatenate(msgBytes, payloadBytes);
		
		return entireMsg;
	}

	/**
	 * Query the type of this SecInfoExchange_Message instance.
	 *
	 * @return The type defined in SecInfoExchange_Message corresponding with
	 * 		   this message
	 */
	public byte getType()
	{
		return Type;
	}

	/**
	 * Get the source address pair of this message.
	 *
	 * @return source address pair.
	 */
	public I_AddressPair getSrc()
	{
		return src;
	}

	/**
	 * Get the overlay ID hash value carried in this object.
	 *
	 * @return the overlay ID hash.
	 */
	public int getOverlayHash()
	{
		return overlayHash;
	}

	/**
	 * Get the certificate stored in this object.
	 *
	 * @return the certificate.
	 */
	public X509Certificate getCertificate()
	{
		return cert;
	}

	/**
	 * Set the certificate variable.
	 *
	 * @param _cert a certificate instance.
	 */
	public void setCertificate(X509Certificate _cert)
	{
		cert = _cert;
	}

	/**
	 * Get the key stored in this object.
	 *
	 * @return the key.
	 */
	public SecretKeySpec getKey()
	{
		return key;
	}

	/**
	 * Set the key variable.
	 *
	 * @param _key a key instance.
	 */
	public void setKey(SecretKeySpec _key)
	{
		key = _key;
		
	}
	
	/**
	 * Get the plaintext protocol message stored in this 
	 * object.
	 *
	 * @return the plaintext protocol message.
	 */
	public I_Message getProtoMsg()
	{
		return protoMsg;
	}

	/**
	 * Set the plaintext protocol message stored in this 
	 * object.
	 *
	 * @param _protoMsg a plaintext protocol message.
	 */
	public void setProtoMsg(I_Message _protoMsg)
	{
		protoMsg = _protoMsg;
	}

	/**
	 * Set the byte array converted from the plaintext protocol 
	 * message stored in this object.
	 *
	 * @param msgBytes the byte array of plaintext protocol message.
	 */
	public void setProtoMsgBytes(byte[] msgBytes)
	{
		protoMsgBytes = msgBytes;
	}

	/**
	 * Get the byte array converted from the plaintext protocol 
	 * message stored in this object.
	 *
	 * @return the byte array of plaintext protocol message.
	 */
	public byte[] getProtoMsgBytes()
	{
		if(protoMsgBytes==null){
			protoMsgBytes = this.protoMsg.toByteArray();
		}
		return protoMsgBytes;
	}

	/**
	 * Get the length of the payload
	 *
	 * @return The payload length
	 */
	public int getPayloadLength()
	{
		return payloadLength;
	}

	/**
	 * Set the creation time for the sender's key
	 *
	 * @param timestamp the creation time of the key
	 */
	public void setKeyTimestamp(long timestamp)
	{
		keyTimestamp = timestamp;
	}

	/** 
	 * Get the creation time of the key in the KeyUpdate or ProtoMsg message
	 *
	 * @return the creation time of the key contained in the message
	 */
	public long getKeyTimestamp()
	{
		return keyTimestamp;
	}
	
	public String toString(){
		switch(this.getType()){
			case SecInfoExchange_Message.CertRequest:
				return "CertRequest";
			case SecInfoExchange_Message.CertReply:
				return "CertReply";
			case SecInfoExchange_Message.KeyRequest:
				return "KeyRequest";
			case SecInfoExchange_Message.KeyUpdate:
				return "KeyUpdate";
			case SecInfoExchange_Message.ProtoMsg:
				return "SecExchg:"+this.protoMsg.toString();
			default:
				return "Unknown security exchange message type";
		}
	}
	/**
	 * @return Returns the sequenceNum.
	 */
	public int getSequenceNum() {
		return sequenceNum;
	}
	/**
	 * @param sequenceNum The sequenceNum to set.
	 */
	public void setSequenceNum(int sequenceNum) {
		this.sequenceNum = sequenceNum;
	}
}
