/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.Vector;
import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Attr;

import org.apache.xpath.XPath;

import hypercast.util.XmlUtil;
import hypercast.util.XmlUtil;

import javax.xml.transform.TransformerException;

/**
 * This class defines the functions of accessing to a set of statistics. 
 * Usually the instance of this class is created by an I_Stats 
 * object and handles the statistics defined in this I_Stats object.
 *
 * @author HyperCast Team
 */

public class StatsProcessor
{
    private final static boolean debug = false;

	/**
	 * The reference to the I_Stats object which contains this StatsProcessor object.
	 */
	private I_Stats statisticsObj;

    /**
     * Vector which contains the statistics (each statistics is represented
     * by a StatsElement instance) defined in the I_Stats object which
     * contains this object.
     */
    private Vector seObjectTable;

	/*
	 * It is not assumed that a writable statistics must be readable.
	 */
	/**
	 * Readable statistic flag.
	 */
	boolean isReadableFlag = true;

	/**
	 * Writable statistic flag.
	 */
	boolean isWritableFlag = false;

	/**
	 * The type of this statistic. For an I_Stats object which contains
	 * a StatsProcessor instance, the type of the statistic representing
	 * this I_Stats object is always "xsd:complexType".
	 */
	final String type = "xsd:complexType";

	
	public StatsProcessor(I_Stats statsObject, boolean readableFlag, boolean writableFlag)
	{
		statisticsObj = statsObject;

		isReadableFlag = readableFlag;
		isWritableFlag = writableFlag;
		
		seObjectTable = new Vector();
	}
    
    /**
     * Add a StatsElement instance into the StatsElement instance vector.
     */
    private void addStatsElement(StatsElement se)
    {

		//if (StatsProcessor.debug)
		//	System.err.println("StatsProcessor:addStatsElement: add StatsElement: "+se.getName());

        seObjectTable.addElement(se);
    }

	/**
	 * Add a StatsElement instance into the StatsElement instance vector.
	 */
	public void addStatsElement(String Name, I_Stats statisticsObj, int minoccur, int maxoccur)
	{

		//if (StatsProcessor.debug)
		//	System.err.println("StatsProcessor:addStatsElement: add StatsElement: "+se.getName());

		if (statisticsObj == null)
		{
			throw new HyperCastWarningRuntimeException ("Try to build statistic "+Name+" for a null I_Stats object.");
		}
		else
		{
			/*
			 * Set the statistic name of the I_Stats object to given value.
			 */
			statisticsObj.setStatsName(Name);

			seObjectTable.addElement(new StatsElement(Name, statisticsObj, minoccur, maxoccur));
		}
	}

    /**
     * Remove a StatsElement instance in the StatsElement instance vector
     * which has given index value.
     */
    public void removeStatsElement(int index)
    {
        seObjectTable.remove(index);
    }

	/**
	 * Remove the StatsElement instance in the StatsElement instance vector.
	 * The I_Stats object contains in this instance matchs the given object
	 * (by equals() operation).
	 */
	public void removeStatsElement(Object obj)
	{
		for (int i = 0; i < seObjectTable.size(); i++)
		{
			StatsElement se = (StatsElement) seObjectTable.elementAt (i);
			//if (se.getStatsObj().equals(obj))
			if (se.getStatsObj() == obj)
			{
				seObjectTable.remove(i);
				return;
			}
		}
	}


	/**
	 * Remove the StatsElement instance in the StatsElement instance vector.
	 * The I_Stats object contains in this instance matchs the given object
	 * (by equals() operation).
	 */
	public void removeStatsElement(I_AddressPair addrPair)
	{
		for (int i = 0; i < seObjectTable.size(); i++)
		{
			StatsElement se = (StatsElement) seObjectTable.elementAt (i);
			if (se.getStatsObj().equals(addrPair))
			{
				seObjectTable.remove(i);
				return;
			}
		}
	}


	/**
	 * Remove the StatsElement instance in the StatsElement instance vector.
	 * The removed StatsElement instance has the given index in the StatsElement
	 * instances which match the given names.
	 */
	public void removeStatsElement(String name, int index)
	{
		Vector matchedStatsElements = getStatsElement(name);
		
		if ((index < 0) || (index >=matchedStatsElements.size()))
		{
			throw new HyperCastWarningRuntimeException ("Try to remove an inexist StatsElement.");
		}

		StatsElement se = (StatsElement) matchedStatsElements.elementAt(index);
		
		seObjectTable.removeElement(se);
	}

    /**
     * Remove all StatsElement instances in the StatsElement instance vector.
     */
    public void removeAllStatsElement()
    {
        seObjectTable.removeAllElements();
    }


	/**
	 * Remove all StatsElement instances in the StatsElement instance vector
	 * which have the given name.
	 */
	public void removeAllStatsElement(String name)
	{
		int size = seObjectTable.size();
		
		for (int i = seObjectTable.size()-1; i>=0; i--)
		{
			StatsElement se = (StatsElement) seObjectTable.elementAt (i);
			if (se.getName().equals(name))
			{
				seObjectTable.remove(i);
			}
		}
	}

    /**
     * Get the size of StatsElement vector stored in this object.
     */
    public int getNumOfStatistics()
    {
        return seObjectTable.size();
    }

	/**
	 * Get the Vector that contains all StatsElement instances in 
	 * the StatsElement instance vector which match the given name.
	 */
	private Vector getStatsElement(String name)
	{
		Vector result = new Vector();

		if (name.equals("*"))
		{
			result = seObjectTable;
		}
		else
		{
			for (int i = 0; i < seObjectTable.size(); i++)
			{
				StatsElement se = (StatsElement) seObjectTable.elementAt (i);
				if (se.getName().equals(name))
				{
					result.addElement(se);

				}
			}
		}
				
		return result;
	}

	/**
	 * Get the StatsElement instance in the StatsElement instance vector 
	 * which has the given index in the vector.
	 */
	private StatsElement getStatsElement(int index)
	{
		if ((index >= 0) && (index < seObjectTable.size()))
		{
			return (StatsElement) seObjectTable.elementAt (index);
		}
		else
		{
			return null;
		}
	}


	/**
	 * Check if the vector contains a StatsElement instance which
	 * has the given name.
	 */
	public boolean containStatsElement(String name)
	{
		for (int i = 0; i < seObjectTable.size(); i++)
		{
			StatsElement se = (StatsElement) seObjectTable.elementAt (i);
			if (se.getName().equals(name))
			{
				return true;
			}
		}
			
		return false;
	}

	/**
	 * Get the Vector that contains all Elements in the node list which match
	 * the given name.
	 */
	private Vector getElementsFromNodeList(NodeList nl, String name)
	{
		Vector result = new Vector();
		
		for (int i=0; i<nl.getLength(); i++)
		{
			Element ele = (Element) nl.item(i);
			String eleName = ele.getNodeName();
			if (eleName.equals(name))
			{
				result.addElement(ele);
			}
		}
		return result;
	}

	/**
	 * Get the Vector that contains all Elements in the Element array which match
	 * the given name.
	 */
	public static Vector getElementsFromElementArray(Element[] el, String name)
	{
		Vector result = new Vector();
		
		for (int i=0; i<el.length; i++)
		{
			String eleName = el[i].getNodeName();
			if (eleName.equals(name))
			{
				result.addElement(el[i]);
			}
		}
		return result;
	}


	/**
	 * Get the Vector that contains all Elements in the Element Vector which match
	 * the given name.
	 */
	public static Vector getElementsFromElementVector(Vector vec, String name)
	{
		Vector result = new Vector();
		
		for (int i=0; i<vec.size(); i++)
		{
			Element ele = (Element) vec.elementAt(i);
			String eleName = ele.getNodeName();
			if (eleName.equals(name))
			{
				result.addElement(ele);
			}
		}
		return result;
	}

	/**
	 * Check if the all StatsElement instances in the given Vector match the given name.
	 */
	private static boolean allStatsElementsInVectorMatchGivenName(Vector vec, String name)
	{
		for (int i=0; i<vec.size(); i++)
		{
			StatsElement se = (StatsElement) vec.elementAt (i);
			if (!se.getName().equals(name))
			{
				return false;
			}
		}
		return true;
	}

	/**
	 * Check if the all Element instances in the given Vector match the given name.
	 */
	public static boolean allElementsInVectorMatchGivenName(Vector vec, String name)
	{
		for (int i=0; i<vec.size(); i++)
		{
			Element ele = (Element) vec.elementAt (i);
			if (!ele.getNodeName().equals(name))
			{
				return false;
			}
		}
		return true;
	}


	/**
	 * Get the name list of all different Element instances in 
	 * the given node list. 
	 */
	private Vector getNameListFromNodeList(NodeList nl)
	{
		Vector result = new Vector();
		
		for (int i=0; i<nl.getLength(); i++)
		{
			String name = nl.item(i).getNodeName();
			if (StatsProcessor.containNameInVector(result, name))
			{
				continue;
			}
			else
			{
				result.addElement(new String(name));
			}
		}
		return result;
	}

	/**
	 * Get the name list of all different Element instances in 
	 * the given Element array. 
	 */
	public static Vector getNameListFromElementArray(Element[] el)
	{
		Vector result = new Vector();
		
		for (int i=0; i<el.length; i++)
		{
			String name = el[i].getNodeName();
			if (StatsProcessor.containNameInVector(result, name))
			{
				continue;
			}
			else
			{
				result.addElement(new String(name));
			}
		}
		return result;
	}

	/**
	 * Check if the given Vector, in which each element is a string,
	 * contains the string matching the given string (name). 
	 */
	private static boolean containNameInVector(Vector vec, String name)
	{
		for (int i = 0; i < vec.size(); i++)
		{
			String se = (String) vec.elementAt (i);
			if (se.equals(name))
			{
				return true;
			}
		}
		return false;
	}

    /**
     * Get name from given array entry expression which is in the format of "name[index]".
     *
     * Given "foo[34]" returns "foo"
     */
    private String getNameFromEntryExpression (final String entryExpression)
    {
        final int indexStartLoc = entryExpression.indexOf ("[");
        return (indexStartLoc < 0) ? entryExpression : entryExpression.substring (0, indexStartLoc);
    }
		
    /**
     * Get index from given array entry expression which is in the format of "name[index]".
     *
     * Given "foo[34]" returns 34
     */
    private int getIndexFromEntryExpression (final String entryExpression)
    {
        int indexStartLoc = entryExpression.indexOf ("[");
        int indexEndLoc = entryExpression.indexOf ("]");

        if (indexStartLoc < 0
            ||
            indexEndLoc < 0
            ||
            (indexStartLoc > 0
             &&
             indexEndLoc > 0
             &&
             indexStartLoc >= indexEndLoc))
        {
            return -1;
        }
        else 
        {
            return Integer.parseInt (entryExpression.substring (indexStartLoc + 1, indexEndLoc));
        }
    }


	/**
	 * Check if the node list (represents new value Element array) match the StatsElement 
	 * instances in the StatsElement instance vector.
	 */
	public boolean newValueArrayIsValid(NodeList nl)
	{
		Vector newValueNameList = this.getNameListFromNodeList(nl);

		for (int i = 0; i < newValueNameList.size(); i++)
		{ 
			String name = (String) newValueNameList.elementAt(i);
				
			Vector MatchedStatsElements = this.getStatsElement(name);
			Vector MatchedNewValueElements = this.getElementsFromNodeList(nl, name);

			if (MatchedStatsElements.size() != MatchedNewValueElements.size())
			{
				return false;
			}
		}
		return true;
	}

	/**
	 * Check if index from given array entry expression has valid value, i.e., >0.
	 */
	private boolean isIndexValidInEntryExpression (final String entryExpression)
	{
		int indexStartLoc = entryExpression.indexOf ("[");
		int indexEndLoc = entryExpression.indexOf ("]");

		if (indexStartLoc < 0
			||
			indexEndLoc < 0
			||
			(indexStartLoc > 0
			 &&
			 indexEndLoc > 0
			 &&
			 indexStartLoc >= indexEndLoc))
		{
			return false;
		}
		else 
		{
			int index = Integer.parseInt (entryExpression.substring (indexStartLoc + 1, indexEndLoc));
			return (index > 0);
		}
	}

	/**
	 * Check if the given array entry expression contains index.
	 */
	private boolean entryExpressionContainIndex (final String entryExpression)
	{
		int indexStartLoc = entryExpression.indexOf ("[");
		int indexEndLoc = entryExpression.indexOf ("]");

		if (indexStartLoc < 0
			||
			indexEndLoc < 0
			||
			(indexStartLoc > 0
			 &&
			 indexEndLoc > 0
			 &&
			 indexStartLoc >= indexEndLoc))
		{
			return false;
		}
		else 
		{
			return true;
		}
	}

    /**
     * Return the Element array which contains the value or sub-tree of stsitstics specified
     * by parameter xpath.
     */
    public Element[] getStatsResult (Document doc, XPath xpath) throws HyperCastStatsException 
    {
        if (StatsProcessor.debug)
            System.err.println("StatsProcessor:getStatsResult: xpath string is ***"+xpath.getPatternString()+"****");

        String xpathStr = xpath.getPatternString();
			
        /*
         * xpathStr may be "/name" or "/name/sub-xpath". "name" may be "*" or the name of the statistic 
         * representing I_Stats object statisticsObj. Sub-xpath is a string with value "statsname" or 
         * "statsname/xxxx" where statsname is the name of a statistic component supported in this object. 
         * For example, in the StatsProcessor stored in OL_SOcket, the xpathStr may be "/*", "/OLSocket", 
         * "/OLSocket/*", "/OLSocket/Node" or "/OLSocket/Node/state", etc.
         * 
         * The StatsProcessor checks "name" and uses it as the name to create top Element. If "name" is
         * "*", the name of the I_Stats object statisticsObj will be used as the name of top Element.
         * As an example, in the StatsProcessor stored in OL_SOcket, if xpathStr is "/OLSocket" or "/*", 
         * the StatsProcessor will create an Element with name "OLSocket", and adds all Elements for its
         * StatsElements instance to this Element to form a tree.
         */
			
        int firstSlash = xpathStr.indexOf("/");

        if (StatsProcessor.debug)
            System.err.println("StatsProcessor:getStatsResult: firstSlash location is "+firstSlash);

        /*
         * xpathStr must starts with "/". In this case, firstSlash is 0.
         */
        if ((firstSlash < 0) || (firstSlash > 0))
            //throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			return new Element[0];

        int secondSlash = xpathStr.substring(firstSlash+1).indexOf("/");

        if (StatsProcessor.debug)
            System.err.println("StatsProcessor:getStatsResult: secondSlash location is "+secondSlash);

		String statsName;
		if (secondSlash < 0)
		{
			statsName = xpathStr.substring(firstSlash+1);
		}
		else
		{		
			statsName = xpathStr.substring(firstSlash+1, secondSlash+1);
		}

		/*
		 * If the statistics name is provided in the xpath string but is not the same as the stats name
		 * of the I_Stats object which contains this StatsProcessor object, throw exception.
		 */
		if (!statsName.equals("*"))
		{
			//if (StatsProcessor.debug)
			//	System.err.println("StatsProcessor:getStatsResult: statsName="+statsName+",statisticsObj.getStatsName()="+statisticsObj.getStatsName());

			if (!statsName.equals(statisticsObj.getStatsName()))
			{
				return new Element[0];
			}
		}

        /************************ xpath has value "/name" or "/*" ******************************/
        /*
         * If secondSlash<0, i.e., second "/" is not found in xpathStr, xpathStr must have 
         * value "/name" or "/*". In this case a top Element with name statsName is created and 
         * the Elements created from the all StatsElement instances in this object are added to the 
         * top Element as children.
         */
        if (secondSlash < 0)
        {
            /*
             * Get the name for the top Element.
             */
            String topName = this.getNameFromEntryExpression(statsName);
			if (topName.equals("*"))
			{
				topName = statisticsObj.getStatsName();
			}

			/*
			 * The statsName can be in theformat "name[1]", i.e., contains index. But the index 
			 * must be 1.
			 */
			int index = this.getIndexFromEntryExpression(statsName);

           
            Element[] valueElement = new Element[1];

            /*
             * If xpathStr has value "/xxxx[index]", we use "xxxx" as the name of top Element.
             * Note that "xxxx[index]" is not a valid Element name.
             */
			valueElement[0] = doc.createElement(topName);

			/*
			 * If statsName contains index, the index only can be 1 and the xpath string looks 
			 * like "/name[1]".
			 */
			if ((entryExpressionContainIndex(statsName)) && (index != 1))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}

			if (StatsProcessor.debug)
				System.err.println("StatsProcessor:getStatsResult: whole StatsElement vector size=: "+seObjectTable.size());
			
            /*
             * Iteration on the StatsElement list. For each StatsElement instance
             * in the vector, get the Element that represents the query result for
             * this instance, and add the obtained Element to the top Element.
             */
            for (int i = 0; i < seObjectTable.size(); i++)
            { 
                StatsElement se = (StatsElement) seObjectTable.elementAt (i);
	
                /*
                 * Find the name of this StatsElement instance, and get the query 
                 * result for this instance.
                 */					
                String subname = se.getName();

				if (StatsProcessor.debug)
					System.err.println("StatsProcessor:getStatsResult: matched StatsElement: "+subname);
					
                XPath newXPath = null;
                try
                {
					newXPath = new XPath("/"+subname, null, null, XPath.MATCH);
                }
                catch (TransformerException te)
                {
                    //throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
                    continue;
                }
						
				I_Stats subComponent = se.getStatsObj();
				Element[] subResultElement = null;
				if (subComponent != null)
				{
					subResultElement = subComponent.getStats(doc, newXPath);
				}
						
                /*
                 * Add the result Element for this StatsElement instance to the top Element.
                 */
                if ((subResultElement != null) && (subResultElement.length != 0) && (subResultElement[0] != null))
                {
                    valueElement[0].appendChild(subResultElement[0]);
                }
            } 

            /*
             * The top Element, which is a Element array of size 1, is returned as the result.
             */
            return valueElement;
        }
        /************************ xpath has value "/name/subxpath" ******************************/
        /*
         * Otherwise, xpathStr has form "/name/sub-xpath" which points to a StatsElement 
         * instance defined in this object. The "sub-xpath" is analyzed and corresponding 
         * action is taken.
         */
        else
        {
            String subxpath = xpathStr.substring(secondSlash+2);
            int thirdSlash = subxpath.indexOf("/");
				
            String subcomponent;
			String remainedpath = null;
            if (thirdSlash > 0)
            {
                /*
                 * The xpathStr looks like "/OL_Socket/Node/State" in OL_Socket, 
                 * for example, then subxpath is "Node/State", and subcomponent is "Node".
                 */
                subcomponent = subxpath.substring(0, thirdSlash);
				remainedpath = subxpath.substring(thirdSlash);
            }
            else
            {
                /*
                 * The xpathStr looks like "/OL_Socket/Node" in OL_Socket, 
                 * for example, then subxpath is "Node", and subcomponent is "Node".
                 */
                subcomponent = subxpath;
				remainedpath = "";
            }

			/*
			 * Get the name of subcomponent.
			 */
			String subName = this.getNameFromEntryExpression(subcomponent);

			/*
			 * Get the index contained in the string subcomponent. It is -1 if 
			 * no index is contained.
			 */
			int index = this.getIndexFromEntryExpression(subcomponent);

			//if (StatsProcessor.debug)
			//	System.err.println("StatsProcessor:getStatsResult: subcomponent="+subcomponent+",subName="+subName+",index="+index);

			/*
			 * The xpath string "/name/*[i]" is not supported.
			 */
			if ((subName.equals("*")) && (entryExpressionContainIndex(subcomponent)))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}

			/*
			 * Get the Vector that contains the StatsElement instances which 
			 * are readable and match subname.
			 */
			//Vector matchedComponents = this.getReadableStatsElement(subName);
			Vector matchedComponents = this.getStatsElement(subName);

			if (StatsProcessor.debug)
				System.err.println("StatsProcessor:getStatsResult: matched StatsElement vector size=: "+matchedComponents.size());

			/*
			 * If no StatsElement instance is found which matchs subname, or the specified 
			 * index is beyond the range of the list, matchedComponents, throw exception.
			 */
			if ((matchedComponents.size() == 0) || 
				((entryExpressionContainIndex(subcomponent)) && ((index <= 0) || (index > matchedComponents.size()))))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}

			Vector resultVec = new Vector();

			/*
			 * Iteration on the StatsElement instances in matchedComponents. For each StatsElement 
			 * instance in the vector, get the Element that represents the query result for
			 * that instance, and add the obtained Element to the result Element array.
			 */
			for (int i = 0; i < matchedComponents.size(); i++)
			{
				/*
				 * If index is larger than 0, matchedComponents contains a list
				 * of StatsElement instances with the same statsName. The index is
				 * used to find the desired StatsElement instance.
				 * 
				 * Note that index starts from 1 in the xpath string.
				 */
				if ((index > 0) && (index != (i+1)))
				{
					continue;
				}
				
				StatsElement se = (StatsElement) matchedComponents.elementAt (i);
	
				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String compname = se.getName();

				if (StatsProcessor.debug)
					System.err.println("StatsProcessor:getStatsResult: matched StatsElement: "+compname);
					
				XPath newXPath = null;
				try
				{
					newXPath = new XPath("/"+compname+remainedpath, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
					continue;
				}

				//if (StatsProcessor.debug)
				//	System.err.println("StatsProcessor:getStatsResult: compname+remainedpath= "+compname+remainedpath);
						
				I_Stats subComponent = se.getStatsObj();
				Element[] subResultElement = null;
				if (subComponent != null)
				{
					if (StatsProcessor.debug)
						System.err.println("StatsProcessor:getStatsResult: call subComponent.getStats() for"+compname);

					subResultElement = subComponent.getStats(doc, newXPath);
				}
						
				/*
				 * Add the result Element for this StatsElement instance to the top Element.
				 */
				if (subResultElement != null)
				{
					//if (StatsProcessor.debug)
					//	System.err.println("StatsProcessor:getStatsResult: add Element to resultRec:"+subResultElement[0].getNodeName());

					for (int j=0; j<subResultElement.length; j++)
					{
						resultVec.addElement(subResultElement[j]);
					}

					//if (StatsProcessor.debug)
					//	System.err.println("StatsProcessor:getStatsResult: add Element to resultRec:resultVec.size()="+resultVec.size());

				}
			} 

			if (StatsProcessor.debug)
				System.err.println("StatsProcessor:getStatsResult: add Element to resultRec:after execution:resultVec.size()="+resultVec.size());


			Element[] valueElement = new Element[resultVec.size()];

			for (int i = 0; i < resultVec.size(); i++)
			{ 
				valueElement[i] = (Element) resultVec.elementAt (i);
			}

			return valueElement;
        }
    }
		
    /**
     * Set the value or sub-tree of stsitstics specified by xpath. The value or sub-tree actually set is
     * returned as an Element or sub-tree.
     */
    public Element[] setStatsResult (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:setStatsResult: xpath string is ***"+xpath.getPatternString()+"****");

		String xpathStr = xpath.getPatternString();
			
		/*
		 * See this.getStatsResult() for the analysis on xpath string.
		 */			
		int firstSlash = xpathStr.indexOf("/");

		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:setStatsResult: firstSlash location is "+firstSlash);

		/*
		 * xpathStr must starts with "/". In this case, firstSlash is 0.
		 */
		if ((firstSlash < 0) || (firstSlash > 0))
			//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			return new Element[0];

		int secondSlash = xpathStr.substring(firstSlash+1).indexOf("/");

		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:setStatsResult: secondSlash location is "+secondSlash);

		String statsName;
		if (secondSlash < 0)
		{
			statsName = xpathStr.substring(firstSlash+1);
		}
		else
		{		
			statsName = xpathStr.substring(firstSlash+1, secondSlash+1);
		}

		/*
		 * If the statistics name is provided in the xpath string but is not the same as the stats name
		 * of the I_Stats object which contains this StatsProcessor object, throw exception.
		 */
		if (!statsName.startsWith("*"))
		{
			if (!statsName.equals(statisticsObj.getStatsName()))
			{
				return new Element[0];
			}
		}


		/************************ xpath has value "/name" or "/*" ******************************/
		/*
		 * If secondSlash<0, i.e., second "/" is not found in xpathStr, xpathStr must have 
		 * value "/name" or "/*". In this case a top Element with name statsName is created and 
		 * the Elements created from the all StatsElement instances in this object are added to the 
		 * top Element as children.
		 */
		if (secondSlash < 0)
		{
			String topName = this.getNameFromEntryExpression(statsName);
			if (topName.equals("*"))
			{
				topName = statisticsObj.getStatsName();
			}

			/*
			 * The statsName can be in theformat "name[1]", i.e., contains index. But the index 
			 * must be 1.
			 */
			int index = this.getIndexFromEntryExpression(statsName);

			/*
			 * If statsName contains index, the index only can be 1 and the xpath string looks 
			 * like "/name[1]".
			 */
			if ((entryExpressionContainIndex(statsName)) && (index != 1))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}

			/*
			 * In this case, the new value array must have only one element whose name matches topName.
			 * Otherwise, an exception is thrown.
			 */
			if (!topName.equals(newValue.getNodeName()))
			{
				throw new HyperCastStatsException(xpathStr, HyperCastStatsException.IncorrectSetValue);
			}


			/*
			 * Create a top Element and set its name to topName.
			 */
			Element[] valueElement = new Element[1];

			/*
			 * If xpathStr has value "/xxxx[index]", we use "xxxx" as the name of top Element
			 * and set index attribute for this Element. Note that "xxxx[index]" is not a valid 
			 * Element name.
			 */
			valueElement[0] = doc.createElement(topName);


			/*
			 * Get the all children of newValue Element (it is a parameter).
			 */
			NodeList nl = newValue.getChildNodes();

			/*
			 * Get the name list of different Element instances in the node list.
			 */
			Vector newValueNameList = this.getNameListFromNodeList(nl);
			
			/*
			 * If the new value Element list doesn't match the StatsElement instance list
			 * in this object, throw exception. See this.newValueArrayIsValid(NodeList nl).
			 */
			if (!newValueArrayIsValid(nl))
			{
				throw new HyperCastStatsException(xpathStr, HyperCastStatsException.IncorrectSetValue);
			}

			/*
			 * For each different name, find the StatsElements instances and the Elements 
			 * in the node list created above which match the name, and recursively call
			 * getStats().
			 */
			for (int i = 0; i < newValueNameList.size(); i++)
			{ 
				String name = (String) newValueNameList.elementAt(i);
				
				Vector MatchedStatsElements = this.getStatsElement(name);
				Vector MatchedNewValueElements = this.getElementsFromNodeList(nl, name);

				for (int j = 0; j < MatchedStatsElements.size(); j++)
				{
					StatsElement se = (StatsElement) MatchedStatsElements.elementAt(j);
					Element newValueElement = (Element) MatchedNewValueElements.elementAt(j);
					
					XPath newXPath = null;
					try
					{
						newXPath = new XPath("/"+name, null, null, XPath.MATCH);
					}
					catch (TransformerException te)
					{
						//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
						continue;
					}

					Element[] subValueElement = se.getStatsObj().setStats(doc, newXPath, newValueElement);
					
					valueElement[0].appendChild(subValueElement[0]);
				}
				
			}
			return valueElement;
		}
		/************************ xpath has value "/name/subxpath" ******************************/
		/*
		 * Otherwise, xpathStr has form "/name/sub-xpath" which points to a StatsElement 
		 * instance defined in this object. The "sub-xpath" is analyzed and corresponding 
		 * action is taken.
		 */
		else
		{
			String subxpath = xpathStr.substring(secondSlash+2);
			int thirdSlash = subxpath.indexOf("/");
				
			String subcomponent;
			String remainedpath = null;
			if (thirdSlash > 0)
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node/State" in OL_Socket, 
				 * for example, then subxpath is "Node/State", and subcomponent is "Node".
				 */
				subcomponent = subxpath.substring(0, thirdSlash);
				remainedpath = subxpath.substring(thirdSlash);
			}
			else
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node" in OL_Socket, 
				 * for example, then subxpath is "Node", and subcomponent is "Node".
				 */
				subcomponent = subxpath;
				remainedpath = "";
			}

			/*
			 * Get the name of subcomponent.
			 */
			String subName = this.getNameFromEntryExpression(subcomponent);

			/*
			 * Get the index contained in the string subcomponent. It is -1 if 
			 * no index is contained.
			 */
			int index = this.getIndexFromEntryExpression(subcomponent);

			/*
			 * The xpath string "/name/*[i]" is not supported.
			 */
			if ((subName.equals("*")) && (index > 0))
			{
				throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			}

			/*
			 * Get the Vector that contains the StatsElement instances which 
			 * are writable and match subname. Note that subname may be "*", 
			 * in that case the following call returns all StatsElement instances
			 * in this object.
			 */
			Vector matchedComponents = this.getStatsElement(subName);

			/*
			 * If no StatsElement instance is found which matchs subname, or the specified 
			 * index is beyond the range of the list, matchedComponents, throw exception.
			 */
			if ((matchedComponents.size() == 0) || 
				((entryExpressionContainIndex(statsName)) && ((index < 1) || (index > matchedComponents.size()))))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}

			String newValueName = newValue.getNodeName();

			if (subcomponent.equals("*") && (!allStatsElementsInVectorMatchGivenName(matchedComponents, newValueName)) ||
				((remainedpath.equals("")) && (!subName.equals(newValueName))))
			
			{
				throw new HyperCastStatsException(xpathStr, HyperCastStatsException.IncorrectSetValue);
			}

			Vector resultVec = new Vector();

			/*
			 * Iteration on the StatsElement instances in matchedComponents. For each StatsElement 
			 * instance in the vector, get the Element that represents the query result for
			 * that instance, and add the obtained Element to the result Element array.
			 */
			for (int j = 0; j < matchedComponents.size(); j++)
			{
				/*
				 * If index is larger than 0, matchedComponents contains a list
				 * of StatsElement instances with the same statsName. The index is
				 * used to find the desired StatsElement instance.
				 * 
				 * Note that index starts from 1 in the xpath string.
				 */
				if ((index > 0) && (index != (j+1)))
				{
					continue;
				}
			
				StatsElement se = (StatsElement) matchedComponents.elementAt (j);

				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String name = se.getName();
				
				XPath newXPath = null;
				try
				{
					newXPath = new XPath("/"+name+remainedpath, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				}
					
				I_Stats subComponent = se.getStatsObj();
				Element[] subResultElement = null;
				if (subComponent != null)
				{
					subResultElement = subComponent.setStats(doc, newXPath, newValue);
				}
				
				/*
				 * Add the result Element for this StatsElement instance to the top Element.
				 */
				if ((subResultElement != null) && (subResultElement.length != 0) && (subResultElement[0] != null))
				{
					resultVec.addElement(subResultElement[0]);
				}
			} 

			Element[] valueElement = new Element[resultVec.size()];

			for (int j = 0; j < resultVec.size(); j++)
			{ 
				valueElement[j] = (Element) resultVec.elementAt (j);
			}

			return valueElement;
		}
    }


    /**
     * Return the Element which contains the read schema element for the stsitstics specified
     * by parameter xpath.
     */
    public Element[] getReadSchemaResult(Document doc, XPath xpath)  throws HyperCastStatsException
    {

		if (StatsProcessor.debug)
		    System.err.println("StatsProcessor:getReadSchemaResult: statisticsName="+statisticsObj.getStatsName()+", xpath string is ***"+xpath.getPatternString()+"****");


		if (!isReadableFlag)
		{
			return new Element[0];
		}

		String xpathStr = xpath.getPatternString();
				
		/*
		 * See this.getStatsResult() for the analysis on xpath string.
		 */			
				
		int firstSlash = xpathStr.indexOf("/");
	
		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:getReadSchemaResult: firstSlash location is "+firstSlash);
	
		/*
		 * xpathStr must starts with "/". In this case, firstSlash is 0.
		 */
		if ((firstSlash < 0) || (firstSlash > 0))
			//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			return new Element[0];
			
		int secondSlash = xpathStr.substring(firstSlash+1).indexOf("/");
	
		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:getReadSchemaResult: secondSlash location is "+secondSlash);
	
		String statsName;
		if (secondSlash < 0)
		{
			statsName = xpathStr.substring(firstSlash+1);
		}
		else
		{		
			statsName = xpathStr.substring(firstSlash+1, secondSlash+1);
		}
	
		/*
		 * If the statistics name is provided in the xpath string but is not the same as the stats name
		 * of the I_Stats object which contains this StatsProcessor object, throw exception.
		 */
		if (!statsName.startsWith("*"))
		{
			if (!statsName.equals(statisticsObj.getStatsName()))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
		}
	
		/************************ xpath has value "/name" or "/*" ******************************/
		/*
		 * If secondSlash<0, i.e., second "/" is not found in xpathStr, xpathStr must have 
		 * value "/name" or "/*". In this case a top Element with name statsName is created and 
		 * the Elements created from the all StatsElement instances in this object are added to the 
		 * top Element as children.
		 */
		if (secondSlash < 0)
		{
			/*
			 * Create a top Element.
			 */
			String topName = this.getNameFromEntryExpression(statsName);
			if (topName.equals("*"))
			{
				topName = statisticsObj.getStatsName();
			}
	
			/*
			 * The statsName can be in theformat "name[1]", i.e., contains index. But the index 
			 * must be 1.
			 */
			int index = this.getIndexFromEntryExpression(statsName);
	
			/*
			 * If statsName contains index, the index only can be 1 and the xpath string looks 
			 * like "/name[1]".
			 */
			if ((entryExpressionContainIndex(statsName)) && (index != 1))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
			
			/*
			 * Obtain schema element for the I_Stats object which contains this object.
			 */
			Element[] valueElement = XmlUtil.createSchemaElement(doc, topName, type, null, null);

			if (valueElement.length == 0)
			{
				return valueElement;
			}

			/*
			 * Above getReadSchemaElement() returns an sub-tree, subTopElement
			 * defined below is the Element in the sub-tree to which all schema
			 * sub-tree created by the sub-components are mounted. 
			 */
			Element subTopElement;
			if (type.equals("xsd:complexType"))
			{
				subTopElement = (Element) valueElement[0].getChildNodes().item(0);
			}
			else
			{
				//subTopElement = valueElement[0];
				return valueElement;
			}
							
			/*
			 * The Hash table which contains the names of StatsElement instances as keys
			 * and the number of StatsElement instances with the same name as elements.
			 */
			Hashtable nameListHash = new Hashtable();
				
			/*
			 * Iteration on the StatsElement list. For each StatsElement instance
			 * in the vector, get the Element that represents the query result for
			 * this instance, and add the obtained Element to the top Element.
			 */
			for (int i = 0; i < seObjectTable.size(); i++)
			{ 
				StatsElement se = (StatsElement) seObjectTable.elementAt (i);

				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String subname = se.getName();

				if (nameListHash.containsKey(subname))
				{
					Integer value = (Integer) nameListHash.get(subname);
					nameListHash.put(subname, new Integer(value.intValue()+1));
					continue;
				}
						
				XPath newXPath = null;
				try
				{
					newXPath = new XPath(xpathStr+"/"+subname, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
					continue;
				}
						
				Element[] subResultElement = this.getReadSchemaResult(doc, newXPath);
						
				/*
				 * Add the result Element for this StatsElement instance to the top Element.
				 */
				if ((subResultElement != null) && (subResultElement.length != 0) && (subResultElement[0] != null))
				{
					subTopElement.appendChild(subResultElement[0]);

					if (!nameListHash.containsKey(subname))
					{
						nameListHash.put(subname, new Integer(1));
					}
				}
			} 
	
			/*
			 * The top Element is returned as the result.
			 */
			return valueElement;
		}
		/************************ xpath has value "/name/subxpath" ******************************/
		/*
		 * Otherwise, xpathStr has form "/name/sub-xpath" which points to a StatsElement 
		 * instance defined in this object. The "sub-xpath" is analyzed and corresponding 
		 * action is taken.
		 */
		else
		{
			String subxpath = xpathStr.substring(secondSlash+2);
			int thirdSlash = subxpath.indexOf("/");
					
			String subcomponent;
			String remainedpath = null;
			if (thirdSlash > 0)
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node/State" in OL_Socket, 
				 * for example, then subxpath is "Node/State", and subcomponent is "Node".
				 */
				subcomponent = subxpath.substring(0, thirdSlash);
				remainedpath = subxpath.substring(thirdSlash);
			}
			else
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node" in OL_Socket, 
				 * for example, then subxpath is "Node", and subcomponent is "Node".
				 */
				subcomponent = subxpath;
				remainedpath = "";
			}
	
			/*
			 * Get the name of subcomponent.
			 */
			String subName = this.getNameFromEntryExpression(subcomponent);
	
			/*
			 * Get the index contained in the string subcomponent. It is -1 if 
			 * no index is contained.
			 */
			int index = this.getIndexFromEntryExpression(subcomponent);

			/*
			 * The xpath string "/name/*[i]" is not supported.
			 */
			if ((subName.equals("*")) && (index > 0))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
	
			/*
			 * Get the Vector that contains the StatsElement instances which 
			 * are readable and match subname.
			 */
			Vector matchedComponents = this.getStatsElement(subName);
	
			/*
			 * If no StatsElement instance is found which matchs subname, or the specified 
			 * index is beyond the range of the list, matchedComponents, throw exception.
			 */
			if ((matchedComponents.size() == 0) || 
				((entryExpressionContainIndex(statsName)) && ((index < 1) || (index > matchedComponents.size()))))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
	
			Vector resultVec = new Vector();
	
			/*
			 * The Hash table which contains the names of StatsElement instances as keys
			 * and the number of StatsElement instances with the same name as elements.
			 */
			Hashtable nameListHash = new Hashtable();
				
			/*
			 * Iteration on the StatsElement instances in matchedComponents. For each StatsElement 
			 * instance in the vector, get the Element that represents the query result for
			 * that instance, and add the obtained Element to the result Element array.
			 */
			for (int i = 0; i < matchedComponents.size(); i++)
			{
				/*
				 * If index is larger than 0, matchedComponents contains a list
				 * of StatsElement instances with the same statsName. The index is
				 * used to find the desired StatsElement instance.
				 * 
				 * Note that index starts from 1 in the xpath string.
				 */
				if ((index > 0) && (index != (i+1)))
				{
					continue;
				}
					
				StatsElement se = (StatsElement) matchedComponents.elementAt (i);
		
				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String compname = se.getName();

				if (nameListHash.containsKey(compname))
				{
					Integer value = (Integer) nameListHash.get(compname);
					nameListHash.put(compname, new Integer(value.intValue()+1));
					continue;
				}

				XPath newXPath = null;
				try
				{
					newXPath = new XPath("/"+compname+remainedpath, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				}

				Element[] subResult = se.getStatsObj().getReadSchema(doc, newXPath);

				for (int j=0; j<subResult.length; j++)
				{
					resultVec.addElement(subResult[j]);
				}

				if (!(nameListHash.containsKey(compname)) && (subResult.length > 0))
				{
					nameListHash.put(compname, new Integer(1));
				}
				
				if (StatsProcessor.debug)
				  System.out.println("StatsProcessor:getReadSchemaResult:subResult.length="+subResult.length+", subResult[0].getNodeName()="+subResult[0].getNodeName()+", compname="+compname);
				if ((subResult.length == 1) && (compname.equals(subResult[0].getAttribute("name"))))
				{
					se.resetSchemaAttribute(subResult[0]);
				}
				
			}
				
			Element[] valueElement = new Element[resultVec.size()];
	
			for (int i = 0; i < resultVec.size(); i++)
			{ 
				valueElement[i] = (Element) resultVec.elementAt (i);
			}
	
			return valueElement;
		}

    }

		
    /**
     * Return the Element which contains the write schema element for the stsitstics specified
     * by parameter xpath.
     */
    public Element[] getWriteSchemaResult(Document doc, XPath xpath)  throws HyperCastStatsException
    {

		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:getWriteSchemaResult: xpath string is ***"+xpath.getPatternString()+"****");


		if (!isWritableFlag)
		{
			return new Element[0];
		}

		String xpathStr = xpath.getPatternString();
				
		/*
		 * See this.getStatsResult() for the analysis on xpath string.
		 */			
				
		int firstSlash = xpathStr.indexOf("/");
	
		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:getWriteSchemaResult: firstSlash location is "+firstSlash);
	
		/*
		 * xpathStr must starts with "/". In this case, firstSlash is 0.
		 */
		if ((firstSlash < 0) || (firstSlash > 0))
			//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			return new Element[0];
			
		int secondSlash = xpathStr.substring(firstSlash+1).indexOf("/");
	
		if (StatsProcessor.debug)
			System.err.println("StatsProcessor:getWriteSchemaResult: secondSlash location is "+secondSlash);
	
		String statsName;
		if (secondSlash < 0)
		{
			statsName = xpathStr.substring(firstSlash+1);
		}
		else
		{		
			statsName = xpathStr.substring(firstSlash+1, secondSlash+1);
		}
	
		/*
		 * If the statistics name is provided in the xpath string but is not the same as the stats name
		 * of the I_Stats object which contains this StatsProcessor object, throw exception.
		 */
		if (!statsName.startsWith("*"))
		{
			if (!statsName.equals(statisticsObj.getStatsName()))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
		}
	
		/************************ xpath has value "/name" or "/*" ******************************/
		/*
		 * If secondSlash<0, i.e., second "/" is not found in xpathStr, xpathStr must have 
		 * value "/name" or "/*". In this case a top Element with name statsName is created and 
		 * the Elements created from the all StatsElement instances in this object are added to the 
		 * top Element as children.
		 */
		if (secondSlash < 0)
		{
			/*
			 * Create a top Element.
			 */
			String topName = this.getNameFromEntryExpression(statsName);
			if (topName.equals("*"))
			{
				topName = statisticsObj.getStatsName();
			}
	
			/*
			 * The statsName can be in theformat "name[1]", i.e., contains index. But the index 
			 * must be 1.
			 */
			int index = this.getIndexFromEntryExpression(statsName);
	
			/*
			 * If statsName contains index, the index only can be 1 and the xpath string looks 
			 * like "/name[1]".
			 */
			if ((entryExpressionContainIndex(statsName)) && (index != 1))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
			
			/*
			 * Obtain schema element for the I_Stats object which contains this object.
			 */
			Element[] valueElement = XmlUtil.createSchemaElement(doc, topName, type, null, null);

			if (valueElement.length == 0)
			{
				return valueElement;
			}

			/*
			 * Above getWriteSchemaElement() returns an sub-tree, subTopElement
			 * defined below is the Element in the sub-tree to which all schema
			 * sub-tree created by the sub-components are mounted. 
			 */
			Element subTopElement;
			if (type.equals("xsd:complexType"))
			{
				subTopElement = (Element) valueElement[0].getChildNodes().item(0);
			}
			else
			{
				//subTopElement = valueElement[0];
				return valueElement;
			}
							
			/*
			 * The Hash table which contains the names of StatsElement instances as keys
			 * and the number of StatsElement instances with the same name as elements.
			 */
			Hashtable nameListHash = new Hashtable();
				
			/*
			 * Iteration on the StatsElement list. For each StatsElement instance
			 * in the vector, get the Element that represents the query result for
			 * this instance, and add the obtained Element to the top Element.
			 */
			for (int i = 0; i < seObjectTable.size(); i++)
			{ 
				StatsElement se = (StatsElement) seObjectTable.elementAt (i);

				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String subname = se.getName();

				if (nameListHash.containsKey(subname))
				{
					Integer value = (Integer) nameListHash.get(subname);
					nameListHash.put(subname, new Integer(value.intValue()+1));
					continue;
				}
						
				XPath newXPath = null;
				try
				{
					newXPath = new XPath(xpathStr+"/"+subname, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
					continue;
				}
						
				Element[] subResultElement = this.getWriteSchemaResult(doc, newXPath);
						
				/*
				 * Add the result Element for this StatsElement instance to the top Element.
				 */
				if ((subResultElement != null) && (subResultElement.length != 0) && (subResultElement[0] != null))
				{
					subTopElement.appendChild(subResultElement[0]);

					if (!nameListHash.containsKey(subname))
					{
						nameListHash.put(subname, new Integer(1));
					}
				}
			} 
	
			/*
			 * The top Element is returned as the result.
			 */
			return valueElement;
		}
		/************************ xpath has value "/name/subxpath" ******************************/
		/*
		 * Otherwise, xpathStr has form "/name/sub-xpath" which points to a StatsElement 
		 * instance defined in this object. The "sub-xpath" is analyzed and corresponding 
		 * action is taken.
		 */
		else
		{
			String subxpath = xpathStr.substring(secondSlash+2);
			int thirdSlash = subxpath.indexOf("/");
					
			String subcomponent;
			String remainedpath = null;
			if (thirdSlash > 0)
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node/State" in OL_Socket, 
				 * for example, then subxpath is "Node/State", and subcomponent is "Node".
				 */
				subcomponent = subxpath.substring(0, thirdSlash);
				remainedpath = subxpath.substring(thirdSlash);
			}
			else
			{
				/*
				 * The xpathStr looks like "/OL_Socket/Node" in OL_Socket, 
				 * for example, then subxpath is "Node", and subcomponent is "Node".
				 */
				subcomponent = subxpath;
				remainedpath = "";
			}
	
			/*
			 * Get the name of subcomponent.
			 */
			String subName = this.getNameFromEntryExpression(subcomponent);
	
			/*
			 * Get the index contained in the string subcomponent. It is -1 if 
			 * no index is contained.
			 */
			int index = this.getIndexFromEntryExpression(subcomponent);

			/*
			 * The xpath string "/name/*[i]" is not supported.
			 */
			if ((subName.equals("*")) && (index > 0))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
	
			/*
			 * Get the Vector that contains the StatsElement instances which 
			 * are readable and match subname.
			 */
			Vector matchedComponents = this.getStatsElement(subName);
	
			/*
			 * If no StatsElement instance is found which matchs subname, or the specified 
			 * index is beyond the range of the list, matchedComponents, throw exception.
			 */
			if ((matchedComponents.size() == 0) || 
				((entryExpressionContainIndex(statsName)) && ((index < 1) || (index > matchedComponents.size()))))
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				return new Element[0];
			}
	
			Vector resultVec = new Vector();
	
			/*
			 * The Hash table which contains the names of StatsElement instances as keys
			 * and the number of StatsElement instances with the same name as elements.
			 */
			Hashtable nameListHash = new Hashtable();
				
			/*
			 * Iteration on the StatsElement instances in matchedComponents. For each StatsElement 
			 * instance in the vector, get the Element that represents the query result for
			 * that instance, and add the obtained Element to the result Element array.
			 */
			for (int i = 0; i < matchedComponents.size(); i++)
			{
				/*
				 * If index is larger than 0, matchedComponents contains a list
				 * of StatsElement instances with the same statsName. The index is
				 * used to find the desired StatsElement instance.
				 * 
				 * Note that index starts from 1 in the xpath string.
				 */
				if ((index > 0) && (index != (i+1)))
				{
					continue;
				}
					
				StatsElement se = (StatsElement) matchedComponents.elementAt (i);
		
				/*
				 * Find the name of this StatsElement instance, and get the query 
				 * result for this instance.
				 */					
				String compname = se.getName();

				if (nameListHash.containsKey(compname))
				{
					Integer value = (Integer) nameListHash.get(compname);
					nameListHash.put(compname, new Integer(value.intValue()+1));
					continue;
				}

				XPath newXPath = null;
				try
				{
					newXPath = new XPath("/"+compname+remainedpath, null, null, XPath.MATCH);
				}
				catch (TransformerException te)
				{
					throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
				}

				Element[] subResult = se.getStatsObj().getWriteSchema(doc, newXPath);

				for (int j=0; j<subResult.length; j++)
				{
					resultVec.addElement(subResult[j]);
				}

				if (!(nameListHash.containsKey(compname)) && (subResult.length > 0))
				{
					nameListHash.put(compname, new Integer(1));
				}
				
				
				if (StatsProcessor.debug)
				  System.out.println("StatsProcessor:getReadSchemaResult:subResult.length="+subResult.length+", subResult[0].getNodeName()="+subResult[0].getNodeName()+", compname="+compname);
				if ((subResult.length == 1) && (compname.equals(subResult[0].getAttribute("name"))))
				{
					se.resetSchemaAttribute(subResult[0]);
				}
				
			}
				
			Element[] valueElement = new Element[resultVec.size()];
	
			for (int i = 0; i < resultVec.size(); i++)
			{ 
				valueElement[i] = (Element) resultVec.elementAt (i);
			}
	
			return valueElement;
		}
    }
}
