/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.adapters;


import java.util.*;

import hypercast.adapters.*;

/** 
 * This class is for storing a set of yet-to-be-processed TimerEvents<p>
 *
 * Implementation: <br>
 *  The queue is stored in a Java Vector - the element with the highest
 *  time is at index 0, the lowest time is at index the_queue.size()-1. 
 *
 * @author HyperCast Team
 * @author  Dongwen Wang
 * @author	Mike D. Nahas
 * @version	2.0, Feb. 28, 2001
 */

public class TimerEventQueue {
    /** Vector of TimeEvents.  
     *  The element with the highest time is at index 0, the lowest time is at index the_queue.size()-1. 
     */ 
    Vector TimeQueue;

    /** Used for random ordering of events that occur at the same time. */
    Random random_number_generator;  

	
	/**
	 * Constructs a TimeEventQueue object
	 */
	public TimerEventQueue() {
		TimeQueue = new Vector();
		random_number_generator = new Random();
	}
    

	/**
	 * Gets the time event with the minimum time
	 * @return	a TimeEvent
	 */
	synchronized public final long getMinEventTime() {
        if (TimeQueue.size() == 0)
            return (long) -1.0;
        else 
            return ((TimerEvent) TimeQueue.lastElement()).getTime();
        
    }
    

	/**
	 * Removes a specified time event from the queue.
	 * @param	id	the id of the time event that to be removed
	 */
	synchronized public final TimerEvent getEvent(Object id) {
        for (int i=0; i<TimeQueue.size(); i++) {
            TimerEvent te = (TimerEvent)TimeQueue.elementAt(i);
            if (te.getTimeID().equals(id)) {
                return te;
            }
        }
        return null;
	}

	
	/**
	 * Removes the top time event in the queue
	 * @return	last TimeEvent in the queue
	 */
	synchronized public final TimerEvent removeTop() {
       TimerEvent return_value = (TimerEvent) TimeQueue.lastElement();
        int s = TimeQueue.size();
        TimeQueue.removeElementAt(s-1);
        return return_value;
    }


	/**
	 * Inserts a time event at a random valid location with the current time at top, 
	 * future events below.
	 */
	synchronized public final void insert(TimerEvent te) {
		int s = TimeQueue.size();
		int top_valid = 0;
		int i;

		for (i = s-1; i >= 0; i--) {
			if (te.getTime() <= ((TimerEvent) TimeQueue.elementAt(i)).getTime()) {
			top_valid = i+1;
		break;
			}
		}

		int bottom_valid = top_valid;
		for (; i >= 0; i--)	{
			if (te.getTime() == ((TimerEvent) TimeQueue.elementAt(i)).getTime())
			bottom_valid = i;
		}

		int insertion_point = bottom_valid;
		if (bottom_valid != top_valid) {
			insertion_point += Math.abs(random_number_generator.nextInt() % (1 + top_valid - bottom_valid));
		}

	/*
	System.out.println();
	System.out.println("Times in queue: ");
	for (i = 0; i < TimeQueue.size(); i++) {
	    System.out.print(i);
	    System.out.print(": ");
	    System.out.println(((TimerEvent) TimeQueue.elementAt(i)).getTime());
	}

	System.out.print("For time :");
	System.out.println(te.getTime());
	System.out.print("Valid insertion places are: ");
	System.out.print(bottom_valid);
	System.out.print(" through ");
	System.out.println(top_valid);
	System.out.print("Chosen: ");
	System.out.println(insertion_point);
	*/

        TimeQueue.insertElementAt(te, insertion_point);
	}


	/**
	 * Removes a specified time event from the queue.
	 * @param	te	the time event that to be removed
	 */
	synchronized public final void remove(TimerEvent te) {
        TimeQueue.removeElement(te);
	}


	/**
	 * Removes a specified time event from the queue.
	 * @param	id	the id of the time event that to be removed
	 */
	synchronized public final TimerEvent remove(Object id) {
        for (int i=0; i<TimeQueue.size(); i++) {
            TimerEvent te = (TimerEvent)TimeQueue.elementAt(i);
            if (te.getTimeID().equals(id)) {
                TimeQueue.removeElementAt(i);
                return te;
            }
        }
        return null;
	}




	/**
	 * Removes all the time events in the queue.
	 */
	synchronized public final void removeEvents(Enumeration e) {
		while (e.hasMoreElements()) {
			TimerEvent delete_me = (TimerEvent) e.nextElement();
			if (isPendingEvent(delete_me)) {
			remove(delete_me);
			}
		}
	}


	/**
	 * Checks if the time event queue is empty.
	 * @return	true if the queue is empty, false otherwise
	 */
	synchronized public final boolean isEmpty() {
		return TimeQueue.isEmpty();
	}

	
	/**
	 * Check if the queue contains a specified time event.
	 * @return	true if the queue is empty, false otherwise
	 * 
	 */
	synchronized public final boolean isPendingEvent(TimerEvent se) {
		return TimeQueue.contains(se);
	}


	/**
	 * Returns all the pending time event int queue.
	 */
	synchronized public final Enumeration getPendingEvents() {
        return TimeQueue.elements();
    }
    

/**
public static void main(String argv[]) {
	TimerEventQueue teq = new TimerEventQueue();
        
    teq.insert(new TimerEvent(1, 3.0f));
    teq.insert(new TimerEvent(2, 5.0f));
    teq.insert(new TimerEvent(3, 1.0f));
    teq.insert(new TimerEvent(4, 2.0f));
    teq.insert(new TimerEvent(-10, 4.0f));
        
    for (int i = 0; i < 5; i++) {
		System.out.println(teq.getMinEventTime());
        teq.removeTop();
	}
}

**/
}
