/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.util;

import java.io.File;
import java.io.FileWriter;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Properties;
import java.util.Enumeration;

/**
 * Dynamically generates a Java source file which contains the default
 * values of all properties used in the specified schema file.
 *
 * This class implements a thin appliction that relies mainly on
 * SchemaDefaultValueCalculator. The application is used during build
 * time to extract constant values from the HyperCast schema
 * object. These values are then written to a dynamically generated
 * Java source file (this file is not found in any source library
 * (e.g. CVS) nor should it ever be edited except by changing this
 * class and running the thin application).
 *
 * By extracting information about constants from the schema file
 * there is a single source of information about constants. This class
 * was implemented in an effort to unify unsynchronized constant
 * information: one version of constant information was kept in the
 * schema, the other was kept in the body of the code.
 *
 * @see SchemaDefaultValueCalculator
 *
 * @author HyperCast Team
 * @version 2005 (version 3.0)
 */
public class ExtractConfigurationDefaults
{
    /**
     * See HyperCast Makefile for usage example.
     * Typical names of files: hypercast.xsd ConfigurationDefaults.java
     */
    public static void main (final String [] args)
        throws MalformedURLException, IOException, FileNotFoundException
    {
        if (args.length != 2)
        {
            System.err.println ("usage: java -classpath jdom.jar ExtractConfigurationDefaults <schema file> <java output file>");
            System.exit (1);
        }
        final String schemaFilename = args[0];
        final String javaFilename = args[1];
        System.out.println ("Extracting configuration constants from schema file \"" + schemaFilename + "\" and generating Java output file \"" + javaFilename + "\"");

        final Properties defaultProperties = (new SchemaDefaultValueCalculator()).getDefaultValues();

        final FileWriter fileWriter = new FileWriter (javaFilename);

        /*
         * Trim leading path information from java filename to extract classname.
         */
        final int lastSeparatorIndex = javaFilename.lastIndexOf (File.separatorChar);
        final int beginningIndexOfFilename = lastSeparatorIndex == -1 ? 0 : lastSeparatorIndex + 1;
        final int endIndexOfFilename = javaFilename.lastIndexOf (".");
        final String classname = javaFilename.substring (beginningIndexOfFilename, endIndexOfFilename);

        final String header =
        	"/***********************************************************************/\n" +
        	"/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */\n" +
        	"/* University of Virginia.                                             */\n" +
        	"/* All rights reserved.                                                */\n" +
        	"/*                                                                     */\n" +
        	"/* This program is being made available under the terms of version 2.1 */\n" +
        	"/* of theGNU Lesser General Public License (\"LGPL\") as published as of */\n" +
        	"/* February, 1999 by the Free Software  Foundation; and subject to the */\n" +
        	"/* restrictions of the LGPL, you are authorized to  modify and/or      */\n" +
        	"/* redistribute it.                                                    */\n" +
        	"/*                                                                     */\n" +
        	"/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE \"AS IS\",*/\n" +
        	"/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */\n" +
        	"/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */\n" +
        	"/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */\n" +
        	"/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */\n" +
        	"/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/\n" +
        	"/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */\n" +
        	"/* See the GNU General Public License for more details.                */\n" +
        	"/*                                                                     */\n" +
        	"/* You should have received a copy of version 2.1 of the GNU General   */\n" +
        	"/* Public License along with this program; if not, you may obtain a    */\n" +
        	"/* copy  from the Free Software Foundation website at the following URL*/\n" +
        	"/* address (link  active as of September 1, 2005):                     */\n" +
        	"/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */\n" +
        	"/* you can request a copy from Free Software Foundation,  Inc., 51     */\n" +
        	"/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */\n" +
        	"/*                                                                     */\n" +
        	"/* This program is part of the HyperCast software distribution. To     */\n" +
        	"/* learn more about HyperCast, please visit the website at             */\n" +
        	"/* http://www.hypercast.org.                                           */\n" +
        	"/***********************************************************************/\n" +
            "\n" +
            "package hypercast;\n" +
            "\n" +
            "import java.util.Properties;\n" +
            "\n" +
            "public class " + classname + "\n" +
            "{\n" +
            "  private static Properties defaultProperties = null;\n" +
            "\n" +
            "  public static synchronized Properties getConfigurationDefaults()\n" +
            "  { \n" +
            "    if (defaultProperties == null)\n" +
            "    {\n" +
            "      defaultProperties = new Properties();\n";

        final StringBuffer body = new StringBuffer();
        for (final Enumeration enumeration = defaultProperties.propertyNames(); enumeration.hasMoreElements();)
        {
            final String propertyName = (String) enumeration.nextElement();
            body.append ("      defaultProperties.setProperty (\"" + propertyName + "\", \"" + defaultProperties.getProperty (propertyName) +"\");\n");
        }

        /*
         * TODO: Figure out a way to make the constant properties
         * immutable/unmodifiable.  The issue here is that
         * although Properties implements Map, Maps are not
         * Properties. That is, an upcast is needed from Map to
         * Properties.
         */
        final String footer =
            //"      defaultProperties = Collections.unmodifiableMap (defaultProperties);\n" +
            "    }\n" + 
            "    return defaultProperties;\n" +
            "  }\n" +
            "}\n";

        final String dynamicJava = header + body + footer;
        fileWriter.write (dynamicJava);
        fileWriter.close();
    }
}
