/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.util;


import hypercast.HyperCastStatsException;
import hypercast.I_Stats;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.io.InputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xml.sax.SAXException;

import java.io.StringWriter;
import java.text.ParseException;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;

import javax.xml.transform.dom.DOMSource;

import javax.xml.transform.stream.StreamResult;



/**
 * This is a class to implement some special functions
 * for Xml document, for example,  writing a
 * Xml document to a file.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, 
 * @author Jianping Wang, 
 * 
 * @version 3.0, Mar 18, 2003
 */

public class XmlUtil
{
    /** Used by JAXP parser to do XML->DOM and DOM->XML translation. */
    private static DocumentBuilderFactory docBuilderFactory;

    /** Used by JAXP parser to do XML->DOM and DOM->XML translation. */
    private static DocumentBuilder docBuilder;
	
    static
    {
        try
        {
            docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilder = docBuilderFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException pce)
        {
            final String errorString =
                "javax.xml.parsers.ParserConfigurationException in XmlUtil: " + pce.toString() + "\n" +
                "Unable to create docBuilder.\n" +
                "Make sure the JAXP parser has been correctly installed.\n" +
                "For more on JAXP see http://java.sun.com/xml/\n" +
                "Exiting.\n";

            System.err.println (errorString);
            System.exit (1);
        }
    }

    /** 
     * Creates a new Document class for the root of DOM trees.  
     */
    public static synchronized Document createDocument()
    {
        return docBuilder.newDocument();
    }
	
    /* *
     * Creates a new Document class from given XML file.
     * 
     * @param xmlFilename	a XML file.
     *
     * @throws FileNotFoundException If the file cannot be opened.
     * @throws NullPointerException If the argument is <code>null</code>
     * @throws ParseException If the string is not a valid XML Document
     */

    public static Document createDocument (final String xmlFilename)
        throws FileNotFoundException, ParseException, IOException
    {
        if (xmlFilename == null)
            throw new NullPointerException();
	
        final File xmlFile = new File (xmlFilename);
        return createDocument (xmlFile);
    }

	
    /**
     * Creates a new Document class from given XML File instance.
     * 
     * @param xmlFile  a XML File instance.
     *
     * @throws FileNotFoundException If the file cannot be opened.
     * @throws NullPointerException If the argument is <code>null</code>
     * @throws IOException on parse error
     */

    public static Document createDocument (final File xmlFile)
        throws FileNotFoundException, ParseException, IOException
    {
        if (xmlFile == null)
            throw new NullPointerException();

		try
		{
             final FileInputStream fileInputStream = new FileInputStream (xmlFile);
		     return createDocument (fileInputStream, xmlFile.getName());
		}
		catch (final FileNotFoundException fnfe)
		  {
			  throw fnfe;
		  }
		  catch (final IOException ioe)
		  {
			  throw new ParseException ("Error when creating XML document from file \"" + xmlFile.getName() + "\". Cause: " + ioe, 0);
		  }
    }

    /**
     * Creates a new Document class from given XML File instance.
     */
    public static Document createDocument(InputStream xmlFileStream, String fileName)
	     throws IOException
	 {		
		final Document doc;
		synchronized (XmlUtil.class)
		{
		     try
			 {
			     doc = docBuilder.parse (xmlFileStream);
			 }
			 catch (IOException ioe)
		     {
				 throw new IOException ("Fails in creating Document instance from file: " + fileName + " (using input stream). Cause: "+ioe);
			 }
			 catch (final SAXException se)
			 {
			    throw new IOException ("parse error for XML file: \"" + fileName + "\"\nCause: " + se);
			 }
		}
		
		removeWhitespace (doc);
		removeComments(doc, doc);
	
	    return doc;
	 }

    /** 
     * Creates a new Document class from given XML string.
     * 
     * @param xmlDocString a XML document in string format.
     *
     * @see XmlUtil#getStringFromDocument(Document)
     */
    //* @throws RuntimeException if any problem occured while reading the XML Document from a byte array.

    public static Document createDocumentFromString (final String xmlDocString)
	throws ParseException
    {
        return createDocument (xmlDocString.getBytes());
    }

    public static Document createDocument (final byte [] array)
	   throws ParseException
   {
   		try
   		{
		   ByteArrayInputStream bais = new ByteArrayInputStream (array);
		   return createDocument ((InputStream) bais, "");
   		}
		catch (final IOException ioe)
		{
		   throw new ParseException ("Error when creating XML document from byte array. Cause: " + ioe, 0);
		}
    }

 /**
  * Method to convert Document to String.
  * 
  * @param doc a Document instance.
  *
  * @see XmlUtil#createDocumentFromString(String)
  */
  public static String getStringFromDocument (Document doc)
  {
	  try
	  {
		 DOMSource domSource = new DOMSource(doc);
		 StringWriter writer = new StringWriter();
		 StreamResult result = new StreamResult(writer);
		 TransformerFactory tf = TransformerFactory.newInstance();
		 Transformer transformer = tf.newTransformer();
		 transformer.transform(domSource, result);
		 return writer.toString();
	  }
	  catch (TransformerException ex)
	  {
		 ex.printStackTrace();
		 return null;
	  }
  }        	

/** Remove comments in a document.
 * 
 * @param d	A Document instance.
 * @param n	Node which represents the subtree from which comments
 * 	are removed.
 */
public static void removeComments(Document d, Node n)
{
	if ((n.getNodeName().equals("#comment")) ||
		((n.getNodeType() == Node.TEXT_NODE) && (n.getNodeValue().trim().length() == 0)))
	{
		Node parent = n.getParentNode();
		if (parent != null)
		{
		  parent.removeChild(n);
		}
		return;
	}
	  
	NodeList children = n.getChildNodes();
	  
	Vector childVec = new Vector();
	for (int i = 0; i < children.getLength(); i++) {
		childVec.addElement(children.item(i));
	}

	for (int i = 0; i < childVec.size(); i++) {
		Node child = (Node) childVec.elementAt(i);
		removeComments(d, child);
	}
}

    /**
     * Remove excess white space in a document.  
     * This function removes all text nodes consisting of only whitespace.  
     * It will also remove all whitespace at the beginning or end of any other text. 
     */
    private static void removeWhitespace (final Node node) // can pass a Document here
    {
        final NodeList children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++)
        {
            final Node child = children.item (i);

            if (child.getNodeType() != Node.TEXT_NODE)
            {
                removeWhitespace (child);
            }
            else
            {
                final String value = child.getNodeValue().trim();

                if (value.equals (""))
                {
                    node.removeChild (child);
                    i--; // so we don't skip the following node.
                }
                else
                {
                    child.setNodeValue (value);
                }
            }
        }
    }

    /**
     * Write Xml document to output stream
     *
     * @param saveDoc	the Xml document to write
     * @param output	the output stream 
     * @throws IOException
     * @throws IOException	if writing fails
     */
    // Perhaps call remove whitespace here?
    public static void writeXml (final Document saveDoc, final OutputStream output) throws IOException{
    	writeXml(saveDoc, output, false);
    }
    
    /**
     *	We process the XML Document in different ways when under different JRE version
     *  So we use this class the control the output of the compiler 
     */
    /* package */ class XmlVersionControl
	{
	    static final int SUN_JRE_1_4=0;
	    static final int JEODE_JRE_1_1=1;
	    
	    static final int JRE_VERSION=SUN_JRE_1_4;
	}
    /**
     * Write Xml document to output stream
     *
     *
     * @param saveDoc	the Xml document to write
     * @param output	the output stream
     * @param omit_declaration	whether to omit the XML delcaration when outputing. 
     *                          Used "true "when want to put several XML docs in a single file.
     * @throws IOException
     * @throws IOException	if writing fails
     */
    public static void writeXml (final Document saveDoc, final OutputStream output, boolean omit_declaration)
        throws IOException
    {
    	if (XmlVersionControl.JRE_VERSION == XmlVersionControl.JEODE_JRE_1_1)
        {
            try
            {
                ((org.apache.crimson.tree.XmlDocument) saveDoc).write(output);
            }
            catch (ClassCastException ce2)
            {
                //System.err.println("ERROR: DOM Socket was passed an unknown type");
                //System.err.println(": "+ ce2.getMessage());
                throw new IOException ("DOM type unknown");
            }
        }else {
            try
            {
                // Prepare the DOM document for writing

                javax.xml.transform.Source source = new javax.xml.transform.dom.DOMSource (saveDoc);
                javax.xml.transform.Result result = new javax.xml.transform.stream.StreamResult (output);
	    
                // Write the DOM document to the file

                javax.xml.transform.Transformer xformer = javax.xml.transform.TransformerFactory.newInstance().newTransformer();

                xformer.setOutputProperty (javax.xml.transform.OutputKeys.INDENT, "yes");
                if(omit_declaration){
                	xformer.setOutputProperty (javax.xml.transform.OutputKeys.OMIT_XML_DECLARATION, "yes");
                }
                
                xformer.transform (source, result);
            }
            catch (Exception e)
            {
                System.err.println ("VERY, VERY BAD IN XmlUtil!: " + e);
                e.printStackTrace (System.err);
                throw new IOException ("DOM cannot be written: " + e);
            }
        }
        
    }

	/**
	 * Find a children node by name
	 * @param node	the node to find the child node in
	 * @param name	the name of child node to find
	 * @return	the child node, or null if not found
	 */
    // TODO: uncommented, possibly temporarily - used by RunControl currently

    public static Node findChildNode(Node node, String name){
        for(int i=0;i<node.getChildNodes().getLength();i++){
            if(node.getChildNodes().item(i).getNodeName().equals(name))
                return node.getChildNodes().item(i);
        }    
        return null;
    }

    /**
     * clear all subnodes of a node, like "rm -rf" but for an XML document/node
     * @param node
     */
    // TODO: uncommented, possibly temporarily
/*
    public static void clean(Node node)
    {
        NodeList nl=node.getChildNodes();
        
        for(int i=0;i<nl.getLength();i++)
        {
            clean(nl.item(i));
            if (nl.item(i).getNodeType()!=Node.ELEMENT_NODE)
            {
                node.removeChild(nl.item(i));
            }
            else
            {
                clean(nl.item(i));
            }
        }        
    }
*/

    /**
     * change the owner of a node and all its subnodes
     * @param doc	the new owner
     * @param node	the node to change owner of
     */
    // TODO: uncommented, possibly temporarily - used by RunControl currently

    public static void changeNodeOwner(Document doc, Node node)
    {
    	NodeList nl=node.getChildNodes();
        
        try
        {
            ((org.apache.crimson.tree.XmlDocument) doc).changeNodeOwner(node);
        }
        catch(ClassCastException ce2)
        {
            System.err.println("ERROR: DOM document was passed an unknown type");
            System.err.println(": "+ ce2.getMessage());
            //throw new IOException("DOM type unknow");
        }
        catch(Exception e2)
        {
        }
		        
        //change the owner of children nodes
        for(int i=0;i<nl.getLength();i++){
            
            changeNodeOwner(doc, nl.item(i));
        }
    }

	/**
	 * A static function to get the stats in an old fashion way
	 * to eliminate the hassle of dealing with XML
	 * @param stats		the I_Stas object
	 * @param name		the name of the stats
	 * @return			the String value of the stats if the result is not a tree
	 * @throws TransformerException
	 * @throws HyperCastStatsException
	 */
	public static String getScalarStatsResult(I_Stats stats, String name) throws HyperCastStatsException, TransformerException{
		Document doc=XmlUtil.createDocument();
		Element[] result=stats.getStats(doc, new XPath (name, null, null, XPath.MATCH));
		return getScalarValue(result);
	}

	/**
	 * Extract scalar value from a single Element result
	 */
	public static String getScalarValue(Element result){
		return getScalarValue(valueToList(result));
	}

	/**
	 * Make a single Element value to a Element list
	 */
	static public Element [] valueToList(Element value){
		Element []vl = new Element[1];
		vl[0] = value;
		return vl;
	}

	public static String getScalarValue(Element result[]){
		if(result == null || result.length ==0)
			return null;
			
		return result[0].getChildNodes().item(0).getNodeValue();
	}

	/** 
	 * Return the element which represents a statistic in the schema document. 
	 * @param doc		The document to create the Element
	 * @param name		The name of the created Element
	 * @param type		The type of the statistic. It can be either SimpleType or ComplexType.
	 * @param restrictionBase	When type is "simpleType", it is to specify the specific type
	 * 					of this statistic, such as "xsd:string". For ComplexType statistic, it
	 * 					is null.
	 * @param patternValue	When type is "simpleType", it is to specify the format of the 
	 * 					value under the specific restriction base. For ComplexType statistic, it
	 * 					is null.
	 * 
	 * @return			new created Element
	 */
	public static Element[] createSchemaElement(Document doc, String name, String type, String restrictionBase, String patternValue)
	{
		if (name == null)
		{
			return new Element[0];
		}
	
		Element sname = doc.createElement("xsd:element");
	
		sname.setAttribute("name", name);
	
		if (type.equals("xsd:complexType"))
		{
			Element complextypeElement = doc.createElement(type);
			sname.appendChild(complextypeElement);
		}
		else if (type.equals("xsd:simpleType"))
		{
			Element simpletypeElement = doc.createElement(type);
			sname.appendChild(simpletypeElement);
				
			Element restrictionBaseElement = doc.createElement("xsd:restriction");
			restrictionBaseElement.setAttribute("base", restrictionBase);
			simpletypeElement.appendChild(restrictionBaseElement);
				
			Element patternValueElement = doc.createElement("xsd:pattern");
			patternValueElement.setAttribute("value", patternValue);
			restrictionBaseElement.appendChild(patternValueElement);
	
			sname.appendChild(simpletypeElement);
		}
		else
		{
			sname.setAttribute("type", type);
		}
	
		Element[] result = new Element[1];
		result[0] = sname;
		return result;
	}

	/**
	 * @return XPath instance cooresponding to the given string.
	 * 
	 * @param xpathStr xpath string.
	 */
	public static XPath createXPath(final String xpathStr)
	{
		try
		{
			return new XPath(xpathStr, null, null, XPath.MATCH);
		}
		catch (TransformerException te)
		{
			//throw new HyperCastConfigException ("\""+xpathStr+"\" is not a valid XPath string.");
			System.err.println("\""+xpathStr+"\" is not a valid XPath string.");
			return null;
		}
	}

	/**
	 * Convert a String to an Xml Element that can be passed to a I_Stats
	 * @param doc		The document to create the Element
	 * @param xpath		The xpath string
	 * @param value		The String value to convert
	 * @return			The Element value
	 */
	public static Element getXmlValue(Document doc, String xpath, String value){
		return getXmlValues(doc, xpath, value)[0];
	}

	public static Element []getXmlValues(Document doc, String xpath, String value){
		
		Element setValue = null;
		
	    final String [] nameField = xpath.split ("/"); // Java 1.4 feature
	
	    if (nameField.length > 0)
	    {
	        final String lastField = nameField [nameField.length - 1];
	        final int sIndex = lastField.indexOf ("[");
	        final int eIndex = lastField.indexOf ("]");
	        
	        if (sIndex < 0
	            &&
	            eIndex < 0)
	        {
	            setValue = doc.createElement (lastField);
	        }
	        else if (sIndex > 0
	                 &&
	                 eIndex > sIndex)
	        {
	            final String elementName = lastField.substring (0, sIndex);
	            setValue = doc.createElement (elementName);
	            setValue.setAttribute ("index", new Integer (lastField.substring (sIndex + 1, eIndex)).toString());
	        }
	        else
	        {
	            throw new IllegalArgumentException ("After splitting xpath, name was found to be invalid when attempting to add scalar stat: " + xpath);
	        }
	        setValue.appendChild (doc.createTextNode (value));
	    }
	    else
	    {
	        throw new IllegalArgumentException ("XPath was found to be invalid when attempting to add scalar stat: " + xpath);
	    }
	    
	    Element []vl = new Element[1];
	    vl[0] = setValue;
	    return vl;
	}

	/**
	 * A static function to set the stats in an old fashion way
	 * to eliminate the hassle of dealing with XML
	 * @param stats		the stats object
	 * @param value		the new value in String format
	 * @return			the value of stats after set
	 * @throws HyperCastStatsException
	 * @throws TransformerException
	 */
	public static String setScalarStatsResult(I_Stats stats, String xpath, String value) throws HyperCastStatsException, TransformerException{
		Document doc=XmlUtil.createDocument();
	    
		Element[] result=stats.setStats(doc, new XPath (xpath, null, null, XPath.MATCH), getXmlValue(doc, xpath, value));
		return result[0].getChildNodes().item(0).getNodeValue();
	}

}
